@@ -3,4 +3,4 @@
 executable="nexuiz-osx-ppc-dedicated"
 
 cd "`dirname \"${0}\"`"
-exec ./${executable} "${@}" +exec server.cfg
+exec ./${executable} +exec server.cfg "${@}"
