@@ -823,14 +823,23 @@ void Collision_TraceLineBrushFloat(trace_t *trace, const vec3_t linestart, const
 	}
 }
 
-void Collision_TracePointBrushFloat(trace_t *trace, const vec3_t point, const colbrushf_t *thatbrush)
+qboolean Collision_PointInsideBrushFloat(const vec3_t point, const colbrushf_t *brush)
 {
 	int nplane;
 	const colplanef_t *plane;
 
-	for (nplane = 0, plane = thatbrush->planes;nplane < thatbrush->numplanes;nplane++, plane++)
+	if (!BoxesOverlap(point, point, brush->mins, brush->maxs))
+		return false;
+	for (nplane = 0, plane = brush->planes;nplane < brush->numplanes;nplane++, plane++)
 		if (DotProduct(plane->normal, point) > plane->dist)
-			return;
+			return false;
+	return true;
+}
+
+void Collision_TracePointBrushFloat(trace_t *trace, const vec3_t point, const colbrushf_t *thatbrush)
+{
+	if (!Collision_PointInsideBrushFloat(point, thatbrush))
+		return;
 
 	trace->startsupercontents |= thatbrush->supercontents;
 	if (trace->hitsupercontentsmask & thatbrush->supercontents)
@@ -1002,7 +1011,9 @@ void Collision_TraceBrushPolygonTransformFloat(trace_t *trace, const colbrushf_t
 
 
 #define MAX_BRUSHFORBOX 16
-static int brushforbox_index = 0;
+static unsigned int brushforbox_index = 0;
+// note: this relies on integer overflow to be consistent with modulo
+// MAX_BRUSHFORBOX, or in other words, MAX_BRUSHFORBOX must be a power of two!
 static colpointf_t brushforbox_point[MAX_BRUSHFORBOX*8];
 static colplanef_t brushforbox_plane[MAX_BRUSHFORBOX*6];
 static colbrushf_t brushforbox_brush[MAX_BRUSHFORBOX];
