@@ -27,8 +27,8 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 // MSVC has a different name for several standard functions
 #ifdef WIN32
-# define strcasecmp stricmp
-# define strncasecmp strnicmp
+# define strcasecmp _stricmp
+# define strncasecmp _strnicmp
 #endif
 
 // Create our own define for Mac OS X
@@ -60,6 +60,7 @@ void SZ_HexDumpToConsole(const sizebuf_t *buf);
 void Com_HexDumpToConsole(const unsigned char *data, int size);
 
 unsigned short CRC_Block(const unsigned char *data, size_t size);
+unsigned short CRC_Block_CaseInsensitive(const unsigned char *data, size_t size); // for hash lookup functions that use strcasecmp for comparison
 
 unsigned char COM_BlockSequenceCRCByteQW(unsigned char *base, int length, int sequence);
 
@@ -149,6 +150,9 @@ typedef enum protocolversion_e
 	PROTOCOL_NEHAHRAMOVIE, // Nehahra movie protocol, a big nasty hack dating back to early days of the Quake Standards Group (but only ever used by neh_gl.exe), this is potentially backwards compatible with quake protocol as long as no extended features are used (but in actuality the neh_gl.exe which wrote this protocol ALWAYS wrote the extended information)
 	PROTOCOL_QUAKE, // quake (aka netquake/normalquake/nq) protocol
 	PROTOCOL_QUAKEWORLD, // quakeworld protocol
+	PROTOCOL_NEHAHRABJP, // same as QUAKEDP but with 16bit modelindex
+	PROTOCOL_NEHAHRABJP2, // same as NEHAHRABJP but with 16bit soundindex
+	PROTOCOL_NEHAHRABJP3, // same as NEHAHRABJP2 but with some changes
 }
 protocolversion_t;
 
@@ -200,15 +204,15 @@ float MSG_ReadAngle (protocolversion_t protocol);
 
 //============================================================================
 
-typedef float (*COM_CharWidthFunc_t) (void *passthrough, char c);
+typedef float (*COM_WordWidthFunc_t) (void *passthrough, const char *w, size_t *length, float maxWidth); // length is updated to the longest fitting string into maxWidth; if maxWidth < 0, all characters are used and length is used as is
 typedef int (*COM_LineProcessorFunc) (void *passthrough, const char *line, size_t length, float width, qboolean isContination);
-int COM_Wordwrap(const char *string, size_t length, float continuationSize, float maxWidth, COM_CharWidthFunc_t charWidth, void *passthroughCW, COM_LineProcessorFunc processLine, void *passthroughPL);
+int COM_Wordwrap(const char *string, size_t length, float continuationSize, float maxWidth, COM_WordWidthFunc_t wordWidth, void *passthroughCW, COM_LineProcessorFunc processLine, void *passthroughPL);
 
 extern char com_token[MAX_INPUTLINE];
 
-int COM_ParseToken_Simple(const char **datapointer, int returnnewline);
-int COM_ParseToken_QuakeC(const char **datapointer, int returnnewline);
-int COM_ParseToken_VM_Tokenize(const char **datapointer, int returnnewline);
+int COM_ParseToken_Simple(const char **datapointer, qboolean returnnewline, qboolean parsebackslash);
+int COM_ParseToken_QuakeC(const char **datapointer, qboolean returnnewline);
+int COM_ParseToken_VM_Tokenize(const char **datapointer, qboolean returnnewline);
 int COM_ParseToken_Console(const char **datapointer);
 
 extern int com_argc;
@@ -271,11 +275,14 @@ typedef enum gamemode_e
 	GAME_NEOTERIC,
 	GAME_OPENQUARTZ, //this game sucks
 	GAME_PRYDON,
-	GAME_NETHERWORLD,
+	GAME_DELUXEQUAKE,
 	GAME_THEHUNTED,
 	GAME_DEFEATINDETAIL2,
 	GAME_DARSANA,
 	GAME_CONTAGIONTHEORY,
+	GAME_EDU2P,
+	GAME_BLADEMASTER,
+	GAME_COUNT
 }
 gamemode_t;
 
@@ -309,7 +316,7 @@ typedef struct stringlist_s
 int matchpattern(const char *in, const char *pattern, int caseinsensitive);
 void stringlistinit(stringlist_t *list);
 void stringlistfreecontents(stringlist_t *list);
-void stringlistappend(stringlist_t *list, char *text);
+void stringlistappend(stringlist_t *list, const char *text);
 void stringlistsort(stringlist_t *list);
 void listdirectory(stringlist_t *list, const char *path);
 
