@@ -3,6 +3,8 @@
 #include "clprogdefs.h"
 #include "csprogs.h"
 #include "cl_collision.h"
+#include "snd_main.h"
+#include "clvm_cmds.h"
 
 //============================================================================
 // Client prog handling
@@ -20,7 +22,6 @@ static char *cl_required_func[] =
 	"CSQC_InputEvent",
 	"CSQC_UpdateView",
 	"CSQC_ConsoleCommand",
-	"CSQC_Shutdown"
 };
 
 static int cl_numrequiredfunc = sizeof(cl_required_func) / sizeof(char*);
@@ -45,10 +46,32 @@ void CL_VM_Error (const char *format, ...)	//[515]: hope it will be never execut
 //	Host_AbortCurrentFrame();	//[515]: hmmm... if server says it needs csqc then client MUST disconnect
 	Host_Error(va("CL_VM_Error: %s", errorstring));
 }
-
+void CL_VM_UpdateDmgGlobals (int dmg_take, int dmg_save, vec3_t dmg_origin)
+{
+	prvm_eval_t *val;
+	if(cl.csqc_loaded)
+	{
+		CSQC_BEGIN
+		val = PRVM_GLOBALFIELDVALUE(prog->globaloffsets.dmg_take);
+		if(val)
+			val->_float = dmg_take;
+		val = PRVM_GLOBALFIELDVALUE(prog->globaloffsets.dmg_save);
+		if(val)
+			val->_float = dmg_save;
+		val = PRVM_GLOBALFIELDVALUE(prog->globaloffsets.dmg_origin);
+		if(val)
+		{
+			val->vector[0] = dmg_origin[0];
+			val->vector[1] = dmg_origin[1];
+			val->vector[2] = dmg_origin[2];
+		}
+		CSQC_END
+	}
+}
 //[515]: set globals before calling R_UpdateView, WEIRD CRAP
 static void CSQC_SetGlobals (void)
 {
+	prvm_eval_t *val;
 	CSQC_BEGIN
 		prog->globals.client->time = cl.time;
 		prog->globals.client->frametime = max(0, cl.time - cl.oldtime);
@@ -56,14 +79,21 @@ static void CSQC_SetGlobals (void)
 		prog->globals.client->clientcommandframe = cl.movecmd[0].sequence;
 		VectorCopy(cl.viewangles, prog->globals.client->input_angles);
 		VectorCopy(cl.viewangles, cl.csqc_angles);
+		// // FIXME: this actually belongs into getinputstate().. [12/17/2007 Black]
 		prog->globals.client->input_buttons = cl.movecmd[0].buttons;
 		VectorSet(prog->globals.client->input_movevalues, cl.movecmd[0].forwardmove, cl.movecmd[0].sidemove, cl.movecmd[0].upmove);
 		//VectorCopy(cl.movement_origin, cl.csqc_origin);
 		Matrix4x4_OriginFromMatrix(&cl.entities[cl.viewentity].render.matrix, cl.csqc_origin);
+
+		// LordHavoc: Spike says not to do this, but without pmove_org the
+		// CSQC is useless as it can't alter the view origin without
+		// completely replacing it
 		VectorCopy(cl.csqc_origin, prog->globals.client->pmove_org);
-		prog->globals.client->maxclients = cl.maxclients;
-		//VectorCopy(cl.movement_velocity, prog->globals.client->pmove_vel);
 		VectorCopy(cl.velocity, prog->globals.client->pmove_vel);
+
+		if ((val = PRVM_GLOBALFIELDVALUE(prog->globaloffsets.view_angles)))
+			VectorCopy(cl.viewangles, val->vector);
+		prog->globals.client->maxclients = cl.maxclients;
 	CSQC_END
 }
 
@@ -96,9 +126,10 @@ extern cvar_t cl_noplayershadow;
 qboolean CSQC_AddRenderEdict(prvm_edict_t *ed)
 {
 	int renderflags;
+	int c;
 	float scale;
 	prvm_eval_t *val;
-	entity_t *e;
+	entity_render_t *entrender;
 	model_t *model;
 	matrix4x4_t tagmatrix, matrix2;
 
@@ -106,22 +137,20 @@ qboolean CSQC_AddRenderEdict(prvm_edict_t *ed)
 	if (!model)
 		return false;
 
-	e = CL_NewTempEntity();
-	if (!e)
+	entrender = CL_NewTempEntity();
+	if (!entrender)
 		return false;
 
-	e->render.model = model;
-	e->render.colormap = (int)ed->fields.client->colormap;
-	e->render.frame = (int)ed->fields.client->frame;
-	e->render.skinnum = (int)ed->fields.client->skin;
-	e->render.effects |= e->render.model->effects;
+	entrender->model = model;
+	entrender->skinnum = (int)ed->fields.client->skin;
+	entrender->effects |= entrender->model->effects;
 	scale = 1;
 	renderflags = 0;
 	if((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.renderflags)) && val->_float)	renderflags = (int)val->_float;
-	if((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.alpha)) && val->_float)		e->render.alpha = val->_float;
-	if((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.scale)) && val->_float)		e->render.scale = scale = val->_float;
-	if((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.colormod)) && VectorLength2(val->vector))	VectorCopy(val->vector, e->render.colormod);
-	if((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.effects)) && val->_float)	e->render.effects |= (int)val->_float;
+	if((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.alpha)) && val->_float)		entrender->alpha = val->_float;
+	if((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.scale)) && val->_float)		entrender->scale = scale = val->_float;
+	if((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.colormod)) && VectorLength2(val->vector))	VectorCopy(val->vector, entrender->colormod);
+	if((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.effects)) && val->_float)	entrender->effects |= (int)val->_float;
 	if((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.tag_entity)) && val->edict)
 	{
 		int tagentity;
@@ -129,8 +158,7 @@ qboolean CSQC_AddRenderEdict(prvm_edict_t *ed)
 		tagentity = val->edict;
 		if((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.tag_index)) && val->_float)
 			tagindex = (int)val->_float;
-		// FIXME: calculate tag matrix
-		Matrix4x4_CreateIdentity(&tagmatrix);
+		CL_GetTagMatrix (&tagmatrix, PRVM_PROG_TO_EDICT(tagentity), tagindex);
 	}
 	else
 		Matrix4x4_CreateIdentity(&tagmatrix);
@@ -147,7 +175,7 @@ qboolean CSQC_AddRenderEdict(prvm_edict_t *ed)
 		vec3_t angles;
 		VectorCopy(ed->fields.client->angles, angles);
 		// if model is alias, reverse pitch direction
-		if (e->render.model->type == mod_alias)
+		if (entrender->model->type == mod_alias)
 			angles[0] = -angles[0];
 
 		// set up the render matrix
@@ -159,57 +187,78 @@ qboolean CSQC_AddRenderEdict(prvm_edict_t *ed)
 	// self.frame1time is the animation base time for the interpolation target
 	// self.frame2 is the interpolation start (previous frame)
 	// self.frame2time is the animation base time for the interpolation start
-	e->render.frame1 = e->render.frame = ed->fields.client->frame;
-	if ((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.frame2))) e->render.frame1 = val->_float;
-	if ((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.frame1time))) e->render.frame2time = val->_float;
-	if ((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.frame2time))) e->render.frame1time = val->_float;
-	if ((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.lerpfrac))) e->render.framelerp = val->_float;
+	entrender->frame1 = entrender->frame2 = ed->fields.client->frame;
+	if ((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.frame2))) entrender->frame2 = val->_float;
+	if ((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.frame1time))) entrender->frame2time = val->_float;
+	if ((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.frame2time))) entrender->frame1time = val->_float;
+	if ((val = PRVM_EDICTFIELDVALUE(ed, prog->fieldoffsets.lerpfrac))) entrender->framelerp = val->_float;
 
 	// concat the matrices to make the entity relative to its tag
-	Matrix4x4_Concat(&e->render.matrix, &tagmatrix, &matrix2);
-	// make the other useful stuff
-	CL_UpdateRenderEntity(&e->render);
+	Matrix4x4_Concat(&entrender->matrix, &tagmatrix, &matrix2);
 
 	if(renderflags)
 	{
-		if(renderflags & RF_VIEWMODEL)	e->render.flags |= RENDER_VIEWMODEL;
-		if(renderflags & RF_EXTERNALMODEL)e->render.flags |= RENDER_EXTERIORMODEL;
-		if(renderflags & RF_DEPTHHACK)	e->render.effects |= EF_NODEPTHTEST;
-		if(renderflags & RF_ADDITIVE)		e->render.effects |= EF_ADDITIVE;
+		if(renderflags & RF_VIEWMODEL)	entrender->flags |= RENDER_VIEWMODEL;
+		if(renderflags & RF_EXTERNALMODEL)entrender->flags |= RENDER_EXTERIORMODEL;
+		if(renderflags & RF_DEPTHHACK)	entrender->effects |= EF_NODEPTHTEST;
+		if(renderflags & RF_ADDITIVE)		entrender->effects |= EF_ADDITIVE;
 	}
 
+	c = (int)ed->fields.client->colormap;
+	if (c <= 0)
+		CL_SetEntityColormapColors(entrender, -1);
+	else if (c <= cl.maxclients && cl.scores != NULL)
+		CL_SetEntityColormapColors(entrender, cl.scores[c-1].colors);
+	else
+		CL_SetEntityColormapColors(entrender, c);
+
 	// either fullbright or lit
-	if (!(e->render.effects & EF_FULLBRIGHT) && !r_fullbright.integer)
-		e->render.flags |= RENDER_LIGHT;
+	if (!(entrender->effects & EF_FULLBRIGHT) && !r_fullbright.integer)
+		entrender->flags |= RENDER_LIGHT;
 	// hide player shadow during intermission or nehahra movie
-	if (!(e->render.effects & (EF_NOSHADOW | EF_ADDITIVE | EF_NODEPTHTEST))
-	 &&  (e->render.alpha >= 1)
-	 && !(e->render.flags & RENDER_VIEWMODEL)
-	 && (!(e->render.flags & RENDER_EXTERIORMODEL) || (!cl.intermission && cls.protocol != PROTOCOL_NEHAHRAMOVIE && !cl_noplayershadow.integer)))
-		e->render.flags |= RENDER_SHADOW;
-	if (e->render.flags & RENDER_VIEWMODEL)
-		e->render.flags |= RENDER_NOSELFSHADOW;
+	if (!(entrender->effects & (EF_NOSHADOW | EF_ADDITIVE | EF_NODEPTHTEST))
+	 &&  (entrender->alpha >= 1)
+	 && !(entrender->flags & RENDER_VIEWMODEL)
+	 && (!(entrender->flags & RENDER_EXTERIORMODEL) || (!cl.intermission && cls.protocol != PROTOCOL_NEHAHRAMOVIE && !cl_noplayershadow.integer)))
+		entrender->flags |= RENDER_SHADOW;
+	if (entrender->flags & RENDER_VIEWMODEL)
+		entrender->flags |= RENDER_NOSELFSHADOW;
+
+	// make the other useful stuff
+	CL_UpdateRenderEntity(entrender);
 
 	return true;
 }
 
-qboolean CL_VM_InputEvent (qboolean pressed, int key)
+qboolean CL_VM_InputEvent (qboolean down, int key, int ascii)
 {
 	qboolean r;
+
 	if(!cl.csqc_loaded)
 		return false;
+
 	CSQC_BEGIN
-		prog->globals.client->time = cl.time;
-		PRVM_G_FLOAT(OFS_PARM0) = pressed;
-		PRVM_G_FLOAT(OFS_PARM1) = key;
-		PRVM_ExecuteProgram(prog->funcoffsets.CSQC_InputEvent, "QC function CSQC_InputEvent is missing");
-		r = CSQC_RETURNVAL;
+		if (!prog->funcoffsets.CSQC_InputEvent)
+			r = false;
+		else
+		{
+			prog->globals.client->time = cl.time;
+			PRVM_G_FLOAT(OFS_PARM0) = !down; // 0 is down, 1 is up
+			PRVM_G_FLOAT(OFS_PARM1) = key;
+			PRVM_G_FLOAT(OFS_PARM2) = ascii;
+			PRVM_ExecuteProgram(prog->funcoffsets.CSQC_InputEvent, "QC function CSQC_InputEvent is missing");
+			r = CSQC_RETURNVAL;
+		}
 	CSQC_END
 	return r;
 }
 
 qboolean CL_VM_UpdateView (void)
 {
+	vec3_t emptyvector;
+	emptyvector[0] = 0;
+	emptyvector[1] = 0;
+	emptyvector[2] = 0;
 //	vec3_t oldangles;
 	if(!cl.csqc_loaded)
 		return false;
@@ -219,10 +268,12 @@ qboolean CL_VM_UpdateView (void)
 		CSQC_SetGlobals();
 		// clear renderable entity and light lists to prevent crashes if the
 		// CSQC_UpdateView function does not call R_ClearScene as it should
-		r_refdef.numentities = 0;
-		r_refdef.numlights = 0;
+		r_refdef.scene.numentities = 0;
+		r_refdef.scene.numlights = 0;
 		PRVM_ExecuteProgram(prog->funcoffsets.CSQC_UpdateView, "QC function CSQC_UpdateView is missing");
 		//VectorCopy(oldangles, cl.viewangles);
+		// Dresk : Reset Dmg Globals Here
+		CL_VM_UpdateDmgGlobals(0, 0, emptyvector);
 	CSQC_END
 	return true;
 }
@@ -231,16 +282,19 @@ extern sizebuf_t vm_tempstringsbuf;
 qboolean CL_VM_ConsoleCommand (const char *cmd)
 {
 	int restorevm_tempstringsbuf_cursize;
-	qboolean r;
+	qboolean r = false;
 	if(!cl.csqc_loaded)
 		return false;
 	CSQC_BEGIN
+	if (prog->funcoffsets.CSQC_ConsoleCommand)
+	{
 		prog->globals.client->time = cl.time;
 		restorevm_tempstringsbuf_cursize = vm_tempstringsbuf.cursize;
 		PRVM_G_INT(OFS_PARM0) = PRVM_SetTempString(cmd);
 		PRVM_ExecuteProgram(prog->funcoffsets.CSQC_ConsoleCommand, "QC function CSQC_ConsoleCommand is missing");
 		vm_tempstringsbuf.cursize = restorevm_tempstringsbuf_cursize;
 		r = CSQC_RETURNVAL;
+	}
 	CSQC_END
 	return r;
 }
@@ -372,6 +426,90 @@ void CL_VM_Parse_CenterPrint (const char *msg)
 	CSQC_END
 }
 
+void CL_VM_UpdateIntermissionState (int intermission)
+{
+	prvm_eval_t *val;
+	if(cl.csqc_loaded)
+	{
+		CSQC_BEGIN
+		val = PRVM_GLOBALFIELDVALUE(prog->globaloffsets.intermission);
+		if(val)
+			val->_float = intermission;
+		CSQC_END
+	}
+}
+void CL_VM_UpdateShowingScoresState (int showingscores)
+{
+	prvm_eval_t *val;
+	if(cl.csqc_loaded)
+	{
+		CSQC_BEGIN
+		val = PRVM_GLOBALFIELDVALUE(prog->globaloffsets.sb_showscores);
+		if(val)
+			val->_float = showingscores;
+		CSQC_END
+	}
+}
+qboolean CL_VM_Event_Sound(int sound_num, int volume, int channel, float attenuation, int ent, vec3_t pos)
+{
+	qboolean r = false;
+	if(cl.csqc_loaded)
+	{
+		CSQC_BEGIN
+		if(prog->funcoffsets.CSQC_Event_Sound)
+		{
+			prog->globals.client->time = cl.time;
+			PRVM_G_FLOAT(OFS_PARM0) = ent;
+			PRVM_G_FLOAT(OFS_PARM1) = channel;
+			PRVM_G_INT(OFS_PARM2) = PRVM_SetTempString(cl.sound_name[sound_num] );
+			PRVM_G_FLOAT(OFS_PARM3) = volume;
+			PRVM_G_FLOAT(OFS_PARM4) = attenuation;
+			VectorCopy(pos, PRVM_G_VECTOR(OFS_PARM5) );
+			PRVM_ExecuteProgram(prog->funcoffsets.CSQC_Event_Sound, "QC function CSQC_Event_Sound is missing");
+			r = CSQC_RETURNVAL;
+		}
+		CSQC_END
+	}
+
+	return r;
+}
+void CL_VM_UpdateCoopDeathmatchGlobals (int gametype)
+{
+	// Avoid global names for clean(er) coding
+	int localcoop;
+	int localdeathmatch;
+
+	prvm_eval_t *val;
+	if(cl.csqc_loaded)
+	{
+		if(gametype == GAME_COOP)
+		{
+			localcoop = 1;
+			localdeathmatch = 0;
+		}
+		else
+		if(gametype == GAME_DEATHMATCH)
+		{
+			localcoop = 0;
+			localdeathmatch = 1;
+		}
+		else
+		{
+			// How did the ServerInfo send an unknown gametype?
+			// Better just assign the globals as 0...
+			localcoop = 0;
+			localdeathmatch = 0;
+		}
+		CSQC_BEGIN
+		val = PRVM_GLOBALFIELDVALUE(prog->globaloffsets.coop);
+		if(val)
+			val->_float = localcoop;
+		val = PRVM_GLOBALFIELDVALUE(prog->globaloffsets.deathmatch);
+		if(val)
+			val->_float = localdeathmatch;
+		CSQC_END
+	}
+}
 float CL_VM_Event (float event)		//[515]: needed ? I'd say "YES", but don't know for what :D
 {
 	float r = 0;
@@ -392,13 +530,19 @@ float CL_VM_Event (float event)		//[515]: needed ? I'd say "YES", but don't know
 void CSQC_ReadEntities (void)
 {
 	unsigned short entnum, oldself, realentnum;
+	if(!cl.csqc_loaded)
+	{
+		Host_Error ("CSQC_ReadEntities: CSQC is not loaded");
+		return;
+	}
+
 	CSQC_BEGIN
 		prog->globals.client->time = cl.time;
 		oldself = prog->globals.client->self;
 		while(1)
 		{
 			entnum = MSG_ReadShort();
-			if(!entnum)
+			if(!entnum || msg_badread)
 				return;
 			realentnum = entnum & 0x7FFF;
 			prog->globals.client->self = cl.csqc_server2csqcentitynumber[realentnum];
@@ -410,21 +554,36 @@ void CSQC_ReadEntities (void)
 					cl.csqc_server2csqcentitynumber[realentnum] = 0;
 				}
 				else
-					Con_Printf("Smth bad happens in csqc...\n");	//[515]: never happens ?
+					Con_Printf("Bad csqc_server2csqcentitynumber map\n");	//[515]: never happens ?
 			}
 			else
 			{
 				if(!prog->globals.client->self)
 				{
-					prvm_edict_t	*ed;
-					ed = PRVM_ED_Alloc();
-					ed->fields.client->entnum = realentnum;
-					prog->globals.client->self = cl.csqc_server2csqcentitynumber[realentnum] = PRVM_EDICT_TO_PROG(ed);
+					if(!prog->funcoffsets.CSQC_Ent_Spawn)
+					{
+						prvm_edict_t	*ed;
+						ed = PRVM_ED_Alloc();
+						ed->fields.client->entnum = realentnum;
+						prog->globals.client->self = cl.csqc_server2csqcentitynumber[realentnum] = PRVM_EDICT_TO_PROG(ed);
+					}
+					else
+					{
+						// entity( float entnum ) CSQC_Ent_Spawn;
+						// the qc function should set entnum, too (this way it also can return world [2/1/2008 Andreas]
+						PRVM_G_FLOAT(OFS_PARM0) = (float) realentnum;
+						// make sure no one gets wrong ideas
+						prog->globals.client->self = 0;
+						PRVM_ExecuteProgram(prog->funcoffsets.CSQC_Ent_Spawn, "QC function CSQC_Ent_Spawn is missing");
+						prog->globals.client->self = cl.csqc_server2csqcentitynumber[realentnum] = PRVM_EDICT( PRVM_G_INT( OFS_RETURN ) );
+					}
 					PRVM_G_FLOAT(OFS_PARM0) = 1;
+					PRVM_ExecuteProgram(prog->funcoffsets.CSQC_Ent_Update, "QC function CSQC_Ent_Update is missing");			
 				}
-				else
+				else {
 					PRVM_G_FLOAT(OFS_PARM0) = 0;
-				PRVM_ExecuteProgram(prog->funcoffsets.CSQC_Ent_Update, "QC function CSQC_Ent_Update is missing");
+					PRVM_ExecuteProgram(prog->funcoffsets.CSQC_Ent_Update, "QC function CSQC_Ent_Update is missing");
+				}
 			}
 		}
 		prog->globals.client->self = oldself;
@@ -433,17 +592,8 @@ void CSQC_ReadEntities (void)
 
 void CL_VM_CB_BeginIncreaseEdicts(void)
 {
-	int i;
-	prvm_edict_t *ent;
-
 	// links don't survive the transition, so unlink everything
-	for (i = 0, ent = prog->edicts;i < prog->max_edicts;i++, ent++)
-	{
-		if (!ent->priv.server->free)
-			World_UnlinkEdict(prog->edicts + i);
-		memset(&ent->priv.server->areagrid, 0, sizeof(ent->priv.server->areagrid));
-	}
-	World_Clear(&cl.world);
+	World_UnlinkAll(&cl.world);
 }
 
 void CL_VM_CB_EndIncreaseEdicts(void)
@@ -501,10 +651,12 @@ void Cmd_ClearCsqcFuncs (void);
 
 void CL_VM_Init (void)
 {
+	const char* csprogsfn;
 	unsigned char *csprogsdata;
 	fs_offset_t csprogsdatasize;
 	int csprogsdatacrc, requiredcrc;
 	int requiredsize;
+	prvm_eval_t *val;
 
 	// reset csqc_progcrc after reading it, so that changing servers doesn't
 	// expect csqc on the next server
@@ -513,19 +665,19 @@ void CL_VM_Init (void)
 	Cvar_SetValueQuick(&csqc_progcrc, -1);
 	Cvar_SetValueQuick(&csqc_progsize, -1);
 
-	cl.csqc_loaded = false;
-	memset(cl.csqc_model_precache, 0, sizeof(cl.csqc_model_precache));
-	memset(&cl.csqc_vidvars, true, sizeof(csqc_vidvars_t));
-
 	// if the server is not requesting a csprogs, then we're done here
 	if (requiredcrc < 0)
 		return;
 
 	// see if the requested csprogs.dat file matches the requested crc
 	csprogsdatacrc = -1;
-	csprogsdata = FS_LoadFile(va("dlcache/%s.%i.%i", csqc_progname.string, requiredsize, requiredcrc), tempmempool, true, &csprogsdatasize);
+	csprogsfn = va("dlcache/%s.%i.%i", csqc_progname.string, requiredsize, requiredcrc);
+	csprogsdata = FS_LoadFile(csprogsfn, tempmempool, true, &csprogsdatasize);
 	if (!csprogsdata)
-		csprogsdata = FS_LoadFile(csqc_progname.string, tempmempool, true, &csprogsdatasize);
+	{
+		csprogsfn = csqc_progname.string;
+		csprogsdata = FS_LoadFile(csprogsfn, tempmempool, true, &csprogsdatasize);
+	}
 	if (csprogsdata)
 	{
 		csprogsdatacrc = CRC_Block(csprogsdata, csprogsdatasize);
@@ -571,6 +723,7 @@ void CL_VM_Init (void)
 	prog->limit_edicts = CL_MAX_EDICTS;
 	prog->reserved_edicts = 0;
 	prog->edictprivate_size = sizeof(edict_engineprivate_t);
+	// TODO: add a shared extension string #define and add real support for csqc extension strings [12/5/2007 Black]
 	prog->extensionstring = vm_sv_extensions;
 	prog->builtins = vm_cl_builtins;
 	prog->numbuiltins = vm_cl_numbuiltins;
@@ -584,40 +737,35 @@ void CL_VM_Init (void)
 	prog->reset_cmd = VM_CL_Cmd_Reset;
 	prog->error_cmd = CL_VM_Error;
 
-	PRVM_LoadProgs(csqc_progname.string, cl_numrequiredfunc, cl_required_func, 0, NULL, 0, NULL);
+	PRVM_LoadProgs(csprogsfn, cl_numrequiredfunc, cl_required_func, 0, NULL, 0, NULL);
 
 	if (!prog->loaded)
 	{
-		CL_VM_Error("CSQC ^2failed to load\n");
+		CL_VM_Error("CSQC %s ^2failed to load\n", csprogsfn);
 		if(!sv.active)
 			CL_Disconnect();
 		return;
 	}
 
-	Con_Printf("CSQC ^5loaded (crc=%i, size=%i)\n", csprogsdatacrc, (int)csprogsdatasize);
+	Con_Printf("CSQC %s ^5loaded (crc=%i, size=%i)\n", csprogsfn, csprogsdatacrc, (int)csprogsdatasize);
 
 	// check if OP_STATE animation is possible in this dat file
 	if (prog->fieldoffsets.nextthink >= 0 && prog->fieldoffsets.frame >= 0 && prog->fieldoffsets.think >= 0 && prog->globaloffsets.self >= 0)
 		prog->flag |= PRVM_OP_STATE;
 
-	if (cl.worldmodel)
-	{
-		VectorCopy(cl.worldmodel->normalmins, cl.world.areagrid_mins);
-		VectorCopy(cl.worldmodel->normalmaxs, cl.world.areagrid_maxs);
-	}
-	else
-	{
-		VectorSet(cl.world.areagrid_mins, -4096, -4096, -4096);
-		VectorSet(cl.world.areagrid_maxs, 4096, 4096, 4096);
-	}
-	World_Clear(&cl.world);
-
 	// set time
 	prog->globals.client->time = cl.time;
 
-	prog->globals.client->mapname = PRVM_SetEngineString(cl.worldmodel->name);
+	prog->globals.client->mapname = cl.worldmodel ? PRVM_SetEngineString(cl.worldmodel->name) : PRVM_SetEngineString("");
 	prog->globals.client->player_localentnum = cl.playerentity;
 
+	// set map description (use world entity 0)
+	val = PRVM_EDICTFIELDVALUE(prog->edicts, prog->fieldoffsets.message);
+	if(val)
+		val->string = PRVM_SetEngineString(cl.levelname);
+	VectorCopy(cl.world.mins, prog->edicts->fields.client->mins);
+	VectorCopy(cl.world.maxs, prog->edicts->fields.client->maxs);
+
 	// call the prog init
 	PRVM_ExecuteProgram(prog->funcoffsets.CSQC_Init, "QC function CSQC_Init is missing");
 
@@ -626,6 +774,9 @@ void CL_VM_Init (void)
 
 	cl.csqc_vidvars.drawcrosshair = false;
 	cl.csqc_vidvars.drawenginesbar = false;
+
+	// Update Coop and Deathmatch Globals (at this point the client knows them from ServerInfo)
+	CL_VM_UpdateCoopDeathmatchGlobals(cl.gametype);
 }
 
 void CL_VM_ShutDown (void)
@@ -637,7 +788,8 @@ void CL_VM_ShutDown (void)
 		return;
 	CSQC_BEGIN
 		prog->globals.client->time = cl.time;
-		PRVM_ExecuteProgram(prog->funcoffsets.CSQC_Shutdown, "QC function CSQC_Shutdown is missing");
+		if (prog->funcoffsets.CSQC_Shutdown)
+			PRVM_ExecuteProgram(prog->funcoffsets.CSQC_Shutdown, "QC function CSQC_Shutdown is missing");
 		PRVM_ResetProg();
 	CSQC_END
 	Con_Print("CSQC ^1unloaded\n");
