@@ -78,7 +78,7 @@ void stringlistfreecontents(stringlist_t *list)
 		Z_Free(list->strings);
 }
 
-void stringlistappend(stringlist_t *list, char *text)
+void stringlistappend(stringlist_t *list, const char *text)
 {
 	size_t textlen;
 	char **oldstrings;
@@ -108,7 +108,7 @@ void stringlistsort(stringlist_t *list)
 	{
 		for (j = i + 1;j < list->numstrings;j++)
 		{
-			if (strcmp(list->strings[i], list->strings[j]) > 0)
+			if (strcasecmp(list->strings[i], list->strings[j]) > 0)
 			{
 				temp = list->strings[i];
 				list->strings[i] = list->strings[j];
@@ -119,6 +119,13 @@ void stringlistsort(stringlist_t *list)
 }
 
 // operating system specific code
+static void adddirentry(stringlist_t *list, const char *name)
+{
+	if (strcmp(name, ".") && strcmp(name, ".."))
+	{
+		stringlistappend(list, name);
+	}
+}
 #ifdef WIN32
 #include <io.h>
 void listdirectory(stringlist_t *list, const char *path)
@@ -133,11 +140,9 @@ void listdirectory(stringlist_t *list, const char *path)
 	hFile = _findfirst(pattern, &n_file);
 	if(hFile == -1)
 		return;
-	// start a new chain with the the first name
-	stringlistappend(list, n_file.name);
-	// iterate through the directory
-	while (_findnext(hFile, &n_file) == 0)
-		stringlistappend(list, n_file.name);
+	do {
+		adddirentry(list, n_file.name );
+	} while (_findnext(hFile, &n_file) == 0);
 	_findclose(hFile);
 
 	// convert names to lowercase because windows does not care, but pattern matching code often does
@@ -156,8 +161,7 @@ void listdirectory(stringlist_t *list, const char *path)
 	if (!dir)
 		return;
 	while ((ent = readdir(dir)))
-		if (strcmp(ent->d_name, ".") && strcmp(ent->d_name, ".."))
-			stringlistappend(list, ent->d_name);
+		adddirentry(list, ent->d_name);
 	closedir(dir);
 }
 #endif
