@@ -22,9 +22,6 @@
 		Boston, MA  02111-1307, USA
 */
 
-// on UNIX platforms we need to define this so that video saving does not cause a SIGFSZ (file size) signal when a video clip exceeds 2GB
-#define _FILE_OFFSET_BITS 64
-
 #include "quakedef.h"
 
 #include <limits.h>
@@ -33,6 +30,7 @@
 #ifdef WIN32
 # include <direct.h>
 # include <io.h>
+# include <shlobj.h>
 #else
 # include <pwd.h>
 # include <sys/stat.h>
@@ -40,6 +38,7 @@
 #endif
 
 #include "fs.h"
+#include "wad.h"
 
 // Win32 requires us to add O_BINARY, but the other OSes don't have it
 #ifndef O_BINARY
@@ -51,6 +50,10 @@
 # define O_NONBLOCK 0
 #endif
 
+// largefile support for Win32
+#ifdef WIN32
+# define lseek _lseeki64
+#endif
 
 /*
 
@@ -273,11 +276,11 @@ char fs_gamedir[MAX_OSPATH];
 char fs_basedir[MAX_OSPATH];
 
 // list of active game directories (empty if not running a mod)
-#define MAX_GAMEDIRS 16
 int fs_numgamedirs = 0;
 char fs_gamedirs[MAX_GAMEDIRS][MAX_QPATH];
 
 cvar_t scr_screenshot_name = {0, "scr_screenshot_name","dp", "prefix name for saved screenshots (changes based on -game commandline, as well as which game mode is running)"};
+cvar_t fs_empty_files_in_pack_mark_deletions = {0, "fs_empty_files_in_pack_mark_deletions", "0", "if enabled, empty files in a pak/pk3 count as not existing but cancel the search in further packs, effectively allowing patch pak/pk3 files to 'delete' files"};
 
 
 /*
@@ -363,14 +366,7 @@ qboolean PK3_OpenLibrary (void)
 		return true;
 
 	// Load the DLL
-	if (! Sys_LoadLibrary (dllnames, &zlib_dll, zlibfuncs))
-	{
-		Con_Printf ("Compressed files support disabled\n");
-		return false;
-	}
-
-	Con_Printf ("Compressed files support enabled\n");
-	return true;
+	return Sys_LoadLibrary (dllnames, &zlib_dll, zlibfuncs);
 }
 
 
@@ -961,34 +957,38 @@ then loads and adds pak1.pak pak2.pak ...
 */
 void FS_AddGameDirectory (const char *dir)
 {
-	stringlist_t *list, *current;
+	int i;
+	stringlist_t list;
 	searchpath_t *search;
 	char pakfile[MAX_OSPATH];
 
 	strlcpy (fs_gamedir, dir, sizeof (fs_gamedir));
 
-	list = listdirectory(dir);
+	stringlistinit(&list);
+	listdirectory(&list, dir);
+	stringlistsort(&list);
 
 	// add any PAK package in the directory
-	for (current = list;current;current = current->next)
+	for (i = 0;i < list.numstrings;i++)
 	{
-		if (!strcasecmp(FS_FileExtension(current->text), "pak"))
+		if (!strcasecmp(FS_FileExtension(list.strings[i]), "pak"))
 		{
-			dpsnprintf (pakfile, sizeof (pakfile), "%s%s", dir, current->text);
+			dpsnprintf (pakfile, sizeof (pakfile), "%s%s", dir, list.strings[i]);
 			FS_AddPack_Fullpath(pakfile, NULL, false);
 		}
 	}
 
-	// add any PK3 package in the director
-	for (current = list;current;current = current->next)
+	// add any PK3 package in the directory
+	for (i = 0;i < list.numstrings;i++)
 	{
-		if (!strcasecmp(FS_FileExtension(current->text), "pk3"))
+		if (!strcasecmp(FS_FileExtension(list.strings[i]), "pk3"))
 		{
-			dpsnprintf (pakfile, sizeof (pakfile), "%s%s", dir, current->text);
+			dpsnprintf (pakfile, sizeof (pakfile), "%s%s", dir, list.strings[i]);
 			FS_AddPack_Fullpath(pakfile, NULL, false);
 		}
 	}
-	freedirectory(list);
+
+	stringlistfreecontents(&list);
 
 	// Add the directory to the search path
 	// (unpacked files have the priority over packed files)
@@ -1006,19 +1006,49 @@ FS_AddGameHierarchy
 */
 void FS_AddGameHierarchy (const char *dir)
 {
-#ifndef WIN32
+	int i;
+	char userdir[MAX_QPATH];
+#ifdef WIN32
+	TCHAR mydocsdir[MAX_PATH + 1];
+#else
 	const char *homedir;
 #endif
 
 	// Add the common game directory
 	FS_AddGameDirectory (va("%s%s/", fs_basedir, dir));
 
-#ifndef WIN32
+	*userdir = 0;
+
 	// Add the personal game directory
+#ifdef WIN32
+	if(SHGetFolderPath(NULL, CSIDL_PERSONAL, NULL, 0, mydocsdir) == S_OK)
+		dpsnprintf(userdir, sizeof(userdir), "%s/My Games/%s/", mydocsdir, gameuserdirname);
+#else
 	homedir = getenv ("HOME");
-	if (homedir != NULL && homedir[0] != '\0')
-		FS_AddGameDirectory (va("%s/.%s/%s/", homedir, gameuserdirname, dir));
+	if(homedir)
+		dpsnprintf(userdir, sizeof(userdir), "%s/.%s/", homedir, gameuserdirname);
 #endif
+
+#ifdef WIN32
+	if(!COM_CheckParm("-mygames"))
+	{
+		int fd = open (va("%s%s/config.cfg", fs_basedir, dir), O_WRONLY | O_CREAT, 0666); // note: no O_TRUNC here!
+		if(fd >= 0)
+		{
+			close(fd);
+			*userdir = 0; // we have write access to the game dir, so let's use it
+		}
+	}
+#endif
+
+	if(COM_CheckParm("-nohome"))
+		*userdir = 0;
+
+	if((i = COM_CheckParm("-userdir")) && i < com_argc - 1)
+		dpsnprintf(userdir, sizeof(userdir), "%s/", com_argv[i+1]);
+
+	if (*userdir)
+		FS_AddGameDirectory(va("%s%s/", userdir, dir));
 }
 
 
@@ -1047,6 +1077,26 @@ const char *FS_FileExtension (const char *in)
 }
 
 
+/*
+============
+FS_FileWithoutPath
+============
+*/
+const char *FS_FileWithoutPath (const char *in)
+{
+	const char *separator, *backslash, *colon;
+
+	separator = strrchr(in, '/');
+	backslash = strrchr(in, '\\');
+	if (!separator || separator < backslash)
+		separator = backslash;
+	colon = strrchr(in, ':');
+	if (!separator || separator < colon)
+		separator = colon;
+	return separator ? separator + 1 : in;
+}
+
+
 /*
 ================
 FS_ClearSearchPath
@@ -1054,12 +1104,18 @@ FS_ClearSearchPath
 */
 void FS_ClearSearchPath (void)
 {
+	// unload all packs and directory information, close all pack files
+	// (if a qfile is still reading a pack it won't be harmed because it used
+	//  dup() to get its own handle already)
 	while (fs_searchpaths)
 	{
 		searchpath_t *search = fs_searchpaths;
 		fs_searchpaths = search->next;
 		if (search->pack)
 		{
+			// close the file
+			close(search->pack->handle);
+			// free any memory associated with it
 			if (search->pack->files)
 				Mem_Free(search->pack->files);
 			Mem_Free(search->pack);
@@ -1071,10 +1127,10 @@ void FS_ClearSearchPath (void)
 
 /*
 ================
-FS_Rescan_f
+FS_Rescan
 ================
 */
-void FS_Rescan_f (void)
+void FS_Rescan (void)
 {
 	int i;
 	qboolean fs_modified = false;
@@ -1132,44 +1188,78 @@ void FS_Rescan_f (void)
 		if (gamemode == GAME_NORMAL || gamemode == GAME_HIPNOTIC || gamemode == GAME_ROGUE)
 			Con_Print("Playing registered version.\n");
 	}
+
+	// unload all wads so that future queries will return the new data
+	W_UnloadAll();
 }
 
+void FS_Rescan_f(void)
+{
+	FS_Rescan();
+}
 
 /*
 ================
-FS_ChangeGameDir
+FS_ChangeGameDirs
 ================
 */
-void Host_SaveConfig_f (void);
-void Host_LoadConfig_f (void);
-qboolean FS_ChangeGameDir(const char *string)
+extern void Host_SaveConfig (void);
+extern void Host_LoadConfig_f (void);
+qboolean FS_ChangeGameDirs(int numgamedirs, char gamedirs[][MAX_QPATH], qboolean complain, qboolean failmissing)
 {
-	// if already using the requested gamedir, do nothing
-	if (fs_numgamedirs == 1 && !strcmp(fs_gamedirs[0], string))
-		return false;
+	int i;
 
-	// if string is nasty, reject it
-	if(FS_CheckNastyPath(string, true)) // overflowed or nasty?
+	if (fs_numgamedirs == numgamedirs)
 	{
-		Con_Printf("FS_ChangeGameDir(\"%s\"): nasty filename rejected\n", string);
-		return false;
+		for (i = 0;i < numgamedirs;i++)
+			if (strcasecmp(fs_gamedirs[i], gamedirs[i]))
+				break;
+		if (i == numgamedirs)
+			return true; // already using this set of gamedirs, do nothing
 	}
 
-	// save the current config
-	Host_SaveConfig_f();
+	if (numgamedirs > MAX_GAMEDIRS)
+	{
+		if (complain)
+			Con_Printf("That is too many gamedirs (%i > %i)\n", numgamedirs, MAX_GAMEDIRS);
+		return false; // too many gamedirs
+	}
 
-	// change to the new gamedir
-	fs_numgamedirs = 1;
-	strlcpy(fs_gamedirs[0], string, sizeof(fs_gamedirs[0]));
+	for (i = 0;i < numgamedirs;i++)
+	{
+		// if string is nasty, reject it
+		if(FS_CheckNastyPath(gamedirs[i], true))
+		{
+			if (complain)
+				Con_Printf("Nasty gamedir name rejected: %s\n", gamedirs[i]);
+			return false; // nasty gamedirs
+		}
+	}
+
+	for (i = 0;i < numgamedirs;i++)
+	{
+		if (!FS_CheckGameDir(gamedirs[i]) && failmissing)
+		{
+			if (complain)
+				Con_Printf("Gamedir missing: %s%s/\n", fs_basedir, gamedirs[i]);
+			return false; // missing gamedirs
+		}
+	}
+
+	Host_SaveConfig();
+
+	fs_numgamedirs = numgamedirs;
+	for (i = 0;i < fs_numgamedirs;i++)
+		strlcpy(fs_gamedirs[i], gamedirs[i], sizeof(fs_gamedirs[i]));
 
 	// reinitialize filesystem to detect the new paks
-	FS_Rescan_f();
+	FS_Rescan();
 
 	// exec the new config
 	Host_LoadConfig_f();
 
-	// reinitialize the loaded sounds
-	S_Reload_f();
+	// unload all sounds so they will be reloaded from the new files as needed
+	S_UnloadAllSounds_f();
 
 	// reinitialize renderer (this reloads hud/console background/etc)
 	R_Modules_Restart();
@@ -1185,6 +1275,8 @@ FS_GameDir_f
 void FS_GameDir_f (void)
 {
 	int i;
+	int numgamedirs;
+	char gamedirs[MAX_GAMEDIRS][MAX_QPATH];
 
 	if (Cmd_Argc() < 2)
 	{
@@ -1195,39 +1287,44 @@ void FS_GameDir_f (void)
 		return;
 	}
 
-	if (cls.state != ca_disconnected || sv.active)
+	numgamedirs = Cmd_Argc() - 1;
+	if (numgamedirs > MAX_GAMEDIRS)
 	{
+		Con_Printf("Too many gamedirs (%i > %i)\n", numgamedirs, MAX_GAMEDIRS);
+		return;
+	}
+
+	for (i = 0;i < numgamedirs;i++)
+		strlcpy(gamedirs[i], Cmd_Argv(i+1), sizeof(gamedirs[i]));
+
+	if ((cls.state == ca_connected && !cls.demoplayback) || sv.active)
+	{
+		// actually, changing during game would work fine, but would be stupid
 		Con_Printf("Can not change gamedir while client is connected or server is running!\n");
 		return;
 	}
 
-	Host_SaveConfig_f();
+	// halt demo playback to close the file
+	CL_Disconnect();
 
-	fs_numgamedirs = 0;
-	for (i = 1;i < Cmd_Argc() && fs_numgamedirs < MAX_GAMEDIRS;i++)
-	{
-		// if string is nasty, reject it
-		if(FS_CheckNastyPath(Cmd_Argv(i), true)) // overflowed or nasty?
-		{
-			Con_Printf("FS_GameDir_f(\"%s\"): nasty filename rejected\n", Cmd_Argv(i));
-			continue;
-		}
+	FS_ChangeGameDirs(numgamedirs, gamedirs, true, true);
+}
 
-		strlcpy(fs_gamedirs[fs_numgamedirs], Cmd_Argv(i), sizeof(fs_gamedirs[fs_numgamedirs]));
-		fs_numgamedirs++;
-	}
 
-	// reinitialize filesystem to detect the new paks
-	FS_Rescan_f();
-
-	// exec the new config
-	Host_LoadConfig_f();
-
-	// reinitialize the loaded sounds
-	S_Reload_f();
-
-	// reinitialize renderer (this reloads hud/console background/etc)
-	R_Modules_Restart();
+/*
+================
+FS_CheckGameDir
+================
+*/
+qboolean FS_CheckGameDir(const char *gamedir)
+{
+	qboolean success;
+	stringlist_t list;
+	stringlistinit(&list);
+	listdirectory(&list, va("%s%s/", fs_basedir, gamedir));
+	success = list.numstrings > 0;
+	stringlistfreecontents(&list);
+	return success;
 }
 
 
@@ -1283,6 +1380,12 @@ void FS_Init (void)
 	if (fs_basedir[0] && fs_basedir[strlen(fs_basedir) - 1] != '/' && fs_basedir[strlen(fs_basedir) - 1] != '\\')
 		strlcat(fs_basedir, "/", sizeof(fs_basedir));
 
+	if (!FS_CheckGameDir(gamedirname1))
+		Con_Printf("WARNING: base gamedir %s%s/ not found!\n", fs_basedir, gamedirname1);
+
+	if (gamedirname2 && !FS_CheckGameDir(gamedirname2))
+		Con_Printf("WARNING: base gamedir %s%s/ not found!\n", fs_basedir, gamedirname2);
+
 	// -game <gamedir>
 	// Adds basedir/gamedir as an override game
 	// LordHavoc: now supports multiple -game directories
@@ -1293,19 +1396,24 @@ void FS_Init (void)
 		if (!strcmp (com_argv[i], "-game") && i < com_argc-1)
 		{
 			i++;
+			if (FS_CheckNastyPath(com_argv[i], true))
+				Sys_Error("-game %s%s/ is a dangerous/non-portable path\n", fs_basedir, com_argv[i]);
+			if (!FS_CheckGameDir(com_argv[i]))
+				Con_Printf("WARNING: -game %s%s/ not found!\n", fs_basedir, com_argv[i]);
 			// add the gamedir to the list of active gamedirs
 			strlcpy (fs_gamedirs[fs_numgamedirs], com_argv[i], sizeof(fs_gamedirs[fs_numgamedirs]));
 			fs_numgamedirs++;
 		}
 	}
 
-	// update the searchpath
-	FS_Rescan_f();
+	// generate the searchpath
+	FS_Rescan();
 }
 
 void FS_Init_Commands(void)
 {
 	Cvar_RegisterVariable (&scr_screenshot_name);
+	Cvar_RegisterVariable (&fs_empty_files_in_pack_mark_deletions);
 
 	Cmd_AddCommand ("gamedir", FS_GameDir_f, "changes active gamedir list (can take multiple arguments), not including base directory (example usage: gamedir ctf)");
 	Cmd_AddCommand ("fs_rescan", FS_Rescan_f, "rescans filesystem for new pack archives and any other changes");
@@ -1321,6 +1429,10 @@ FS_Shutdown
 */
 void FS_Shutdown (void)
 {
+	// close all pack files and such
+	// (hopefully there aren't any other open files, but they'll be cleaned up
+	//  by the OS anyway)
+	FS_ClearSearchPath();
 	Mem_FreePool (&fs_mempool);
 }
 
@@ -1599,6 +1711,17 @@ static searchpath_t *FS_FindFile (const char *name, int* index, qboolean quiet)
 				// Found it
 				if (!diff)
 				{
+					if (fs_empty_files_in_pack_mark_deletions.integer && pak->files[middle].realsize == 0)
+					{
+						// yes, but the first one is empty so we treat it as not being there
+						if (!quiet && developer.integer >= 10)
+							Con_Printf("FS_FindFile: %s is marked as deleted\n", name);
+
+						if (index != NULL)
+							*index = -1;
+						return NULL;
+					}
+
 					if (!quiet && developer.integer >= 10)
 						Con_Printf("FS_FindFile: %s in %s\n",
 									pak->files[middle].name, pak->filename);
@@ -1688,6 +1811,17 @@ Open a file. The syntax is the same as fopen
 */
 qfile_t* FS_Open (const char* filepath, const char* mode, qboolean quiet, qboolean nonblocking)
 {
+#ifdef FS_FIX_PATHS
+	char fixedFileName[MAX_QPATH];
+	char *d;
+	strlcpy( fixedFileName, filepath, MAX_QPATH );
+	// try to fix common mistakes (\ instead of /)
+	for( d = fixedFileName ; *d ; d++ )
+		if( *d == '\\' )
+			*d = '/';
+	filepath = fixedFileName;
+#endif
+
 	if (FS_CheckNastyPath(filepath, false))
 	{
 		Con_Printf("FS_Open(\"%s\", \"%s\", %s): nasty filename rejected\n", filepath, mode, quiet ? "true" : "false");
@@ -1796,12 +1930,12 @@ fs_offset_t FS_Read (qfile_t* file, void* buffer, size_t buffersize)
 	if (file->buff_ind < file->buff_len)
 	{
 		count = file->buff_len - file->buff_ind;
+		count = ((fs_offset_t)buffersize > count) ? count : (fs_offset_t)buffersize;
+		done += count;
+		memcpy (buffer, &file->buff[file->buff_ind], count);
+		file->buff_ind += count;
 
-		done += ((fs_offset_t)buffersize > count) ? count : (fs_offset_t)buffersize;
-		memcpy (buffer, &file->buff[file->buff_ind], done);
-		file->buff_ind += done;
-
-		buffersize -= done;
+		buffersize -= count;
 		if (buffersize == 0)
 			return done;
 	}
@@ -2013,7 +2147,7 @@ Get the next character of a file
 */
 int FS_Getc (qfile_t* file)
 {
-	char c;
+	unsigned char c;
 
 	if (FS_Read (file, &c, 1) != 1)
 		return EOF;
@@ -2070,7 +2204,7 @@ int FS_Seek (qfile_t* file, fs_offset_t offset, int whence)
 		default:
 			return -1;
 	}
-	if (offset < 0 || offset > (long) file->real_length)
+	if (offset < 0 || offset > file->real_length)
 		return -1;
 
 	// If we have the data in our read buffer, we don't need to actually seek
@@ -2198,6 +2332,8 @@ unsigned char *FS_LoadFile (const char *path, mempool_t *pool, qboolean quiet, f
 		buf[filesize] = '\0';
 		FS_Read (file, buf, filesize);
 		FS_Close (file);
+		if (developer_loadfile.integer)
+			Con_Printf("loaded file \"%s\" (%u bytes)\n", path, (unsigned int)filesize);
 	}
 
 	if (filesizepointer)
@@ -2224,7 +2360,7 @@ qboolean FS_WriteFile (const char *filename, void *data, fs_offset_t len)
 		return false;
 	}
 
-	Con_DPrintf("FS_WriteFile: %s\n", filename);
+	Con_DPrintf("FS_WriteFile: %s (%u bytes)\n", filename, (unsigned int)len);
 	FS_Write (file, data, len);
 	FS_Close (file);
 	return true;
@@ -2293,6 +2429,30 @@ void FS_DefaultExtension (char *path, const char *extension, size_t size_path)
 }
 
 
+/*
+==================
+FS_FileType
+
+Look for a file in the packages and in the filesystem
+==================
+*/
+int FS_FileType (const char *filename)
+{
+	searchpath_t *search;
+	char fullpath[MAX_QPATH];
+
+	search = FS_FindFile (filename, NULL, true);
+	if(!search)
+		return FS_FILETYPE_NONE;
+
+	if(search->pack)
+		return FS_FILETYPE_FILE; // TODO can't check directories in paks yet, maybe later
+
+	dpsnprintf(fullpath, sizeof(fullpath), "%s%s", search->filename, filename);
+	return FS_SysFileType(fullpath);
+}
+
+
 /*
 ==================
 FS_FileExists
@@ -2313,28 +2473,36 @@ FS_SysFileExists
 Look for a file in the filesystem only
 ==================
 */
-qboolean FS_SysFileExists (const char *path)
+int FS_SysFileType (const char *path)
 {
 #if WIN32
-	int desc;
+	DWORD result = GetFileAttributes(path);
 
-	// TODO: use another function instead, to avoid opening the file
-	desc = open (path, O_RDONLY | O_BINARY);
-	if (desc < 0)
-		return false;
+	if(result == INVALID_FILE_ATTRIBUTES)
+		return FS_FILETYPE_NONE;
 
-	close (desc);
-	return true;
+	if(result & FILE_ATTRIBUTE_DIRECTORY)
+		return FS_FILETYPE_DIRECTORY;
+
+	return FS_FILETYPE_FILE;
 #else
 	struct stat buf;
 
 	if (stat (path,&buf) == -1)
-		return false;
+		return FS_FILETYPE_NONE;
 
-	return true;
+	if(S_ISDIR(buf.st_mode))
+		return FS_FILETYPE_DIRECTORY;
+
+	return FS_FILETYPE_FILE;
 #endif
 }
 
+qboolean FS_SysFileExists (const char *path)
+{
+	return FS_SysFileType (path) != FS_FILETYPE_NONE;
+}
+
 void FS_mkdir (const char *path)
 {
 #if WIN32
@@ -2356,8 +2524,9 @@ fssearch_t *FS_Search(const char *pattern, int caseinsensitive, int quiet)
 	fssearch_t *search;
 	searchpath_t *searchpath;
 	pack_t *pak;
-	int i, basepathlength, numfiles, numchars;
-	stringlist_t *dir, *dirfile, *liststart, *listcurrent, *listtemp;
+	int i, basepathlength, numfiles, numchars, resultlistindex, dirlistindex;
+	stringlist_t resultlist;
+	stringlist_t dirlist;
 	const char *slash, *backslash, *colon, *separator;
 	char *basepath;
 	char netpath[MAX_OSPATH];
@@ -2372,10 +2541,9 @@ fssearch_t *FS_Search(const char *pattern, int caseinsensitive, int quiet)
 		return NULL;
 	}
 
+	stringlistinit(&resultlist);
+	stringlistinit(&dirlist);
 	search = NULL;
-	liststart = NULL;
-	listcurrent = NULL;
-	listtemp = NULL;
 	slash = strrchr(pattern, '/');
 	backslash = strrchr(pattern, '\\');
 	colon = strrchr(pattern, ':');
@@ -2402,16 +2570,14 @@ fssearch_t *FS_Search(const char *pattern, int caseinsensitive, int quiet)
 				{
 					if (matchpattern(temp, (char *)pattern, true))
 					{
-						for (listtemp = liststart;listtemp;listtemp = listtemp->next)
-							if (!strcmp(listtemp->text, temp))
+						for (resultlistindex = 0;resultlistindex < resultlist.numstrings;resultlistindex++)
+							if (!strcmp(resultlist.strings[resultlistindex], temp))
 								break;
-						if (listtemp == NULL)
+						if (resultlistindex == resultlist.numstrings)
 						{
-							listcurrent = stringlistappend(listcurrent, temp);
-							if (liststart == NULL)
-								liststart = listcurrent;
-							if (!quiet)
-								Con_DPrintf("SearchPackFile: %s : %s\n", pak->filename, temp);
+							stringlistappend(&resultlist, temp);
+							if (!quiet && developer_loading.integer)
+								Con_Printf("SearchPackFile: %s : %s\n", pak->filename, temp);
 						}
 					}
 					// strip off one path element at a time until empty
@@ -2434,59 +2600,52 @@ fssearch_t *FS_Search(const char *pattern, int caseinsensitive, int quiet)
 		{
 			// get a directory listing and look at each name
 			dpsnprintf(netpath, sizeof (netpath), "%s%s", searchpath->filename, basepath);
-			if ((dir = listdirectory(netpath)))
-			{
-				for (dirfile = dir;dirfile;dirfile = dirfile->next)
+			stringlistinit(&dirlist);
+			listdirectory(&dirlist, netpath);
+			for (dirlistindex = 0;dirlistindex < dirlist.numstrings;dirlistindex++)
+  			{
+				dpsnprintf(temp, sizeof(temp), "%s%s", basepath, dirlist.strings[dirlistindex]);
+				if (matchpattern(temp, (char *)pattern, true))
 				{
-					dpsnprintf(temp, sizeof(temp), "%s%s", basepath, dirfile->text);
-					if (matchpattern(temp, (char *)pattern, true))
+					for (resultlistindex = 0;resultlistindex < resultlist.numstrings;resultlistindex++)
+						if (!strcmp(resultlist.strings[resultlistindex], temp))
+							break;
+					if (resultlistindex == resultlist.numstrings)
 					{
-						for (listtemp = liststart;listtemp;listtemp = listtemp->next)
-							if (!strcmp(listtemp->text, temp))
-								break;
-						if (listtemp == NULL)
-						{
-							listcurrent = stringlistappend(listcurrent, temp);
-							if (liststart == NULL)
-								liststart = listcurrent;
-							if (!quiet)
-								Con_DPrintf("SearchDirFile: %s\n", temp);
-						}
+						stringlistappend(&resultlist, temp);
+						if (!quiet && developer_loading.integer)
+							Con_Printf("SearchDirFile: %s\n", temp);
 					}
 				}
-				freedirectory(dir);
 			}
+			stringlistfreecontents(&dirlist);
 		}
 	}
 
-	if (liststart)
+	if (resultlist.numstrings)
 	{
-		liststart = stringlistsort(liststart);
-		numfiles = 0;
+		stringlistsort(&resultlist);
+		numfiles = resultlist.numstrings;
 		numchars = 0;
-		for (listtemp = liststart;listtemp;listtemp = listtemp->next)
-		{
-			numfiles++;
-			numchars += (int)strlen(listtemp->text) + 1;
-		}
+		for (resultlistindex = 0;resultlistindex < resultlist.numstrings;resultlistindex++)
+			numchars += (int)strlen(resultlist.strings[resultlistindex]) + 1;
 		search = (fssearch_t *)Z_Malloc(sizeof(fssearch_t) + numchars + numfiles * sizeof(char *));
 		search->filenames = (char **)((char *)search + sizeof(fssearch_t));
 		search->filenamesbuffer = (char *)((char *)search + sizeof(fssearch_t) + numfiles * sizeof(char *));
 		search->numfilenames = (int)numfiles;
 		numfiles = 0;
 		numchars = 0;
-		for (listtemp = liststart;listtemp;listtemp = listtemp->next)
+		for (resultlistindex = 0;resultlistindex < resultlist.numstrings;resultlistindex++)
 		{
 			size_t textlen;
 			search->filenames[numfiles] = search->filenamesbuffer + numchars;
-			textlen = strlen(listtemp->text) + 1;
-			memcpy(search->filenames[numfiles], listtemp->text, textlen);
+			textlen = strlen(resultlist.strings[resultlistindex]) + 1;
+			memcpy(search->filenames[numfiles], resultlist.strings[resultlistindex], textlen);
 			numfiles++;
 			numchars += (int)textlen;
 		}
-		if (liststart)
-			stringlistfree(liststart);
 	}
+	stringlistfreecontents(&resultlist);
 
 	Mem_Free(basepath);
 	return search;
@@ -2601,3 +2760,78 @@ const char *FS_WhichPack(const char *filename)
 	else
 		return 0;
 }
+
+/*
+====================
+FS_IsRegisteredQuakePack
+
+Look for a proof of purchase file file in the requested package
+
+If it is found, this file should NOT be downloaded.
+====================
+*/
+qboolean FS_IsRegisteredQuakePack(const char *name)
+{
+	searchpath_t *search;
+	pack_t *pak;
+
+	// search through the path, one element at a time
+	for (search = fs_searchpaths;search;search = search->next)
+	{
+		if (search->pack && !strcasecmp(FS_FileWithoutPath(search->filename), name))
+		{
+			int (*strcmp_funct) (const char* str1, const char* str2);
+			int left, right, middle;
+
+			pak = search->pack;
+			strcmp_funct = pak->ignorecase ? strcasecmp : strcmp;
+
+			// Look for the file (binary search)
+			left = 0;
+			right = pak->numfiles - 1;
+			while (left <= right)
+			{
+				int diff;
+
+				middle = (left + right) / 2;
+				diff = !strcmp_funct (pak->files[middle].name, "gfx/pop.lmp");
+
+				// Found it
+				if (!diff)
+					return true;
+
+				// If we're too far in the list
+				if (diff > 0)
+					right = middle - 1;
+				else
+					left = middle + 1;
+			}
+
+			// we found the requested pack but it is not registered quake
+			return false;
+		}
+	}
+
+	return false;
+}
+
+int FS_CRCFile(const char *filename, size_t *filesizepointer)
+{
+	int crc = -1;
+	unsigned char *filedata;
+	fs_offset_t filesize;
+	if (filesizepointer)
+		*filesizepointer = 0;
+	if (!filename || !*filename)
+		return crc;
+	filedata = FS_LoadFile(filename, tempmempool, true, &filesize);
+	if (filedata)
+	{
+		if (filesizepointer)
+			*filesizepointer = filesize;
+		crc = CRC_Block(filedata, filesize);
+		Mem_Free(filedata);
+	}
+	return crc;
+}
+
