@@ -19,6 +19,8 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
 #include "quakedef.h"
+#include "sv_demo.h"
+#include "image.h"
 
 int current_skill;
 cvar_t sv_cheats = {0, "sv_cheats", "0", "enables cheat commands in any game, and cheat impulses in dpmod"};
@@ -28,6 +30,7 @@ cvar_t rcon_address = {0, "rcon_address", "", "server address to send rcon comma
 cvar_t team = {CVAR_USERINFO | CVAR_SAVE, "team", "none", "QW team (4 character limit, example: blue)"};
 cvar_t skin = {CVAR_USERINFO | CVAR_SAVE, "skin", "", "QW player skin name (example: base)"};
 cvar_t noaim = {CVAR_USERINFO | CVAR_SAVE, "noaim", "1", "QW option to disable vertical autoaim"};
+cvar_t r_fixtrans_auto = {0, "r_fixtrans_auto", "0", "automatically fixtrans textures (when set to 2, it also saves the fixed versions to a fixtrans directory)"};
 qboolean allowcheats = false;
 
 extern qboolean host_shuttingdown;
@@ -46,7 +49,6 @@ void Host_Quit_f (void)
 		Sys_Quit (0);
 }
 
-
 /*
 ==================
 Host_Status_f
@@ -81,6 +83,7 @@ void Host_Status_f (void)
 	print ("version:  %s build %s\n", gamename, buildstring);
 	print ("protocol: %i (%s)\n", Protocol_NumberForEnum(sv.protocol), Protocol_NameForEnum(sv.protocol));
 	print ("map:      %s\n", sv.name);
+	print ("timing:   %s\n", Host_TimingReport());
 	print ("players:  %i active (%i max)\n\n", players, svs.maxclients);
 	for (j = 0, client = svs.clients;j < svs.maxclients;j++, client++)
 	{
@@ -97,7 +100,7 @@ void Host_Status_f (void)
 		}
 		else
 			hours = 0;
-		print ("#%-3u %-16.16s^%i  %3i  %2i:%02i:%02i\n", j+1, client->name, STRING_COLOR_DEFAULT, client->frags, hours, minutes, seconds);
+		print ("#%-3u %-16.16s  %3i  %2i:%02i:%02i\n", j+1, client->name, client->frags, hours, minutes, seconds);
 		print ("   %s\n", client->netconnection ? client->netconnection->address : "botclient");
 	}
 }
@@ -261,6 +264,10 @@ void Host_Map_f (void)
 		return;
 	}
 
+	// GAME_DELUXEQUAKE - clear warpmark (used by QC)
+	if (gamemode == GAME_DELUXEQUAKE)
+		Cvar_Set("warpmark", "");
+
 	cls.demonum = -1;		// stop demo loop in case this fails
 
 	CL_Disconnect ();
@@ -428,33 +435,6 @@ LOAD / SAVE GAME
 
 #define	SAVEGAME_VERSION	5
 
-/*
-===============
-Host_SavegameComment
-
-Writes a SAVEGAME_COMMENT_LENGTH character comment describing the current
-===============
-*/
-void Host_SavegameComment (char *text)
-{
-	int		i;
-	char	kills[20];
-
-	for (i=0 ; i<SAVEGAME_COMMENT_LENGTH ; i++)
-		text[i] = ' ';
-	// LordHavoc: added min() to prevent overflow
-	memcpy (text, cl.levelname, min(strlen(cl.levelname), SAVEGAME_COMMENT_LENGTH));
-	sprintf (kills,"kills:%3i/%3i", cl.stats[STAT_MONSTERS], cl.stats[STAT_TOTALMONSTERS]);
-	memcpy (text+22, kills, strlen(kills));
-	// convert space to _ to make stdio happy
-	// LordHavoc: convert control characters to _ as well
-	for (i=0 ; i<SAVEGAME_COMMENT_LENGTH ; i++)
-		if (text[i] <= ' ')
-			text[i] = '_';
-	text[SAVEGAME_COMMENT_LENGTH] = '\0';
-}
-
-
 /*
 ===============
 Host_Savegame_f
@@ -467,34 +447,29 @@ void Host_Savegame_f (void)
 	int		i;
 	char	comment[SAVEGAME_COMMENT_LENGTH+1];
 
-	if (cls.state != ca_connected || !sv.active)
+	if (!sv.active)
 	{
-		Con_Print("Not playing a local game.\n");
+		Con_Print("Can't save - no server running.\n");
 		return;
 	}
 
-	if (cl.intermission)
+	if (cl.islocalgame)
 	{
-		Con_Print("Can't save in intermission.\n");
-		return;
-	}
-
-	for (i = 0;i < svs.maxclients;i++)
-	{
-		if (svs.clients[i].active)
+		// singleplayer checks
+		if (cl.intermission)
 		{
-			if (i > 0)
-			{
-				Con_Print("Can't save multiplayer games.\n");
-				return;
-			}
-			if (svs.clients[i].edict->fields.server->deadflag)
-			{
-				Con_Print("Can't savegame with a dead player\n");
-				return;
-			}
+			Con_Print("Can't save in intermission.\n");
+			return;
+		}
+
+		if (svs.clients[0].active && svs.clients[0].edict->fields.server->deadflag)
+		{
+			Con_Print("Can't savegame with a dead player\n");
+			return;
 		}
 	}
+	else
+		Con_Print("Warning: saving a multiplayer game may have strange results when restored (to properly resume, all players must join in the same player slots and then the game can be reloaded).\n");
 
 	if (Cmd_Argc() != 2)
 	{
@@ -519,8 +494,19 @@ void Host_Savegame_f (void)
 		return;
 	}
 
+	SV_VM_Begin();
+
 	FS_Printf(f, "%i\n", SAVEGAME_VERSION);
-	Host_SavegameComment (comment);
+
+	memset(comment, 0, sizeof(comment));
+	sprintf(comment, "%-21.21s kills:%3i/%3i", PRVM_GetString(prog->edicts->fields.server->message), (int)prog->globals.server->killed_monsters, (int)prog->globals.server->total_monsters);
+	// convert space to _ to make stdio happy
+	// LordHavoc: convert control characters to _ as well
+	for (i=0 ; i<SAVEGAME_COMMENT_LENGTH ; i++)
+		if (comment[i] <= ' ')
+			comment[i] = '_';
+	comment[SAVEGAME_COMMENT_LENGTH] = '\0';
+
 	FS_Printf(f, "%s\n", comment);
 	for (i=0 ; i<NUM_SPAWN_PARMS ; i++)
 		FS_Printf(f, "%f\n", svs.clients[0].spawn_parms[i]);
@@ -537,8 +523,6 @@ void Host_Savegame_f (void)
 			FS_Print(f,"m\n");
 	}
 
-	SV_VM_Begin();
-
 	PRVM_ED_WriteGlobals (f);
 	for (i=0 ; i<prog->num_edicts ; i++)
 		PRVM_ED_Write (f, PRVM_EDICT_NUM(i));
@@ -581,6 +565,13 @@ void Host_Loadgame_f (void)
 
 	Con_Printf("Loading game from %s...\n", filename);
 
+	// stop playing demos
+	if (cls.demoplayback)
+		CL_Disconnect ();
+
+	// remove menu
+	key_dest = key_game;
+
 	cls.demonum = -1;		// stop demo loop in case this fails
 
 	t = text = (char *)FS_LoadFile (filename, tempmempool, false, NULL);
@@ -591,7 +582,7 @@ void Host_Loadgame_f (void)
 	}
 
 	// version
-	COM_ParseToken_Simple(&t, false);
+	COM_ParseToken_Simple(&t, false, false);
 	version = atoi(com_token);
 	if (version != SAVEGAME_VERSION)
 	{
@@ -601,25 +592,25 @@ void Host_Loadgame_f (void)
 	}
 
 	// description
-	COM_ParseToken_Simple(&t, false);
+	COM_ParseToken_Simple(&t, false, false);
 
 	for (i = 0;i < NUM_SPAWN_PARMS;i++)
 	{
-		COM_ParseToken_Simple(&t, false);
+		COM_ParseToken_Simple(&t, false, false);
 		spawn_parms[i] = atof(com_token);
 	}
 	// skill
-	COM_ParseToken_Simple(&t, false);
+	COM_ParseToken_Simple(&t, false, false);
 // this silliness is so we can load 1.06 save files, which have float skill values
 	current_skill = (int)(atof(com_token) + 0.5);
 	Cvar_SetValue ("skill", (float)current_skill);
 
 	// mapname
-	COM_ParseToken_Simple(&t, false);
+	COM_ParseToken_Simple(&t, false, false);
 	strlcpy (mapname, com_token, sizeof(mapname));
 
 	// time
-	COM_ParseToken_Simple(&t, false);
+	COM_ParseToken_Simple(&t, false, false);
 	time = atof(com_token);
 
 	allowcheats = sv_cheats.integer != 0;
@@ -640,7 +631,7 @@ void Host_Loadgame_f (void)
 	{
 		// light style
 		oldt = t;
-		COM_ParseToken_Simple(&t, false);
+		COM_ParseToken_Simple(&t, false, false);
 		// if this is a 64 lightstyle savegame produced by Quake, stop now
 		// we have to check this because darkplaces saves 256 lightstyle savegames
 		if (com_token[0] == '{')
@@ -658,7 +649,7 @@ void Host_Loadgame_f (void)
 	for(;;)
 	{
 		oldt = t;
-		COM_ParseToken_Simple(&t, false);
+		COM_ParseToken_Simple(&t, false, false);
 		if (com_token[0] == '{')
 		{
 			t = oldt;
@@ -673,10 +664,10 @@ void Host_Loadgame_f (void)
 	for (;;)
 	{
 		start = t;
-		while (COM_ParseToken_Simple(&t, false))
+		while (COM_ParseToken_Simple(&t, false, false))
 			if (!strcmp(com_token, "}"))
 				break;
-		if (!COM_ParseToken_Simple(&start, false))
+		if (!COM_ParseToken_Simple(&start, false, false))
 		{
 			// end of file
 			break;
@@ -725,7 +716,7 @@ void Host_Loadgame_f (void)
 	SV_VM_End();
 
 	// make sure we're connected to loopback
-	if (cls.state == ca_disconnected || !(cls.state == ca_connected && cls.netcon != NULL && LHNETADDRESS_GetAddressType(&cls.netcon->peeraddress) == LHNETADDRESSTYPE_LOOP))
+	if (sv.active && cls.state == ca_disconnected)
 		CL_EstablishConnection("local:1");
 }
 
@@ -754,11 +745,6 @@ void Host_Name_f (void)
 	else
 		strlcpy (newName, Cmd_Args(), sizeof (newName));
 
-	for (i = 0, j = 0;newName[i];i++)
-		if (newName[i] != '\r' && newName[i] != '\n')
-			newName[j++] = newName[i];
-	newName[j] = 0;
-
 	if (cmd_source == src_command)
 	{
 		Cvar_Set ("_cl_name", newName);
@@ -776,6 +762,11 @@ void Host_Name_f (void)
 	// point the string back at updateclient->name to keep it safe
 	strlcpy (host_client->name, newName, sizeof (host_client->name));
 
+	for (i = 0, j = 0;host_client->name[i];i++)
+		if (host_client->name[i] != '\r' && host_client->name[i] != '\n')
+			host_client->name[j++] = host_client->name[i];
+	host_client->name[j] = 0;
+
 	COM_StringLengthNoColors(host_client->name, 0, &valid_colors);
 	if(!valid_colors) // NOTE: this also proves the string is not empty, as "" is a valid colored string
 	{
@@ -796,16 +787,42 @@ void Host_Name_f (void)
 		}
 	}
 
+	// find the last color tag offset and decide if we need to add a reset tag
+	for (i = 0, j = -1;host_client->name[i];i++)
+	{
+		if (host_client->name[i] == STRING_COLOR_TAG)
+		{
+			if (host_client->name[i+1] >= '0' && host_client->name[i+1] <= '9')
+			{
+				j = i;
+				// if this happens to be a reset  tag then we don't need one
+				if (host_client->name[i+1] == '0' + STRING_COLOR_DEFAULT)
+					j = -1;
+				i++;
+				continue;
+			}
+			if (host_client->name[i+1] == STRING_COLOR_TAG)
+			{
+				i++;
+				continue;
+			}
+		}
+	}
+	// does not end in the default color string, so add it
+	if (j >= 0 && strlen(host_client->name) < sizeof(host_client->name) - 2)
+		memcpy(host_client->name + strlen(host_client->name), STRING_COLOR_DEFAULT_STR, strlen(STRING_COLOR_DEFAULT_STR) + 1);
+
 	host_client->edict->fields.server->netname = PRVM_SetEngineString(host_client->name);
 	if (strcmp(host_client->old_name, host_client->name))
 	{
 		if (host_client->spawned)
-			SV_BroadcastPrintf("%s^%i changed name to %s\n", host_client->old_name, STRING_COLOR_DEFAULT, host_client->name);
+			SV_BroadcastPrintf("%s changed name to %s\n", host_client->old_name, host_client->name);
 		strlcpy(host_client->old_name, host_client->name, sizeof(host_client->old_name));
 		// send notification to all clients
 		MSG_WriteByte (&sv.reliable_datagram, svc_updatename);
 		MSG_WriteByte (&sv.reliable_datagram, host_client - svs.clients);
 		MSG_WriteString (&sv.reliable_datagram, host_client->name);
+		SV_WriteNetnameIntoDemo(host_client);
 	}
 }
 
@@ -970,13 +987,13 @@ void Host_Say(qboolean teamonly)
 	}
 	// note this uses the chat prefix \001
 	if (!fromServer && !teamonly)
-		dpsnprintf (text, sizeof(text), "\001%s" STRING_COLOR_DEFAULT_STR ": %s", host_client->name, p1);
+		dpsnprintf (text, sizeof(text), "\001%s: %s", host_client->name, p1);
 	else if (!fromServer && teamonly)
-		dpsnprintf (text, sizeof(text), "\001(%s" STRING_COLOR_DEFAULT_STR "): %s", host_client->name, p1);
+		dpsnprintf (text, sizeof(text), "\001(%s): %s", host_client->name, p1);
 	else if(*(sv_adminnick.string))
-		dpsnprintf (text, sizeof(text), "\001<%s" STRING_COLOR_DEFAULT_STR "> %s", sv_adminnick.string, p1);
+		dpsnprintf (text, sizeof(text), "\001<%s> %s", sv_adminnick.string, p1);
 	else
-		dpsnprintf (text, sizeof(text), "\001<%s" STRING_COLOR_DEFAULT_STR "> %s", hostname.string, p1);
+		dpsnprintf (text, sizeof(text), "\001<%s> %s", hostname.string, p1);
 	p2 = text + strlen(text);
 	while ((const char *)p2 > (const char *)text && (p2[-1] == '\r' || p2[-1] == '\n' || (p2[-1] == '\"' && quoted)))
 	{
@@ -1248,7 +1265,7 @@ void Host_BottomColor_f(void)
 	Host_Color(-1, atoi(Cmd_Argv(1)));
 }
 
-cvar_t cl_rate = {CVAR_SAVE | CVAR_NQUSERINFOHACK, "_cl_rate", "10000", "internal storage cvar for current rate (changed by rate command)"};
+cvar_t cl_rate = {CVAR_SAVE | CVAR_NQUSERINFOHACK, "_cl_rate", "20000", "internal storage cvar for current rate (changed by rate command)"};
 void Host_Rate_f(void)
 {
 	int rate;
@@ -1364,6 +1381,7 @@ void Host_PreSpawn_f (void)
 		SZ_Write (&host_client->netconnection->message, sv.signon.data, sv.signon.cursize);
 		MSG_WriteByte (&host_client->netconnection->message, svc_signonnum);
 		MSG_WriteByte (&host_client->netconnection->message, 2);
+		host_client->sendsignon = 0;		// enable unlimited sends again
 	}
 
 	// reset the name change timer because the client will send name soon
@@ -1400,9 +1418,6 @@ void Host_Spawn_f (void)
 	if (sv.loadgame)
 	{
 		// loaded games are fully initialized already
-		// if this is the last client to be connected, unpause
-		sv.paused = false;
-
 		if (prog->funcoffsets.RestoreGame)
 		{
 			Con_DPrint("Calling RestoreGame\n");
@@ -1425,8 +1440,8 @@ void Host_Spawn_f (void)
 		prog->globals.server->self = PRVM_EDICT_TO_PROG(host_client->edict);
 		PRVM_ExecuteProgram (prog->globals.server->ClientConnect, "QC function ClientConnect is missing");
 
-		if (svs.maxclients > 1 || cls.state == ca_dedicated)
-			Con_Printf("%s^%i entered the game\n", host_client->name, STRING_COLOR_DEFAULT);
+		if (cls.state == ca_dedicated)
+			Con_Printf("%s connected\n", host_client->name);
 
 		PRVM_ExecuteProgram (prog->globals.server->PutClientInServer, "QC function PutClientInServer is missing");
 	}
@@ -1493,7 +1508,6 @@ void Host_Spawn_f (void)
 		MSG_WriteAngle (&host_client->netconnection->message, host_client->edict->fields.server->v_angle[0], sv.protocol);
 		MSG_WriteAngle (&host_client->netconnection->message, host_client->edict->fields.server->v_angle[1], sv.protocol);
 		MSG_WriteAngle (&host_client->netconnection->message, 0, sv.protocol);
-		sv.loadgame = false; // we're basically done with loading now
 	}
 	else
 	{
@@ -1517,6 +1531,20 @@ Host_Begin_f
 void Host_Begin_f (void)
 {
 	host_client->spawned = true;
+
+	// LordHavoc: note: this code also exists in SV_DropClient
+	if (sv.loadgame)
+	{
+		int i;
+		for (i = 0;i < svs.maxclients;i++)
+			if (svs.clients[i].active && !svs.clients[i].spawned)
+				break;
+		if (i == svs.maxclients)
+		{
+			Con_Printf("Loaded game, everyone rejoined - unpausing\n");
+			sv.paused = sv.loadgame = false; // we're basically done with loading now
+		}
+	}
 }
 
 //===========================================================================
@@ -1580,7 +1608,7 @@ void Host_Kick_f (void)
 		if (Cmd_Argc() > 2)
 		{
 			message = Cmd_Args();
-			COM_ParseToken_Simple(&message, false);
+			COM_ParseToken_Simple(&message, false, false);
 			if (byNumber)
 			{
 				message++;							// skip the #
@@ -1925,7 +1953,7 @@ void Host_Startdemos_f (void)
 		Con_Printf("Max %i demos in demoloop\n", MAX_DEMOS);
 		c = MAX_DEMOS;
 	}
-	Con_Printf("%i demo(s) in loop\n", c);
+	Con_DPrintf("%i demo(s) in loop\n", c);
 
 	for (i=1 ; i<c+1 ; i++)
 		strlcpy (cls.demos[i-1], Cmd_Argv(i), sizeof (cls.demos[i-1]));
@@ -2321,6 +2349,8 @@ void Host_Packet_f (void) // credit: taken from QuakeWorld
 	}
 
 	mysocket = NetConn_ChooseClientSocketForAddress(&address);
+	if (!mysocket)
+		mysocket = NetConn_ChooseServerSocketForAddress(&address);
 	if (mysocket)
 		NetConn_Write(mysocket, send, out - send, &address);
 }
@@ -2481,6 +2511,9 @@ void Host_InitCommands (void)
 	Cmd_AddCommand_WithClientCommand ("pings", NULL, Host_Pings_f, "command sent by clients to request updated ping and packetloss of players on scoreboard (originally from QW, but also used on NQ servers)");
 	Cmd_AddCommand ("pingplreport", Host_PingPLReport_f, "command sent by server containing client ping and packet loss values for scoreboard, triggered by pings command from client (not used by QW servers)");
 
+	Cmd_AddCommand ("fixtrans", Image_FixTransparentPixels_f, "change alpha-zero pixels in an image file to sensible values, and write out a new TGA (warning: SLOW)");
+	Cvar_RegisterVariable (&r_fixtrans_auto);
+
 	Cvar_RegisterVariable (&team);
 	Cvar_RegisterVariable (&skin);
 	Cvar_RegisterVariable (&noaim);
@@ -2489,3 +2522,6 @@ void Host_InitCommands (void)
 	Cvar_RegisterVariable(&sv_adminnick);
 }
 
+void Host_NoOperation_f(void)
+{
+}
