@@ -53,6 +53,9 @@ static const keyname_t   keynames[] = {
 	{"ENTER", K_ENTER},
 	{"ESCAPE", K_ESCAPE},
 	{"SPACE", K_SPACE},
+
+	// spacer so it lines up with keys.h
+
 	{"BACKSPACE", K_BACKSPACE},
 	{"UPARROW", K_UPARROW},
 	{"DOWNARROW", K_DOWNARROW},
@@ -85,12 +88,47 @@ static const keyname_t   keynames[] = {
 
 	{"PAUSE", K_PAUSE},
 
-	{"MWHEELUP", K_MWHEELUP},
-	{"MWHEELDOWN", K_MWHEELDOWN},
+	{"NUMLOCK", K_NUMLOCK},
+	{"CAPSLOCK", K_CAPSLOCK},
+	{"SCROLLOCK", K_SCROLLOCK},
+
+	{"KP_INS",			K_KP_INS },
+	{"KP_0", K_KP_0},
+	{"KP_END",			K_KP_END },
+	{"KP_1", K_KP_1},
+	{"KP_DOWNARROW",	K_KP_DOWNARROW },
+	{"KP_2", K_KP_2},
+	{"KP_PGDN",			K_KP_PGDN },
+	{"KP_3", K_KP_3},
+	{"KP_LEFTARROW",	K_KP_LEFTARROW },
+	{"KP_4", K_KP_4},
+	{"KP_5", K_KP_5},
+	{"KP_RIGHTARROW",	K_KP_RIGHTARROW },
+	{"KP_6", K_KP_6},
+	{"KP_HOME",			K_KP_HOME },
+	{"KP_7", K_KP_7},
+	{"KP_UPARROW",		K_KP_UPARROW },
+	{"KP_8", K_KP_8},
+	{"KP_PGUP",			K_KP_PGUP },
+	{"KP_9", K_KP_9},
+	{"KP_DEL",			K_KP_DEL },
+	{"KP_PERIOD", K_KP_PERIOD},
+	{"KP_SLASH",		K_KP_SLASH },
+	{"KP_DIVIDE", K_KP_DIVIDE},
+	{"KP_MULTIPLY", K_KP_MULTIPLY},
+	{"KP_MINUS", K_KP_MINUS},
+	{"KP_PLUS", K_KP_PLUS},
+	{"KP_ENTER", K_KP_ENTER},
+	{"KP_EQUALS", K_KP_EQUALS},
+
+
 
 	{"MOUSE1", K_MOUSE1},
+
 	{"MOUSE2", K_MOUSE2},
 	{"MOUSE3", K_MOUSE3},
+	{"MWHEELUP", K_MWHEELUP},
+	{"MWHEELDOWN", K_MWHEELDOWN},
 	{"MOUSE4", K_MOUSE4},
 	{"MOUSE5", K_MOUSE5},
 	{"MOUSE6", K_MOUSE6},
@@ -105,39 +143,8 @@ static const keyname_t   keynames[] = {
 	{"MOUSE15", K_MOUSE15},
 	{"MOUSE16", K_MOUSE16},
 
-	{"NUMLOCK", K_NUMLOCK},
-	{"CAPSLOCK", K_CAPSLOCK},
-	{"SCROLLOCK", K_SCROLLOCK},
 
-	{"KP_HOME",			K_KP_HOME },
-	{"KP_UPARROW",		K_KP_UPARROW },
-	{"KP_PGUP",			K_KP_PGUP },
-	{"KP_LEFTARROW",	K_KP_LEFTARROW },
-	{"KP_RIGHTARROW",	K_KP_RIGHTARROW },
-	{"KP_END",			K_KP_END },
-	{"KP_DOWNARROW",	K_KP_DOWNARROW },
-	{"KP_PGDN",			K_KP_PGDN },
-	{"KP_INS",			K_KP_INS },
-	{"KP_DEL",			K_KP_DEL },
-	{"KP_SLASH",		K_KP_SLASH },
 
-	{"KP_0", K_KP_0},
-	{"KP_1", K_KP_1},
-	{"KP_2", K_KP_2},
-	{"KP_3", K_KP_3},
-	{"KP_4", K_KP_4},
-	{"KP_5", K_KP_5},
-	{"KP_6", K_KP_6},
-	{"KP_7", K_KP_7},
-	{"KP_8", K_KP_8},
-	{"KP_9", K_KP_9},
-	{"KP_PERIOD", K_KP_PERIOD},
-	{"KP_DIVIDE", K_KP_DIVIDE},
-	{"KP_MULTIPLY", K_KP_MULTIPLY},
-	{"KP_MINUS", K_KP_MINUS},
-	{"KP_PLUS", K_KP_PLUS},
-	{"KP_ENTER", K_KP_ENTER},
-	{"KP_EQUALS", K_KP_EQUALS},
 
 	{"JOY1",  K_JOY1},
 	{"JOY2",  K_JOY2},
@@ -156,6 +163,11 @@ static const keyname_t   keynames[] = {
 	{"JOY15", K_JOY15},
 	{"JOY16", K_JOY16},
 
+
+
+
+
+
 	{"AUX1", K_AUX1},
 	{"AUX2", K_AUX2},
 	{"AUX3", K_AUX3},
@@ -502,6 +514,8 @@ qboolean	chat_team;
 char		chat_buffer[MAX_INPUTLINE];
 unsigned int	chat_bufferlen = 0;
 
+extern int Nicks_CompleteChatLine(char *buffer, size_t size, unsigned int pos);
+
 static void
 Key_Message (int key, char ascii)
 {
@@ -531,6 +545,11 @@ Key_Message (int key, char ascii)
 		return;
 	}
 
+	if(key == K_TAB) {
+		chat_bufferlen = Nicks_CompleteChatLine(chat_buffer, sizeof(chat_buffer), chat_bufferlen);
+		return;
+	}
+
 	if (chat_bufferlen == sizeof (chat_buffer) - 1)
 		return;							// all full
 
@@ -617,6 +636,8 @@ Key_SetBinding (int keynum, int bindmap, const char *binding)
 		Z_Free (keybindings[bindmap][keynum]);
 		keybindings[bindmap][keynum] = NULL;
 	}
+	if(!binding[0]) // make "" binds be removed --blub
+		return;
 // allocate memory for new binding
 	l = strlen (binding);
 	newbinding = (char *)Z_Malloc (l + 1);
@@ -797,17 +818,21 @@ void
 Key_WriteBindings (qfile_t *f)
 {
 	int         i, j;
+	char bindbuf[MAX_INPUTLINE];
+	const char *p;
 
 	for (j = 0; j < MAX_BINDMAPS; j++)
 	{
 		for (i = 0; i < (int)(sizeof(keybindings[0])/sizeof(keybindings[0][0])); i++)
 		{
-			if (keybindings[j][i])
+			p = keybindings[j][i];
+			if (p)
 			{
+				Cmd_QuoteString(bindbuf, sizeof(bindbuf), p, "\"\\");
 				if (j == 0)
-					FS_Printf(f, "bind %s \"%s\"\n", Key_KeynumToString (i), keybindings[j][i]);
+					FS_Printf(f, "bind %s \"%s\"\n", Key_KeynumToString (i), bindbuf);
 				else
-					FS_Printf(f, "in_bind %d %s \"%s\"\n", j, Key_KeynumToString (i), keybindings[j][i]);
+					FS_Printf(f, "in_bind %d %s \"%s\"\n", j, Key_KeynumToString (i), bindbuf);
 			}
 		}
 	}
@@ -850,7 +875,7 @@ const char *Key_GetBind (int key)
 	return bind;
 }
 
-qboolean CL_VM_InputEvent (qboolean pressed, int key);
+qboolean CL_VM_InputEvent (qboolean down, int key, int ascii);
 
 /*
 ===================
@@ -872,9 +897,13 @@ Key_Event (int key, char ascii, qboolean down)
 	if (!bind)
 		bind = keybindings[key_bmap2][key];
 
+	if (developer.integer >= 1000)
+		Con_Printf("Key_Event(%i, '%c', %s) keydown %i bind \"%s\"\n", key, ascii, down ? "down" : "up", keydown[key], bind ? bind : "");
+
+#if 0
 	if(key_dest == key_game)
 	{
-		q = CL_VM_InputEvent(!down, key);
+		q = CL_VM_InputEvent(down, key);
 		if(q)
 		{
 			if (down)
@@ -884,6 +913,7 @@ Key_Event (int key, char ascii, qboolean down)
 			return;
 		}
 	}
+#endif
 
 	if (down)
 	{
@@ -895,13 +925,6 @@ Key_Event (int key, char ascii, qboolean down)
 	{
 		// clear repeat count now that the key is released
 		keydown[key] = 0;
-		// key up events only generate commands if the game key binding is a button
-		// command (leading + sign).  These will occur even in console mode, to
-		// keep the character from continuing an action started before a console
-		// switch.  Button commands include the kenum as a parameter, so multiple
-		// downs can be matched with ups
-		if (bind && bind[0] == '+')
-			Cbuf_AddText(va("-%s %i\n", bind + 1, key));
 	}
 
 	// key_consoleactive is a flag not a key_dest because the console is a
@@ -938,10 +961,13 @@ Key_Event (int key, char ascii, qboolean down)
 					Key_Message (key, ascii);
 				break;
 			case key_menu:
+			case key_menu_grabbed:
 				MR_KeyEvent (key, ascii, down);
 				break;
 			case key_game:
-				if (down)
+				// csqc has priority over toggle menu if it wants to (e.g. handling escape for UI stuff in-game.. :sick:)
+				q = CL_VM_InputEvent(down, key, ascii);
+				if (!q && down)
 					MR_ToggleMenu_f ();
 				break;
 			default:
@@ -950,7 +976,8 @@ Key_Event (int key, char ascii, qboolean down)
 		return;
 	}
 
-	// send function keydowns to interpreter no matter what mode is
+	// send function keydowns to interpreter no matter what mode is (unless the menu has specifically grabbed the keyboard, for rebinding keys)
+	if (key_dest != key_menu_grabbed)
 	if (key >= K_F1 && key <= K_F12 && down)
 	{
 		// ignore key repeats on F1-F12 binds
@@ -983,6 +1010,17 @@ Key_Event (int key, char ascii, qboolean down)
 		return;
 	}
 
+
+	// FIXME: actually the up-bind should only be called if the button was actually pressed while key_dest == key_game [12/17/2007 Black]
+	//			 especially CL_VM_InputEvent should be able to prevent it from being called (to intercept the binds)
+	// key up events only generate commands if the game key binding is a button
+	// command (leading + sign).  These will occur even in console mode, to
+	// keep the character from continuing an action started before a console
+	// switch.  Button commands include the kenum as a parameter, so multiple
+	// downs can be matched with ups
+	if (!down && bind && bind[0] == '+')
+		Cbuf_AddText(va("-%s %i\n", bind + 1, key));
+
 	// ignore binds while a video is played, let the video system handle the key event
 	if (cl_videoplaying)
 	{
@@ -998,11 +1036,13 @@ Key_Event (int key, char ascii, qboolean down)
 				Key_Message (key, ascii);
 			break;
 		case key_menu:
+		case key_menu_grabbed:
 			MR_KeyEvent (key, ascii, down);
 			break;
 		case key_game:
-			// ignore key repeats on binds
-			if (bind && keydown[key] == 1 && down)
+			q = CL_VM_InputEvent(down, key, ascii);
+			// ignore key repeats on binds and only send the bind if the event hasnt been already processed by csqc
+			if (!q && bind && keydown[key] == 1 && down)
 			{
 				// button commands add keynum as a parm
 				if (bind[0] == '+')
