@@ -48,7 +48,8 @@ enum m_state_e {
 	m_options_effects,
 	m_options_graphics,
 	m_options_colorcontrol,
-	m_reset
+	m_reset,
+	m_modlist
 };
 
 extern enum m_state_e m_state;
@@ -75,6 +76,7 @@ void MP_Shutdown (void);*/
 //
 // menu router
 //
+
 void MR_Init_Commands (void);
 void MR_Init (void);
 void MR_Restart (void);
@@ -82,5 +84,15 @@ extern void (*MR_KeyEvent) (int key, char ascii, qboolean downevent);
 extern void (*MR_Draw) (void);
 extern void (*MR_ToggleMenu_f) (void);
 extern void (*MR_Shutdown) (void);
+
+typedef struct video_resolution_s
+{
+	const char *type;
+	int width, height;
+	int conwidth, conheight;
+	double pixelheight; // pixel aspect
+}
+video_resolution_t;
+extern video_resolution_t video_resolutions[];
 #endif
 
