@@ -75,22 +75,14 @@ skinframe_t;
 
 #define MAX_SKINS 256
 
-typedef struct overridetagname_s
-{
-	char name[MAX_QPATH];
-}
-overridetagname_t;
-
-// a replacement set of tag names, per skin
-typedef struct overridetagnameset_s
-{
-	int num_overridetagnames;
-	overridetagname_t *data_overridetagnames;
-}
-overridetagnameset_t;
-
 struct md3vertex_s;
-struct trivertex_s;
+struct trivertx_s;
+typedef struct texvecvertex_s
+{
+	signed char svec[3];
+	signed char tvec[3];
+}
+texvecvertex_t;
 
 // used for mesh lists in q1bsp/q3bsp map models
 // (the surfaces reference portions of these meshes)
@@ -125,6 +117,7 @@ typedef struct surfmesh_s
 	int num_morphframes;
 	struct md3vertex_s *data_morphmd3vertex;
 	struct trivertx_s *data_morphmdlvertex;
+	struct texvecvertex_s *data_morphtexvecvertex;
 	float *data_morphmd2framesize6f;
 	float num_morphmdlframescale[3];
 	float num_morphmdlframetranslate[3];
@@ -186,17 +179,21 @@ shadowmesh_t;
 // TODO: support these features more directly
 #define Q3TEXTUREFLAG_TWOSIDED 1
 #define Q3TEXTUREFLAG_NOPICMIP 16
+#define Q3TEXTUREFLAG_POLYGONOFFSET 32
+#define Q3TEXTUREFLAG_REFRACTION 256
+#define Q3TEXTUREFLAG_REFLECTION 512
+#define Q3TEXTUREFLAG_WATERSHADER 1024
 
 #define Q3PATHLENGTH 64
 #define TEXTURE_MAXFRAMES 64
 #define Q3WAVEPARMS 4
 #define Q3DEFORM_MAXPARMS 3
-#define Q3SHADER_MAXLAYERS 8
+#define Q3SHADER_MAXLAYERS 2 // FIXME support more than that (currently only two are used, so why keep more in RAM?)
 #define Q3RGBGEN_MAXPARMS 3
 #define Q3ALPHAGEN_MAXPARMS 1
 #define Q3TCGEN_MAXPARMS 6
 #define Q3TCMOD_MAXPARMS 6
-#define Q3MAXTCMODS 4
+#define Q3MAXTCMODS 8
 #define Q3MAXDEFORMS 4
 
 typedef enum q3wavefunc_e
@@ -330,7 +327,8 @@ typedef struct q3shaderinfo_layer_s
 	int clampmap;
 	float framerate;
 	int numframes;
-	char texturename[TEXTURE_MAXFRAMES][Q3PATHLENGTH];
+	int texflags;
+	char** texturename;
 	int blendfunc[2];
 	q3shaderinfo_layer_rgbgen_t rgbgen;
 	q3shaderinfo_layer_alphagen_t alphagen;
@@ -357,10 +355,19 @@ typedef struct q3shaderinfo_s
 	qboolean lighting;
 	qboolean vertexalpha;
 	qboolean textureblendalpha;
-	q3shaderinfo_layer_t *primarylayer, *backgroundlayer;
+	int primarylayer, backgroundlayer;
 	q3shaderinfo_layer_t layers[Q3SHADER_MAXLAYERS];
 	char skyboxname[Q3PATHLENGTH];
 	q3shaderinfo_deform_t deforms[Q3MAXDEFORMS];
+
+	// reflection
+	float reflectmin; // when refraction is used, minimum amount of reflection (when looking straight down)
+	float reflectmax; // when refraction is used, maximum amount of reflection (when looking parallel to water)
+	float refractfactor; // amount of refraction distort (1.0 = like the cvar specifies)
+	vec4_t refractcolor4f; // color tint of refraction (including alpha factor)
+	float reflectfactor; // amount of reflection distort (1.0 = like the cvar specifies)
+	vec4_t reflectcolor4f; // color tint of reflection (including alpha factor)
+	float r_water_wateralpha; // additional wateralpha to apply when r_water is active
 }
 q3shaderinfo_t;
 
@@ -411,9 +418,10 @@ typedef struct texture_s
 	// current material flags (updated each bmodel render)
 	int currentmaterialflags;
 
-	// current PolygonOffset values for rendering this material
-	float currentpolygonfactor;
-	float currentpolygonoffset;
+	// PolygonOffset values for rendering this material
+	// (these are added to the r_refdef values and submodel values)
+	float biaspolygonfactor;
+	float biaspolygonoffset;
 
 	// textures to use when rendering this material
 	skinframe_t *currentskinframe;
@@ -458,6 +466,10 @@ typedef struct texture_s
 	float specularpower;
 	// color tint (colormod * currentalpha) used for rtlighting this material
 	float dlightcolor[3];
+	// color tint (colormod * 2) used for lightmapped lighting on this material
+	// includes alpha as 4th component
+	// replaces role of gl_Color in GLSL shader
+	float lightmapcolor[4];
 
 	// from q3 shaders
 	int customblendfunc[2];
@@ -471,6 +483,15 @@ typedef struct texture_s
 	int supercontents;
 	int surfaceparms;
 	int textureflags;
+
+	// reflection
+	float reflectmin; // when refraction is used, minimum amount of reflection (when looking straight down)
+	float reflectmax; // when refraction is used, maximum amount of reflection (when looking parallel to water)
+	float refractfactor; // amount of refraction distort (1.0 = like the cvar specifies)
+	vec4_t refractcolor4f; // color tint of refraction (including alpha factor)
+	float reflectfactor; // amount of reflection distort (1.0 = like the cvar specifies)
+	vec4_t reflectcolor4f; // color tint of reflection (including alpha factor)
+	float r_water_wateralpha; // additional wateralpha to apply when r_water is active
 }
 texture_t;
 
@@ -486,7 +507,7 @@ typedef struct msurface_lightmapinfo_s
 {
 	// texture mapping properties used by this surface
 	mtexinfo_t *texinfo; // q1bsp
-	// index into r_refdef.lightstylevalue array, 255 means not used (black)
+	// index into r_refdef.scene.lightstylevalue array, 255 means not used (black)
 	unsigned char styles[MAXLIGHTMAPS]; // q1bsp
 	// RGB lighting data [numstyles][height][width][3]
 	unsigned char *samples; // q1bsp
@@ -552,12 +573,19 @@ model_sprite_t;
 
 struct trace_s;
 
+typedef struct model_brush_lightstyleinfo_s
+{
+	int style;
+	int value;
+	int numsurfaces;
+	int *surfacelist;
+}
+model_brush_lightstyleinfo_t;
+
 typedef struct model_brush_s
 {
 	// true if this model is a HalfLife .bsp file
 	qboolean ishlbsp;
-	// true if this model is a Martial Concert .bsp file
-	qboolean ismcbsp;
 	// string of entity definitions (.map format)
 	char *entities;
 
@@ -616,7 +644,7 @@ typedef struct model_brush_s
 	int (*SuperContentsFromNativeContents)(struct model_s *model, int nativecontents);
 	int (*NativeContentsFromSuperContents)(struct model_s *model, int supercontents);
 	unsigned char *(*GetPVS)(struct model_s *model, const vec3_t p);
-	int (*FatPVS)(struct model_s *model, const vec3_t org, vec_t radius, unsigned char *pvsbuffer, int pvsbufferlength);
+	int (*FatPVS)(struct model_s *model, const vec3_t org, vec_t radius, unsigned char *pvsbuffer, int pvsbufferlength, qboolean merge);
 	int (*BoxTouchingPVS)(struct model_s *model, const unsigned char *pvs, const vec3_t mins, const vec3_t maxs);
 	int (*BoxTouchingLeafPVS)(struct model_s *model, const unsigned char *pvs, const vec3_t mins, const vec3_t maxs);
 	int (*BoxTouchingVisibleLeafs)(struct model_s *model, const unsigned char *visibleleafs, const vec3_t mins, const vec3_t maxs);
@@ -645,9 +673,6 @@ model_brush_t;
 
 typedef struct model_brushq1_s
 {
-	// lightmap format, set to r_lightmaprgba when model is loaded
-	int				lightmaprgba;
-
 	dmodel_t		*submodels;
 
 	int				numvertexes;
@@ -663,7 +688,7 @@ typedef struct model_brushq1_s
 	int				*surfedges;
 
 	int				numclipnodes;
-	dclipnode_t		*clipnodes;
+	mclipnode_t		*clipnodes;
 
 	hull_t			hulls[MAX_MAP_HULLS];
 
@@ -675,11 +700,8 @@ typedef struct model_brushq1_s
 	unsigned char			*nmaplightdata; // deluxemap file
 
 	// lightmap update chains for light styles
-	int				light_styles;
-	unsigned char			*light_style;
-	int				*light_stylevalue;
-	msurface_t		***light_styleupdatechains;
-	msurface_t		**light_styleupdatechainsbuffer;
+	int				num_lightstyles;
+	model_brush_lightstyleinfo_t *data_lightstyleinfo;
 }
 model_brushq1_t;
 
@@ -744,6 +766,8 @@ typedef struct model_brushq3_s
 }
 model_brushq3_t;
 
+struct frameblend_s;
+
 typedef struct model_s
 {
 	// name and path of model, for example "progs/player.mdl"
@@ -813,12 +837,20 @@ typedef struct model_s
 	msurface_lightmapinfo_t *data_surfaces_lightmapinfo;
 	// all surfaces belong to this mesh
 	surfmesh_t		surfmesh;
+	// data type of model
+	const char		*modeldatatypestring;
+	// generates vertex data for a given frameblend
+	void(*AnimateVertices)(const struct model_s *model, const struct frameblend_s *frameblend, float *vertex3f, float *normal3f, float *svector3f, float *tvector3f);
 	// draw the model's sky polygons (only used by brush models)
 	void(*DrawSky)(struct entity_render_s *ent);
+	// draw refraction/reflection textures for the model's water polygons (only used by brush models)
+	void(*DrawAddWaterPlanes)(struct entity_render_s *ent);
 	// draw the model using lightmap/dlight shading
 	void(*Draw)(struct entity_render_s *ent);
 	// draw the model to the depth buffer (no color rendering at all)
 	void(*DrawDepth)(struct entity_render_s *ent);
+	// draw any enabled debugging effects on this model (such as showing triangles, normals, collision brushes...)
+	void(*DrawDebug)(struct entity_render_s *ent);
 	// gathers info on which clusters and surfaces are lit by light, as well as calculating a bounding box
 	void(*GetLightInfo)(struct entity_render_s *ent, vec3_t relativelightorigin, float lightradius, vec3_t outmins, vec3_t outmaxs, int *outleaflist, unsigned char *outleafpvs, int *outnumleafspointer, int *outsurfacelist, unsigned char *outsurfacepvs, int *outnumsurfacespointer, unsigned char *outshadowtrispvs, unsigned char *outlighttrispvs);
 	// compile a shadow volume for the model based on light source
@@ -829,6 +861,8 @@ typedef struct model_s
 	void(*DrawLight)(struct entity_render_s *ent, int numsurfaces, const int *surfacelist, const unsigned char *trispvs);
 	// trace a box against this model
 	void (*TraceBox)(struct model_s *model, int frame, struct trace_s *trace, const vec3_t start, const vec3_t boxmins, const vec3_t boxmaxs, const vec3_t end, int hitsupercontentsmask);
+	// find the supercontents value at a point in this model
+	int (*PointSuperContents)(struct model_s *model, int frame, const vec3_t point);
 	// fields belonging to some types of model
 	model_sprite_t	sprite;
 	model_brush_t	brush;
@@ -837,8 +871,6 @@ typedef struct model_s
 	model_brushq2_t	brushq2;
 	*/
 	model_brushq3_t	brushq3;
-	// skin files can have different tags for each skin
-	overridetagnameset_t	*data_overridetagnamesforskin;
 	// flags this model for offseting sounds to the model center (used by brush models)
 	int soundfromcenter;
 }
@@ -888,7 +920,7 @@ void Mod_ShadowMesh_Free(shadowmesh_t *mesh);
 
 void Mod_LoadQ3Shaders(void);
 q3shaderinfo_t *Mod_LookupQ3Shader(const char *name);
-qboolean Mod_LoadTextureFromQ3Shader(texture_t *texture, const char *name, qboolean q1bsp, qboolean q3bsp, qboolean md3);
+qboolean Mod_LoadTextureFromQ3Shader(texture_t *texture, const char *name, qboolean warnmissing, qboolean fallback, int defaulttexflags);
 
 extern cvar_t r_mipskins;
 
@@ -923,9 +955,11 @@ int Mod_Q1BSP_SuperContentsFromNativeContents(struct model_s *model, int nativec
 
 // a lot of model formats use the Q1BSP code, so here are the prototypes...
 struct entity_render_s;
+void R_Q1BSP_DrawAddWaterPlanes(struct entity_render_s *ent);
 void R_Q1BSP_DrawSky(struct entity_render_s *ent);
 void R_Q1BSP_Draw(struct entity_render_s *ent);
 void R_Q1BSP_DrawDepth(struct entity_render_s *ent);
+void R_Q1BSP_DrawDebug(struct entity_render_s *ent);
 void R_Q1BSP_GetLightInfo(struct entity_render_s *ent, vec3_t relativelightorigin, float lightradius, vec3_t outmins, vec3_t outmaxs, int *outleaflist, unsigned char *outleafpvs, int *outnumleafspointer, int *outsurfacelist, unsigned char *outsurfacepvs, int *outnumsurfacespointer, unsigned char *outshadowtrispvs, unsigned char *outlighttrispvs);
 void R_Q1BSP_CompileShadowVolume(struct entity_render_s *ent, vec3_t relativelightorigin, vec3_t relativelightdirection, float lightradius, int numsurfaces, const int *surfacelist);
 void R_Q1BSP_DrawShadowVolume(struct entity_render_s *ent, vec3_t relativelightorigin, vec3_t relativelightdirection, float lightradius, int numsurfaces, const int *surfacelist, const vec3_t lightmins, const vec3_t lightmaxs);
@@ -934,7 +968,6 @@ void R_Q1BSP_DrawLight(struct entity_render_s *ent, int numsurfaces, const int *
 // alias models
 struct frameblend_s;
 void Mod_AliasInit(void);
-void Mod_Alias_GetMesh_Vertices(const model_t *model, const struct frameblend_s *frameblend, float *vertex3f, float *normal3f, float *svector3f, float *tvector3f);
 int Mod_Alias_GetTagMatrix(const model_t *model, int poseframe, int tagindex, matrix4x4_t *outmatrix);
 int Mod_Alias_GetTagIndexForName(const model_t *model, unsigned int skin, const char *tagname);
 
