@@ -38,7 +38,7 @@ static cvar_t sv_masters [] =
 	{CVAR_SAVE, "sv_master4", "", "user-chosen master server 4"},
 	{0, "sv_masterextra1", "ghdigital.com", "default master server 1 (admin: LordHavoc)"}, // admin: LordHavoc
 	{0, "sv_masterextra2", "dpmaster.deathmask.net", "default master server 2 (admin: Willis)"}, // admin: Willis
-	{0, "sv_masterextra3", "excalibur.nvg.ntnu.no", "default master server 3 (admin: tChr)"}, // admin: tChr
+	{0, "sv_masterextra3", "dpmaster.tchr.no", "default master server 3 (admin: tChr)"}, // admin: tChr
 	{0, NULL, NULL, NULL}
 };
 
@@ -48,19 +48,11 @@ static cvar_t sv_qwmasters [] =
 	{CVAR_SAVE, "sv_qwmaster2", "", "user-chosen qwmaster server 2"},
 	{CVAR_SAVE, "sv_qwmaster3", "", "user-chosen qwmaster server 3"},
 	{CVAR_SAVE, "sv_qwmaster4", "", "user-chosen qwmaster server 4"},
-	{0, "sv_qwmasterextra1", "192.246.40.37:27000", "id Limbo (admin: id Software)"},
-	{0, "sv_qwmasterextra2", "192.246.40.37:27002", "id CTF (admin: id Software)"},
-	{0, "sv_qwmasterextra3", "192.246.40.37:27003", "id TeamFortress (admin: id Software)"},
-	{0, "sv_qwmasterextra4", "192.246.40.37:27004", "id Miscilaneous (admin: id Software)"},
-	{0, "sv_qwmasterextra5", "192.246.40.37:27006", "id Deathmatch Only (admin: id Software)"},
-	{0, "sv_qwmasterextra6", "150.254.66.120:27000", "Poland's master server. (admin: unknown)"},
-	{0, "sv_qwmasterextra7", "62.112.145.129:27000", "Ocrana master server. (admin: unknown)"},
-	{0, "sv_qwmasterextra8", "master.edome.net", "edome master server. (admin: unknown)"},
-	{0, "sv_qwmasterextra9", "qwmaster.barrysworld.com", "barrysworld master server. (admin: unknown)"},
-	{0, "sv_qwmasterextra10", "qwmaster.ocrana.de:27000", "Ocrana2 master server. (admin: unknown)"},
-	{0, "sv_qwmasterextra11", "213.221.174.165:27000", "unknown1 master server. (admin: unknown)"},
-	{0, "sv_qwmasterextra12", "195.74.0.8", "unknown2 master server. (admin: unknown)"},
-	{0, "sv_qwmasterextra13", "204.182.161.2", "unknown3 master server. (admin: unknown)"},
+	{0, "sv_qwmasterextra1", "master.quakeservers.net:27000", "Global master server. (admin: unknown)"},
+	{0, "sv_qwmasterextra2", "asgaard.morphos-team.net:27000", "Global master server. (admin: unknown)"},
+	{0, "sv_qwmasterextra3", "qwmaster.ocrana.de:27000", "German master server. (admin: unknown)"},
+	{0, "sv_qwmasterextra4", "masterserver.exhale.de:27000", "German master server. (admin: unknown)"},
+	{0, "sv_qwmasterextra5", "kubus.rulez.pl:27000", "Poland master server. (admin: unknown)"},
 	{0, NULL, NULL, NULL}
 };
 
@@ -69,7 +61,6 @@ static double nextheartbeattime = 0;
 sizebuf_t net_message;
 static unsigned char net_message_buf[NET_MAXMESSAGE];
 
-cvar_t net_banlist = {CVAR_SAVE, "net_banlist","", "space separated list of networks/IPs to be banned in a.b.c.d/nn or a.b.c.d/e.f.g.h notation"};
 cvar_t net_messagetimeout = {0, "net_messagetimeout","300", "drops players who have not sent any packets for this many seconds"};
 cvar_t net_connecttimeout = {0, "net_connecttimeout","10", "after requesting a connection, the client must reply within this many seconds or be dropped (cuts down on connect floods)"};
 cvar_t net_connectfloodblockingtimeout = {0, "net_connectfloodblockingtimeout", "5", "when a connection packet is received, it will block all future connect packets from that IP address for this many seconds (cuts down on connect floods)"};
@@ -82,9 +73,12 @@ static cvar_t cl_netpacketloss_receive = {0, "cl_netpacketloss_receive","0", "dr
 static cvar_t net_slist_queriespersecond = {0, "net_slist_queriespersecond", "20", "how many server information requests to send per second"};
 static cvar_t net_slist_queriesperframe = {0, "net_slist_queriesperframe", "4", "maximum number of server information requests to send each rendered frame (guards against low framerates causing problems)"};
 static cvar_t net_slist_timeout = {0, "net_slist_timeout", "4", "how long to listen for a server information response before giving up"};
+static cvar_t net_slist_pause = {0, "net_slist_pause", "0", "when set to 1, the server list won't update until it is set back to 0"};
 static cvar_t net_slist_maxtries = {0, "net_slist_maxtries", "3", "how many times to ask the same server for information (more times gives better ping reports but takes longer)"};
 
 static cvar_t gameversion = {0, "gameversion", "0", "version of game data (mod-specific), when client and server gameversion mismatch in the server browser the server is shown as incompatible"};
+static cvar_t rcon_restricted_password = {CVAR_PRIVATE, "rcon_restricted_password", "", "password to authenticate rcon commands in restricted mode"};
+static cvar_t rcon_restricted_commands = {0, "rcon_restricted_commands", "", "allowed commands for rcon when the restricted mode password was used"};
 
 /* statistic counters */
 static int packetsSent = 0;
@@ -106,6 +100,7 @@ int serverreplycount = 0;
 
 // this is only false if there are still servers left to query
 static qboolean serverlist_querysleep = true;
+static qboolean serverlist_paused = false;
 // this is pushed a second or two ahead of realtime whenever a master server
 // reply is received, to avoid issuing queries while master replies are still
 // flooding in (which would make a mess of the ping times)
@@ -192,6 +187,9 @@ static qboolean _ServerList_Entry_Compare( serverlist_entry_t *A, serverlist_ent
 		case SLIF_NUMHUMANS:
 			result = A->info.numhumans - B->info.numhumans;
 			break;
+		case SLIF_FREESLOTS:
+			result = A->info.freeslots - B->info.freeslots;
+			break;
 		case SLIF_PROTOCOL:
 			result = A->info.protocol - B->info.protocol;
 			break;
@@ -199,16 +197,16 @@ static qboolean _ServerList_Entry_Compare( serverlist_entry_t *A, serverlist_ent
 			result = strcmp( B->info.cname, A->info.cname );
 			break;
 		case SLIF_GAME:
-			result = strcmp( B->info.game, A->info.game );
+			result = strcasecmp( B->info.game, A->info.game );
 			break;
 		case SLIF_MAP:
-			result = strcmp( B->info.map, A->info.map );
+			result = strcasecmp( B->info.map, A->info.map );
 			break;
 		case SLIF_MOD:
-			result = strcmp( B->info.mod, A->info.mod );
+			result = strcasecmp( B->info.mod, A->info.mod );
 			break;
 		case SLIF_NAME:
-			result = strcmp( B->info.name, A->info.name );
+			result = strcasecmp( B->info.name, A->info.name );
 			break;
 		default:
 			Con_DPrint( "_ServerList_Entry_Compare: Bad serverlist_sortbyfield!\n" );
@@ -297,6 +295,8 @@ static qboolean _ServerList_Entry_Mask( serverlist_mask_t *mask, serverlist_info
 		return false;
 	if( !_ServerList_CompareInt( info->numhumans, mask->tests[SLIF_NUMHUMANS], mask->info.numhumans ) )
 		return false;
+	if( !_ServerList_CompareInt( info->freeslots, mask->tests[SLIF_FREESLOTS], mask->info.freeslots ) )
+		return false;
 	if( !_ServerList_CompareInt( info->protocol, mask->tests[SLIF_PROTOCOL], mask->info.protocol ))
 		return false;
 	if( *mask->info.cname
@@ -387,9 +387,12 @@ void ServerList_RebuildViewList(void)
 	int i;
 
 	serverlist_viewcount = 0;
-	for( i = 0 ; i < serverlist_cachecount ; i++ )
-		if( serverlist_cache[i].query == SQS_QUERIED )
-			ServerList_ViewList_Insert( &serverlist_cache[i] );
+	for( i = 0 ; i < serverlist_cachecount ; i++ ) {
+		serverlist_entry_t *entry = &serverlist_cache[i];
+		// also display entries that are currently being refreshed [11/8/2007 Black]
+		if( entry->query == SQS_QUERIED || entry->query == SQS_REFRESHING )
+			ServerList_ViewList_Insert( entry );
+	}
 }
 
 void ServerList_ResetMasks(void)
@@ -405,6 +408,22 @@ void ServerList_ResetMasks(void)
 		serverlist_ormasks[i].info.numbots = -1;
 }
 
+void ServerList_GetPlayerStatistics(int *numplayerspointer, int *maxplayerspointer)
+{
+	int i;
+	int numplayers = 0, maxplayers = 0;
+	for (i = 0;i < serverlist_cachecount;i++)
+	{
+		if (serverlist_cache[i].query == SQS_QUERIED)
+		{
+			numplayers += serverlist_cache[i].info.numhumans;
+			maxplayers += serverlist_cache[i].info.maxplayers;
+		}
+	}
+	*numplayerspointer = numplayers;
+	*maxplayerspointer = maxplayers;
+}
+
 #if 0
 static void _ServerList_Test(void)
 {
@@ -421,16 +440,26 @@ static void _ServerList_Test(void)
 }
 #endif
 
-void ServerList_QueryList(qboolean querydp, qboolean queryqw)
+void ServerList_QueryList(qboolean resetcache, qboolean querydp, qboolean queryqw, qboolean consoleoutput)
 {
 	masterquerytime = realtime;
 	masterquerycount = 0;
 	masterreplycount = 0;
-	serverquerycount = 0;
-	serverreplycount = 0;
-	serverlist_cachecount = 0;
-	serverlist_viewcount = 0;
-	serverlist_consoleoutput = false;
+	if( resetcache ) {
+		serverquerycount = 0;
+		serverreplycount = 0;
+		serverlist_cachecount = 0;
+		serverlist_viewcount = 0;
+	} else {
+		// refresh all entries
+		int n;
+		for( n = 0 ; n < serverlist_cachecount ; n++ ) {
+			serverlist_entry_t *entry = &serverlist_cache[ n ];
+			entry->query = SQS_REFRESHING;
+			entry->querycounter = 0;
+		}
+	}
+	serverlist_consoleoutput = consoleoutput;
 
 	//_ServerList_Test();
 
@@ -506,7 +535,7 @@ qboolean NetConn_CanSend(netconn_t *conn)
 	}
 }
 
-int NetConn_SendUnreliableMessage(netconn_t *conn, sizebuf_t *data, protocolversion_t protocol, int rate)
+int NetConn_SendUnreliableMessage(netconn_t *conn, sizebuf_t *data, protocolversion_t protocol, int rate, qboolean quakesignon_suppressreliables)
 {
 	int totallen = 0;
 
@@ -538,18 +567,18 @@ int NetConn_SendUnreliableMessage(netconn_t *conn, sizebuf_t *data, protocolvers
 			sendreliable = true;
 		}
 		// outgoing unreliable packet number, and outgoing reliable packet number (0 or 1)
-		*((int *)(sendbuffer + 0)) = LittleLong((unsigned int)conn->qw.outgoing_sequence | ((unsigned int)sendreliable<<31));
+		*((int *)(sendbuffer + 0)) = LittleLong((unsigned int)conn->outgoing_unreliable_sequence | ((unsigned int)sendreliable<<31));
 		// last received unreliable packet number, and last received reliable packet number (0 or 1)
 		*((int *)(sendbuffer + 4)) = LittleLong((unsigned int)conn->qw.incoming_sequence | ((unsigned int)conn->qw.incoming_reliable_sequence<<31));
 		packetLen = 8;
-		conn->qw.outgoing_sequence++;
+		conn->outgoing_unreliable_sequence++;
 		// client sends qport in every packet
 		if (conn == cls.netcon)
 		{
 			*((short *)(sendbuffer + 8)) = LittleShort(cls.qw_qport);
 			packetLen += 2;
 			// also update cls.qw_outgoing_sequence
-			cls.qw_outgoing_sequence = conn->qw.outgoing_sequence;
+			cls.qw_outgoing_sequence = conn->outgoing_unreliable_sequence;
 		}
 		if (packetLen + (sendreliable ? conn->sendMessageLength : 0) > 1400)
 		{
@@ -565,7 +594,7 @@ int NetConn_SendUnreliableMessage(netconn_t *conn, sizebuf_t *data, protocolvers
 			conn->outgoing_reliablesize[conn->outgoing_packetcounter] += conn->sendMessageLength;
 			memcpy(sendbuffer + packetLen, conn->sendMessage, conn->sendMessageLength);
 			packetLen += conn->sendMessageLength;
-			conn->qw.last_reliable_sequence = conn->qw.outgoing_sequence;
+			conn->qw.last_reliable_sequence = conn->outgoing_unreliable_sequence;
 		}
 
 		// add the unreliable message if possible
@@ -623,7 +652,7 @@ int NetConn_SendUnreliableMessage(netconn_t *conn, sizebuf_t *data, protocolvers
 		}
 
 		// if we have a new reliable message to send, do so
-		if (!conn->sendMessageLength && conn->message.cursize)
+		if (!conn->sendMessageLength && conn->message.cursize && !quakesignon_suppressreliables)
 		{
 			if (conn->message.cursize > (int)sizeof(conn->sendMessage))
 			{
@@ -686,10 +715,10 @@ int NetConn_SendUnreliableMessage(netconn_t *conn, sizebuf_t *data, protocolvers
 
 			header = (unsigned int *)sendbuffer;
 			header[0] = BigLong(packetLen | NETFLAG_UNRELIABLE);
-			header[1] = BigLong(conn->nq.unreliableSendSequence);
+			header[1] = BigLong(conn->outgoing_unreliable_sequence);
 			memcpy(sendbuffer + NET_HEADERSIZE, data->data, data->cursize);
 
-			conn->nq.unreliableSendSequence++;
+			conn->outgoing_unreliable_sequence++;
 
 			conn->outgoing_unreliablesize[conn->outgoing_packetcounter] += packetLen;
 
@@ -712,6 +741,16 @@ int NetConn_SendUnreliableMessage(netconn_t *conn, sizebuf_t *data, protocolvers
 	return 0;
 }
 
+qboolean NetConn_HaveClientPorts(void)
+{
+	return !!cl_numsockets;
+}
+
+qboolean NetConn_HaveServerPorts(void)
+{
+	return !!sv_numsockets;
+}
+
 void NetConn_CloseClientPorts(void)
 {
 	for (;cl_numsockets > 0;cl_numsockets--)
@@ -749,7 +788,10 @@ void NetConn_OpenClientPorts(void)
 	port = bound(0, cl_netport.integer, 65535);
 	if (cl_netport.integer != port)
 		Cvar_SetValueQuick(&cl_netport, port);
-	Con_Printf("Client using port %i\n", port);
+	if(port == 0)
+		Con_Printf("Client using an automatically assigned port\n");
+	else
+		Con_Printf("Client using port %i\n", port);
 	NetConn_OpenClientPort("local:2", 0);
 	NetConn_OpenClientPort(net_address.string, port);
 	//NetConn_OpenClientPort(net_address_ipv6.string, port);
@@ -1064,7 +1106,7 @@ static int NetConn_ReceivedMessage(netconn_t *conn, unsigned char *data, int len
 							unsigned int *header;
 
 							conn->sendMessageLength -= MAX_PACKETFRAGMENT;
-							memcpy(conn->sendMessage, conn->sendMessage+MAX_PACKETFRAGMENT, conn->sendMessageLength);
+							memmove(conn->sendMessage, conn->sendMessage+MAX_PACKETFRAGMENT, conn->sendMessageLength);
 
 							if (conn->sendMessageLength <= MAX_PACKETFRAGMENT)
 							{
@@ -1166,7 +1208,6 @@ void NetConn_ConnectionEstablished(lhnetsocket_t *mysocket, lhnetaddress_t *peer
 	cls.signon = 0;				// need all the signon messages before playing
 	cls.protocol = initialprotocol;
 	// reset move sequence numbering on this new connection
-	cls.movesequence = 1;
 	cls.servermovesequence = 0;
 	if (cls.protocol == PROTOCOL_QUAKEWORLD)
 		Cmd_ForwardStringToServer("new");
@@ -1180,7 +1221,7 @@ void NetConn_ConnectionEstablished(lhnetsocket_t *mysocket, lhnetaddress_t *peer
 		msg.data = buf;
 		msg.maxsize = sizeof(buf);
 		MSG_WriteChar(&msg, clc_nop);
-		NetConn_SendUnreliableMessage(cls.netcon, &msg, cls.protocol, 10000);
+		NetConn_SendUnreliableMessage(cls.netcon, &msg, cls.protocol, 10000, false);
 	}
 }
 
@@ -1191,74 +1232,19 @@ int NetConn_IsLocalGame(void)
 	return false;
 }
 
-static int NetConn_IsBanned(lhnetaddress_t *peeraddress)
-{
-	const char *banlist = net_banlist.string;
-	const char *p;
-	char buf[128];
-	size_t len;
-	lhnetaddress_t address, netmask;
-
-	if(!*banlist)
-		return false; // no ban list -> no parsing
-
-	if(peeraddress->addresstype == LHNETADDRESSTYPE_LOOP)
-		return false; // can't ban local
-
-	if(peeraddress->addresstype == LHNETADDRESSTYPE_NONE)
-		return false; // WTF?
-
-	while(*banlist)
-	{
-		while(*banlist == ' ')
-			++banlist;
-		if(!*banlist)
-			break;
-		p = strchr(banlist, ' ');
-		if(!p)
-			p = banlist + strlen(banlist);
-
-		len = p - banlist;
-		if(len < sizeof(buf))
-		{
-			memcpy(buf, banlist, len);
-			buf[len] = 0;
-
-			if(LHNETADDRESS_NetworkFromString(&address, &netmask, buf))
-			{
-				if(developer.integer >= 100)
-				{
-					char addressstring[128];
-					LHNETADDRESS_ToString(&address, addressstring, sizeof(addressstring), false);
-					Con_DPrintf("Banlist check: address %s/", addressstring);
-					LHNETADDRESS_ToString(&netmask, addressstring, sizeof(addressstring), false);
-					Con_DPrintf("%s against ", addressstring);
-					LHNETADDRESS_ToString(peeraddress, addressstring, sizeof(addressstring), false);
-					Con_DPrintf("%s\n", addressstring);
-				}
-				if(LHNETADDRESS_InNetwork(peeraddress, &address, &netmask))
-					return true;
-			}
-			else
-				Con_DPrintf("invalid address in net_banlist: %s\n", buf);
-		}
-		else
-			Con_DPrintf("invalid address in net_banlist: (too long)\n");
-
-		banlist = p;
-	}
-
-	return false;
-}
-
 static int NetConn_ClientParsePacket_ServerList_ProcessReply(const char *addressstring)
 {
 	int n;
 	int pingtime;
+	serverlist_entry_t *entry = NULL;
+
 	// search the cache for this server and update it
-	for (n = 0;n < serverlist_cachecount;n++)
-		if (!strcmp(addressstring, serverlist_cache[n].info.cname))
+	for (n = 0;n < serverlist_cachecount;n++) {
+		entry = &serverlist_cache[ n ];
+		if (!strcmp(addressstring, entry->info.cname))
 			break;
+	}
+
 	if (n == serverlist_cachecount)
 	{
 		// LAN search doesnt require an answer from the master server so we wont
@@ -1268,43 +1254,93 @@ static int NetConn_ClientParsePacket_ServerList_ProcessReply(const char *address
 		if (serverlist_cachecount == SERVERLIST_TOTALSIZE)
 			return -1;
 
-		memset(&serverlist_cache[serverlist_cachecount], 0, sizeof(serverlist_cache[serverlist_cachecount]));
+		entry = &serverlist_cache[n];
+
+		memset(entry, 0, sizeof(*entry));
 		// store the data the engine cares about (address and ping)
-		strlcpy(serverlist_cache[serverlist_cachecount].info.cname, addressstring, sizeof(serverlist_cache[serverlist_cachecount].info.cname));
-		serverlist_cache[serverlist_cachecount].info.ping = 100000;
-		serverlist_cache[serverlist_cachecount].querytime = realtime;
+		strlcpy(entry->info.cname, addressstring, sizeof(entry->info.cname));
+		entry->info.ping = 100000;
+		entry->querytime = realtime;
 		// if not in the slist menu we should print the server to console
 		if (serverlist_consoleoutput)
 			Con_Printf("querying %s\n", addressstring);
 		++serverlist_cachecount;
 	}
 	// if this is the first reply from this server, count it as having replied
-	if (serverlist_cache[n].info.ping == 100000)
-		serverreplycount++;
-	pingtime = (int)((realtime - serverlist_cache[n].querytime) * 1000.0 + 0.5);
+	pingtime = (int)((realtime - entry->querytime) * 1000.0 + 0.5);
 	pingtime = bound(0, pingtime, 9999);
-	// update the ping
-	serverlist_cache[n].info.ping = min(serverlist_cache[n].info.ping, pingtime);
+	if (entry->query == SQS_REFRESHING) {
+		entry->info.ping = pingtime;
+		entry->query = SQS_QUERIED;
+	} else {
+		// convert to unsigned to catch the -1
+		// I still dont like this but its better than the old 10000 magic ping number - as in easier to type and read :( [11/8/2007 Black]
+		entry->info.ping = min((unsigned) entry->info.ping, (unsigned) pingtime);
+		serverreplycount++;
+	}
+	
 	// other server info is updated by the caller
 	return n;
 }
 
 static void NetConn_ClientParsePacket_ServerList_UpdateCache(int n)
 {
-	serverlist_info_t *info = &serverlist_cache[n].info;
+	serverlist_entry_t *entry = &serverlist_cache[n];
+	serverlist_info_t *info = &entry->info;
 	// update description strings for engine menu and console output
-	dpsnprintf(serverlist_cache[n].line1, sizeof(serverlist_cache[n].line1), "^%c%5d^7 ^%c%3u^7/%3u %-65.65s", info->ping >= 300 ? '1' : (info->ping >= 200 ? '3' : '7'), (int)info->ping, ((info->numhumans > 0 && info->numhumans < info->maxplayers) ? (info->numhumans >= 4 ? '7' : '3') : '1'), info->numplayers, info->maxplayers, info->name);
-	dpsnprintf(serverlist_cache[n].line2, sizeof(serverlist_cache[n].line2), "^4%-21.21s %-19.19s ^%c%-17.17s^4 %-20.20s", info->cname, info->game, (info->gameversion != gameversion.integer) ? '1' : '4', info->mod, info->map);
-	if (serverlist_cache[n].query == SQS_QUERIED)
-		ServerList_ViewList_Remove(&serverlist_cache[n]);
+	dpsnprintf(entry->line1, sizeof(serverlist_cache[n].line1), "^%c%5d^7 ^%c%3u^7/%3u %-65.65s", info->ping >= 300 ? '1' : (info->ping >= 200 ? '3' : '7'), (int)info->ping, ((info->numhumans > 0 && info->numhumans < info->maxplayers) ? (info->numhumans >= 4 ? '7' : '3') : '1'), info->numplayers, info->maxplayers, info->name);
+	dpsnprintf(entry->line2, sizeof(serverlist_cache[n].line2), "^4%-21.21s %-19.19s ^%c%-17.17s^4 %-20.20s", info->cname, info->game, (info->gameversion != gameversion.integer) ? '1' : '4', info->mod, info->map);
+	if (entry->query == SQS_QUERIED)
+	{
+		if(!serverlist_paused)
+			ServerList_ViewList_Remove(entry);
+	}
 	// if not in the slist menu we should print the server to console (if wanted)
 	else if( serverlist_consoleoutput )
 		Con_Printf("%s\n%s\n", serverlist_cache[n].line1, serverlist_cache[n].line2);
 	// and finally, update the view set
-	ServerList_ViewList_Insert( &serverlist_cache[n] );
+	if(!serverlist_paused)
+		ServerList_ViewList_Insert( entry );
+	//	update the entry's state
 	serverlist_cache[n].query = SQS_QUERIED;
 }
 
+// returns true, if it's sensible to continue the processing
+static qboolean NetConn_ClientParsePacket_ServerList_PrepareQuery( int protocol, const char *ipstring ) {
+	int n;
+	serverlist_entry_t *entry;
+
+	//	ignore the rest of the message if the serverlist is full
+	if( serverlist_cachecount == SERVERLIST_TOTALSIZE )
+		return false;
+	//	also ignore	it	if	we	have already queried	it	(other master server	response)
+	for( n =	0 ; n	< serverlist_cachecount	; n++	)
+		if( !strcmp( ipstring, serverlist_cache[ n ].info.cname ) )
+			break;
+
+	entry = &serverlist_cache[n];
+
+	if( n < serverlist_cachecount ) {
+		// the entry has already been queried once or 
+		return true;
+	}
+
+	memset(entry, 0, sizeof(entry));
+	entry->protocol =	protocol;
+	//	store	the data	the engine cares about (address and	ping)
+	strlcpy (entry->info.cname, ipstring, sizeof(entry->info.cname));
+	
+	// no, then reset the ping right away
+	entry->info.ping = -1;
+	// we also want to increase the serverlist_cachecount then
+	serverlist_cachecount++;
+	serverquerycount++;
+
+	entry->query =	SQS_QUERYING;
+
+	return true;
+}
+
 static int NetConn_ClientParsePacket(lhnetsocket_t *mysocket, unsigned char *data, int length, lhnetaddress_t *peeraddress)
 {
 	qboolean fromserver;
@@ -1372,9 +1408,6 @@ static int NetConn_ClientParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 			serverlist_info_t *info;
 			int n;
 
-			if(NetConn_IsBanned(peeraddress))
-				return true; // just ignore
-
 			string += 13;
 			// search the cache for this server and update it
 			n = NetConn_ClientParsePacket_ServerList_ProcessReply(addressstring2);
@@ -1401,6 +1434,7 @@ static int NetConn_ClientParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 			if ((s = SearchInfostring(string, "sv_maxclients")) != NULL) info->maxplayers = atoi(s);
 			if ((s = SearchInfostring(string, "gameversion"  )) != NULL) info->gameversion = atoi(s);
 			info->numhumans = info->numplayers - max(0, info->numbots);
+			info->freeslots = info->maxplayers - info->numplayers;
 
 			NetConn_ClientParsePacket_ServerList_UpdateCache(n);
 
@@ -1416,30 +1450,12 @@ static int NetConn_ClientParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 				Con_Print("received DarkPlaces server list...\n");
 			while (length >= 7 && data[0] == '\\' && (data[1] != 0xFF || data[2] != 0xFF || data[3] != 0xFF || data[4] != 0xFF) && data[5] * 256 + data[6] != 0)
 			{
-				int n;
-
 				dpsnprintf (ipstring, sizeof (ipstring), "%u.%u.%u.%u:%u", data[1], data[2], data[3], data[4], data[5] * 256 + data[6]);
 				if (serverlist_consoleoutput && developer_networking.integer)
 					Con_Printf("Requesting info from DarkPlaces server %s\n", ipstring);
-				// ignore the rest of the message if the serverlist is full
-				if( serverlist_cachecount == SERVERLIST_TOTALSIZE )
+				
+				if( !NetConn_ClientParsePacket_ServerList_PrepareQuery( PROTOCOL_DARKPLACES7, ipstring ) ) {
 					break;
-				// also ignore it if we have already queried it (other master server response)
-				for( n = 0 ; n < serverlist_cachecount ; n++ )
-					if( !strcmp( ipstring, serverlist_cache[ n ].info.cname ) )
-						break;
-				if( n >= serverlist_cachecount )
-				{
-					serverquerycount++;
-
-					memset(&serverlist_cache[serverlist_cachecount], 0, sizeof(serverlist_cache[serverlist_cachecount]));
-					serverlist_cache[serverlist_cachecount].protocol = PROTOCOL_DARKPLACES7;
-					// store the data the engine cares about (address and ping)
-					strlcpy (serverlist_cache[serverlist_cachecount].info.cname, ipstring, sizeof (serverlist_cache[serverlist_cachecount].info.cname));
-					serverlist_cache[serverlist_cachecount].info.ping = 100000;
-					serverlist_cache[serverlist_cachecount].query = SQS_QUERYING;
-
-					++serverlist_cachecount;
 				}
 
 				// move on to next address in packet
@@ -1461,30 +1477,12 @@ static int NetConn_ClientParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 				Con_Printf("received QuakeWorld server list from %s...\n", addressstring2);
 			while (length >= 6 && (data[0] != 0xFF || data[1] != 0xFF || data[2] != 0xFF || data[3] != 0xFF) && data[4] * 256 + data[5] != 0)
 			{
-				int n;
-
 				dpsnprintf (ipstring, sizeof (ipstring), "%u.%u.%u.%u:%u", data[0], data[1], data[2], data[3], data[4] * 256 + data[5]);
 				if (serverlist_consoleoutput && developer_networking.integer)
 					Con_Printf("Requesting info from QuakeWorld server %s\n", ipstring);
-				// ignore the rest of the message if the serverlist is full
-				if( serverlist_cachecount == SERVERLIST_TOTALSIZE )
+				
+				if( !NetConn_ClientParsePacket_ServerList_PrepareQuery( PROTOCOL_QUAKEWORLD, ipstring ) ) {
 					break;
-				// also ignore it if we have already queried it (other master server response)
-				for( n = 0 ; n < serverlist_cachecount ; n++ )
-					if( !strcmp( ipstring, serverlist_cache[ n ].info.cname ) )
-						break;
-				if( n >= serverlist_cachecount )
-				{
-					serverquerycount++;
-
-					memset(&serverlist_cache[serverlist_cachecount], 0, sizeof(serverlist_cache[serverlist_cachecount]));
-					serverlist_cache[serverlist_cachecount].protocol = PROTOCOL_QUAKEWORLD;
-					// store the data the engine cares about (address and ping)
-					strlcpy (serverlist_cache[serverlist_cachecount].info.cname, ipstring, sizeof (serverlist_cache[serverlist_cachecount].info.cname));
-					serverlist_cache[serverlist_cachecount].info.ping = 100000;
-					serverlist_cache[serverlist_cachecount].query = SQS_QUERYING;
-
-					++serverlist_cachecount;
 				}
 
 				// move on to next address in packet
@@ -1555,6 +1553,7 @@ static int NetConn_ClientParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 			if ((s = SearchInfostring(string, "hostname"     )) != NULL) strlcpy(info->name, s, sizeof (info->name));else info->name[0] = 0;
 			info->protocol = 0;
 			info->numplayers = 0; // updated below
+			info->numhumans = 0; // updated below
 			if ((s = SearchInfostring(string, "maxclients"   )) != NULL) info->maxplayers = atoi(s);else info->maxplayers  = 0;
 			if ((s = SearchInfostring(string, "gameversion"  )) != NULL) info->gameversion = atoi(s);else info->gameversion = 0;
 
@@ -1571,6 +1570,7 @@ static int NetConn_ClientParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 					if (s >= string + length)
 						break;
 					info->numplayers++;
+					info->numhumans++;
 					s++;
 				}
 			}
@@ -1684,6 +1684,10 @@ void NetConn_QueryQueueFrame(void)
 	double timeouttime;
 	static double querycounter = 0;
 
+	if(!net_slist_pause.integer && serverlist_paused)
+		ServerList_RebuildViewList();
+	serverlist_paused = net_slist_pause.integer;
+
 	if (serverlist_querysleep)
 		return;
 
@@ -1702,41 +1706,42 @@ void NetConn_QueryQueueFrame(void)
 		return;
 	}
 
-	// scan serverlist and issue queries as needed
-    serverlist_querysleep = true;
+	//	scan serverlist and issue queries as needed
+	serverlist_querysleep = true;
 
-	timeouttime = realtime - net_slist_timeout.value;
-	for( index = 0, queries = 0 ; index < serverlist_cachecount && queries < maxqueries ; index++ )
+	timeouttime	= realtime - net_slist_timeout.value;
+	for( index = 0, queries	= 0 ;	index	< serverlist_cachecount	&&	queries < maxqueries	; index++ )
 	{
 		serverlist_entry_t *entry = &serverlist_cache[ index ];
-		if( entry->query != SQS_QUERYING )
+		if( entry->query != SQS_QUERYING && entry->query != SQS_REFRESHING )
 		{
 			continue;
 		}
 
-        serverlist_querysleep = false;
-		if( entry->querycounter != 0 && entry->querytime > timeouttime )
+		serverlist_querysleep	= false;
+		if( entry->querycounter	!=	0 && entry->querytime >	timeouttime	)
 		{
 			continue;
 		}
 
-		if( entry->querycounter != (unsigned) net_slist_maxtries.integer )
+		if( entry->querycounter	!=	(unsigned) net_slist_maxtries.integer )
 		{
-			lhnetaddress_t address;
+			lhnetaddress_t	address;
 			int socket;
 
 			LHNETADDRESS_FromString(&address, entry->info.cname, 0);
-			if (entry->protocol == PROTOCOL_QUAKEWORLD)
+			if	(entry->protocol == PROTOCOL_QUAKEWORLD)
 			{
-				for (socket = 0; socket < cl_numsockets ; socket++)
+				for (socket	= 0; socket	< cl_numsockets ;	socket++)
 					NetConn_WriteString(cl_sockets[socket], "\377\377\377\377status\n", &address);
 			}
 			else
 			{
-				for (socket = 0; socket < cl_numsockets ; socket++)
+				for (socket	= 0; socket	< cl_numsockets ;	socket++)
 					NetConn_WriteString(cl_sockets[socket], "\377\377\377\377getinfo", &address);
 			}
 
+			//	update the entry fields
 			entry->querytime = realtime;
 			entry->querycounter++;
 
@@ -1748,6 +1753,13 @@ void NetConn_QueryQueueFrame(void)
 		}
 		else
 		{
+			// have we tried to refresh this server?
+			if( entry->query == SQS_REFRESHING ) {
+				// yes, so update the reply count (since its not responding anymore)
+				serverreplycount--;
+				if(!serverlist_paused)
+					ServerList_ViewList_Remove(entry);
+			}
 			entry->query = SQS_TIMEDOUT;
 		}
 	}
@@ -1787,7 +1799,7 @@ void NetConn_ClientFrame(void)
 		while (cl_sockets[i] && (length = NetConn_Read(cl_sockets[i], readbuffer, sizeof(readbuffer), &peeraddress)) > 0)
 			NetConn_ClientParsePacket(cl_sockets[i], readbuffer, length, &peeraddress);
 	NetConn_QueryQueueFrame();
-	if (cls.netcon && realtime > cls.netcon->timeout)
+	if (cls.netcon && realtime > cls.netcon->timeout && !sv.active)
 	{
 		Con_Print("Connection timed out\n");
 		CL_Disconnect();
@@ -1963,6 +1975,41 @@ void NetConn_ClearConnectFlood(lhnetaddress_t *peeraddress)
 	}
 }
 
+// returns a string describing the user level, or NULL for auth failure
+const char *RCon_Authenticate(const char *password, const char *s, const char *endpos)
+{
+	const char *text;
+
+	if(!strcmp(rcon_password.string, password))
+		return "rcon";
+	
+	if(strcmp(rcon_restricted_password.string, password))
+		return NULL;
+
+	for(text = s; text != endpos; ++text)
+		if(*text > 0 && (*text < ' ' || *text == ';'))
+			return NULL; // block possible exploits against the parser/alias expansion
+
+	while(s != endpos)
+	{
+		size_t l = strlen(s);
+		if(l)
+		{
+			text = s;
+
+			if (!COM_ParseToken_Console(&text))
+				return NULL;
+
+			// com_token now contains the command
+			if(!strstr(va(" %s ", rcon_restricted_commands.string), va(" %s ", com_token)))
+				return NULL;
+		}
+		s += l + 1;
+	}
+
+	return "restricted rcon";
+}
+
 extern void SV_SendServerinfo (client_t *client);
 static int NetConn_ServerParsePacket(lhnetsocket_t *mysocket, unsigned char *data, int length, lhnetaddress_t *peeraddress)
 {
@@ -2041,13 +2088,6 @@ static int NetConn_ServerParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 			if (i == MAX_CHALLENGES)
 				return true;
 
-			if(NetConn_IsBanned(peeraddress))
-			{
-				Con_Printf("Rejecting connection from banned client %s.\n", addressstring2);
-				NetConn_WriteString(mysocket, "\377\377\377\377reject You are not welcome on this server.", peeraddress);
-				return true;
-			}
-
 			// check engine protocol
 			if (strcmp(SearchInfostring(string, "protocol"), "darkplaces 3"))
 			{
@@ -2152,38 +2192,64 @@ static int NetConn_ServerParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 		{
 			int i;
 			char *s = string + 5;
+			char *endpos = string + length + 1; // one behind the NUL, so adding strlen+1 will eventually reach it
 			char password[64];
 			for (i = 0;*s > ' ';s++)
 				if (i < (int)sizeof(password) - 1)
 					password[i++] = *s;
+			if(*s <= ' ' && s != endpos) // skip leading ugly space
+				++s;
 			password[i] = 0;
-			if (password[0] > ' ' && !strcmp(rcon_password.string, password))
+			if (password[0] > ' ')
 			{
-				// looks like a legitimate rcon command with the correct password
-				Con_Printf("server received rcon command from %s:\n%s\n", host_client ? host_client->name : addressstring2, s);
-				rcon_redirect = true;
-				rcon_redirect_bufferpos = 0;
-				Cmd_ExecuteString(s, src_command);
-				rcon_redirect_buffer[rcon_redirect_bufferpos] = 0;
-				rcon_redirect = false;
-				// print resulting text to client
-				// if client is playing, send a reliable reply instead of
-				// a command packet
-				if (host_client)
+				const char *userlevel = RCon_Authenticate(password, s, endpos);
+				if(userlevel)
 				{
-					// if the netconnection is loop, then this is the
-					// local player on a listen mode server, and it would
-					// result in duplicate printing to the console
-					// (not that the local player should be using rcon
-					//  when they have the console)
-					if (host_client->netconnection && LHNETADDRESS_GetAddressType(&host_client->netconnection->peeraddress) != LHNETADDRESSTYPE_LOOP)
-						SV_ClientPrintf("%s", rcon_redirect_buffer);
+					// looks like a legitimate rcon command with the correct password
+					char *s_ptr = s;
+					Con_Printf("server received %s command from %s: ", userlevel, host_client ? host_client->name : addressstring2);
+					while(s_ptr != endpos)
+					{
+						size_t l = strlen(s_ptr);
+						if(l)
+							Con_Printf(" %s;", s_ptr);
+						s_ptr += l + 1;
+					}
+					Con_Printf("\n");
+					rcon_redirect = true;
+					rcon_redirect_bufferpos = 0;
+					while(s != endpos)
+					{
+						size_t l = strlen(s);
+						if(l)
+							Cmd_ExecuteString(s, src_command);
+						s += l + 1;
+					}
+					rcon_redirect_buffer[rcon_redirect_bufferpos] = 0;
+					rcon_redirect = false;
+					// print resulting text to client
+					// if client is playing, send a reliable reply instead of
+					// a command packet
+					if (host_client)
+					{
+						// if the netconnection is loop, then this is the
+						// local player on a listen mode server, and it would
+						// result in duplicate printing to the console
+						// (not that the local player should be using rcon
+						//  when they have the console)
+						if (host_client->netconnection && LHNETADDRESS_GetAddressType(&host_client->netconnection->peeraddress) != LHNETADDRESSTYPE_LOOP)
+							SV_ClientPrintf("%s", rcon_redirect_buffer);
+					}
+					else
+					{
+						// qw print command
+						dpsnprintf(response, sizeof(response), "\377\377\377\377n%s", rcon_redirect_buffer);
+						NetConn_WriteString(mysocket, response, peeraddress);
+					}
 				}
 				else
 				{
-					// qw print command
-					dpsnprintf(response, sizeof(response), "\377\377\377\377n%s", rcon_redirect_buffer);
-					NetConn_WriteString(mysocket, response, peeraddress);
+					Con_Printf("server denied rcon access to %s\n", host_client ? host_client->name : addressstring2);
 				}
 			}
 			return true;
@@ -2206,7 +2272,7 @@ static int NetConn_ServerParsePacket(lhnetsocket_t *mysocket, unsigned char *dat
 	// protocol
 	// (this protects more modern protocols against being used for
 	//  Quake packet flood Denial Of Service attacks)
-	if (length >= 5 && (i = BigLong(*((int *)data))) && (i & (~NETFLAG_LENGTH_MASK)) == (int)NETFLAG_CTL && (i & NETFLAG_LENGTH_MASK) == length && (sv.protocol == PROTOCOL_QUAKE || sv.protocol == PROTOCOL_QUAKEDP || sv.protocol == PROTOCOL_NEHAHRAMOVIE || sv.protocol == PROTOCOL_DARKPLACES1 || sv.protocol == PROTOCOL_DARKPLACES2 || sv.protocol == PROTOCOL_DARKPLACES3))
+	if (length >= 5 && (i = BigLong(*((int *)data))) && (i & (~NETFLAG_LENGTH_MASK)) == (int)NETFLAG_CTL && (i & NETFLAG_LENGTH_MASK) == length && (sv.protocol == PROTOCOL_QUAKE || sv.protocol == PROTOCOL_QUAKEDP || sv.protocol == PROTOCOL_NEHAHRAMOVIE || sv.protocol == PROTOCOL_NEHAHRABJP || sv.protocol == PROTOCOL_NEHAHRABJP2 || sv.protocol == PROTOCOL_NEHAHRABJP3 || sv.protocol == PROTOCOL_DARKPLACES1 || sv.protocol == PROTOCOL_DARKPLACES2 || sv.protocol == PROTOCOL_DARKPLACES3))
 	{
 		int c;
 		int protocolnumber;
@@ -2592,7 +2658,7 @@ static void Net_Heartbeat_f(void)
 void PrintStats(netconn_t *conn)
 {
 	if ((cls.state == ca_connected && cls.protocol == PROTOCOL_QUAKEWORLD) || (sv.active && sv.protocol == PROTOCOL_QUAKEWORLD))
-		Con_Printf("address=%21s canSend=%u sendSeq=%6u recvSeq=%6u\n", conn->address, !conn->sendMessageLength, conn->qw.outgoing_sequence, conn->qw.incoming_sequence);
+		Con_Printf("address=%21s canSend=%u sendSeq=%6u recvSeq=%6u\n", conn->address, !conn->sendMessageLength, conn->outgoing_unreliable_sequence, conn->qw.incoming_sequence);
 	else
 		Con_Printf("address=%21s canSend=%u sendSeq=%6u recvSeq=%6u\n", conn->address, !conn->sendMessageLength, conn->nq.sendSequence, conn->nq.receiveSequence);
 }
@@ -2614,6 +2680,16 @@ void Net_Stats_f(void)
 		PrintStats(conn);
 }
 
+void Net_Refresh_f(void)
+{
+	if (m_state != m_slist) {
+		Con_Print("Sending new requests to master servers\n");
+		ServerList_QueryList(false, true, false, true);
+		Con_Print("Listening for replies...\n");
+	} else
+		ServerList_QueryList(false, true, false, false);
+}
+
 void Net_Slist_f(void)
 {
 	ServerList_ResetMasks();
@@ -2621,11 +2697,10 @@ void Net_Slist_f(void)
 	serverlist_sortdescending = false;
     if (m_state != m_slist) {
 		Con_Print("Sending requests to master servers\n");
-		ServerList_QueryList(true, false);
-		serverlist_consoleoutput = true;
+		ServerList_QueryList(true, true, false, true);
 		Con_Print("Listening for replies...\n");
 	} else
-		ServerList_QueryList(true, false);
+		ServerList_QueryList(true, true, false, false);
 }
 
 void Net_SlistQW_f(void)
@@ -2635,30 +2710,11 @@ void Net_SlistQW_f(void)
 	serverlist_sortdescending = false;
     if (m_state != m_slist) {
 		Con_Print("Sending requests to master servers\n");
-		ServerList_QueryList(false, true);
+		ServerList_QueryList(true, false, true, true);
 		serverlist_consoleoutput = true;
 		Con_Print("Listening for replies...\n");
 	} else
-		ServerList_QueryList(false, true);
-}
-
-void Net_Banlist_Test_f(void)
-{
-	int i;
-	for(i = 1; i < Cmd_Argc(); ++i)
-	{
-		lhnetaddress_t addr;
-		const char *addrstr = Cmd_Argv(i);
-		if(LHNETADDRESS_FromString(&addr, addrstr, 0))
-		{
-			if(NetConn_IsBanned(&addr))
-				Con_Printf("%s is banned\n", addrstr);
-			else
-				Con_Printf("%s is free to join\n", addrstr);
-		}
-		else
-			Con_Printf("Invalid address %s\n", addrstr);
-	}
+		ServerList_QueryList(true, false, true, false);
 }
 
 void NetConn_Init(void)
@@ -2669,13 +2725,15 @@ void NetConn_Init(void)
 	Cmd_AddCommand("net_stats", Net_Stats_f, "print network statistics");
 	Cmd_AddCommand("net_slist", Net_Slist_f, "query dp master servers and print all server information");
 	Cmd_AddCommand("net_slistqw", Net_SlistQW_f, "query qw master servers and print all server information");
+	Cmd_AddCommand("net_refresh", Net_Refresh_f, "query dp master servers and refresh all server information");
 	Cmd_AddCommand("heartbeat", Net_Heartbeat_f, "send a heartbeat to the master server (updates your server information)");
-	Cmd_AddCommand("net_banlist_test", Net_Banlist_Test_f, "tests if a given IP is banned");
-	Cvar_RegisterVariable(&net_banlist);
+	Cvar_RegisterVariable(&rcon_restricted_password);
+	Cvar_RegisterVariable(&rcon_restricted_commands);
 	Cvar_RegisterVariable(&net_slist_queriespersecond);
 	Cvar_RegisterVariable(&net_slist_queriesperframe);
 	Cvar_RegisterVariable(&net_slist_timeout);
 	Cvar_RegisterVariable(&net_slist_maxtries);
+	Cvar_RegisterVariable(&net_slist_pause);
 	Cvar_RegisterVariable(&net_messagetimeout);
 	Cvar_RegisterVariable(&net_connecttimeout);
 	Cvar_RegisterVariable(&net_connectfloodblockingtimeout);
