@@ -266,6 +266,43 @@ void PRVM_StackTrace (void)
 }
 
 
+void PRVM_CallProfile ()
+{
+	mfunction_t *f, *best;
+	int i;
+	double max;
+	double sum;
+
+	Con_Printf( "%s Call Profile:\n", PRVM_NAME );
+
+	sum = 0;
+	do
+	{
+		max = 0;
+		best = NULL;
+		for (i=0 ; i<prog->progs->numfunctions ; i++)
+		{
+			f = &prog->functions[i];
+			if (max < f->totaltime)
+			{
+				max = f->totaltime;
+				best = f;
+			}
+		}
+		if (best)
+		{
+			sum += best->totaltime;
+			Con_Printf("%9.4f %s\n", best->totaltime, PRVM_GetString(best->s_name));
+			best->totaltime = 0;
+		}
+	} while (best);
+
+	Con_Printf("Total time since last profile reset: %9.4f\n", Sys_DoubleTime() - prog->starttime);
+	Con_Printf("       - used by QC code of this VM: %9.4f\n", sum);
+
+	prog->starttime = Sys_DoubleTime();
+}
+
 void PRVM_Profile (int maxfunctions, int mininstructions)
 {
 	mfunction_t *f, *best;
@@ -305,6 +342,29 @@ void PRVM_Profile (int maxfunctions, int mininstructions)
 	} while (best);
 }
 
+/*
+============
+PRVM_CallProfile_f
+
+============
+*/
+void PRVM_CallProfile_f (void)
+{
+	if (Cmd_Argc() != 2)
+	{
+		Con_Print("prvm_callprofile <program name>\n");
+		return;
+	}
+
+	PRVM_Begin;
+	if(!PRVM_SetProgFromString(Cmd_Argv(1)))
+		return;
+
+	PRVM_CallProfile();
+
+	PRVM_End;
+}
+
 /*
 ============
 PRVM_Profile_f
@@ -363,11 +423,14 @@ void PRVM_PrintState(void)
 	PRVM_StackTrace ();
 }
 
-void PRVM_Crash()
+extern sizebuf_t vm_tempstringsbuf;
+void PRVM_Crash(void)
 {
 	if (prog == NULL)
 		return;
 
+	prog->funcoffsets.SV_Shutdown = 0; // don't call SV_Shutdown on crash
+
 	if( prog->depth > 0 )
 	{
 		Con_Printf("QuakeC crash report for %s:\n", PRVM_NAME);
@@ -378,6 +441,9 @@ void PRVM_Crash()
 	prog->depth = 0;
 	prog->localstack_used = 0;
 
+	// delete all tempstrings (FIXME: is this safe in VM->engine->VM recursion?)
+	vm_tempstringsbuf.cursize = 0;
+
 	// reset the prog pointer
 	prog = NULL;
 }
@@ -481,11 +547,12 @@ PRVM_ExecuteProgram
 #define OPA ((prvm_eval_t *)&prog->globals.generic[(unsigned short) st->a])
 #define OPB ((prvm_eval_t *)&prog->globals.generic[(unsigned short) st->b])
 #define OPC ((prvm_eval_t *)&prog->globals.generic[(unsigned short) st->c])
+#ifdef PRVM_BOUNDSCHECK_CVAR
 extern cvar_t prvm_boundscheck;
+#endif
 extern cvar_t prvm_traceqc;
 extern cvar_t prvm_statementprofiling;
-extern int		PRVM_ED_FindFieldOffset (const char *field);
-extern ddef_t*	PRVM_ED_FindGlobal(const char *name);
+extern sizebuf_t vm_tempstringsbuf;
 void PRVM_ExecuteProgram (func_t fnum, const char *errormessage)
 {
 	dstatement_t	*st, *startst;
@@ -493,16 +560,23 @@ void PRVM_ExecuteProgram (func_t fnum, const char *errormessage)
 	prvm_edict_t	*ed;
 	prvm_eval_t	*ptr;
 	int		jumpcount, cachedpr_trace, exitdepth;
+	int		restorevm_tempstringsbuf_cursize;
+	double  calltime;
+
+	calltime = Sys_DoubleTime();
 
 	if (!fnum || fnum >= (unsigned int)prog->progs->numfunctions)
 	{
-		if (prog->self && PRVM_G_INT(prog->self->ofs))
-			PRVM_ED_Print(PRVM_PROG_TO_EDICT(PRVM_G_INT(prog->self->ofs)));
+		if (prog->globaloffsets.self >= 0 && PRVM_GLOBALFIELDVALUE(prog->globaloffsets.self)->edict)
+			PRVM_ED_Print(PRVM_PROG_TO_EDICT(PRVM_GLOBALFIELDVALUE(prog->globaloffsets.self)->edict), NULL);
 		PRVM_ERROR ("PRVM_ExecuteProgram: %s", errormessage);
 	}
 
 	f = &prog->functions[fnum];
 
+	// after executing this function, delete all tempstrings it created
+	restorevm_tempstringsbuf_cursize = vm_tempstringsbuf.cursize;
+
 	prog->trace = prvm_traceqc.integer;
 
 	// we know we're done when pr_depth drops to this
@@ -524,7 +598,9 @@ chooseexecprogram:
 	if (prvm_statementprofiling.integer)
 	{
 #define PRVMSTATEMENTPROFILING 1
+#ifdef PRVM_BOUNDSCHECK_CVAR
 		if (prvm_boundscheck.integer)
+#endif
 		{
 #define PRVMBOUNDSCHECK 1
 			if (prog->trace)
@@ -539,6 +615,7 @@ chooseexecprogram:
 			}
 #undef PRVMBOUNDSCHECK
 		}
+#ifdef PRVM_BOUNDSCHECK_CVAR
 		else
 		{
 			if (prog->trace)
@@ -552,11 +629,14 @@ chooseexecprogram:
 #include "prvm_execprogram.h"
 			}
 		}
+#endif
 #undef PRVMSTATEMENTPROFILING
 	}
 	else
 	{
+#ifdef PRVM_BOUNDSCHECK_CVAR
 		if (prvm_boundscheck.integer)
+#endif
 		{
 #define PRVMBOUNDSCHECK 1
 			if (prog->trace)
@@ -571,6 +651,7 @@ chooseexecprogram:
 			}
 #undef PRVMBOUNDSCHECK
 		}
+#ifdef PRVM_BOUNDSCHECK_CVAR
 		else
 		{
 			if (prog->trace)
@@ -584,5 +665,16 @@ chooseexecprogram:
 #include "prvm_execprogram.h"
 			}
 		}
+#endif
 	}
+
+cleanup:
+	if (developer.integer >= 200 && vm_tempstringsbuf.cursize > restorevm_tempstringsbuf_cursize)
+		Con_Printf("PRVM_ExecuteProgram: %s used %i bytes of tempstrings\n", PRVM_GetString(prog->functions[fnum].s_name), vm_tempstringsbuf.cursize - restorevm_tempstringsbuf_cursize);
+	// delete tempstrings created by this function
+	vm_tempstringsbuf.cursize = restorevm_tempstringsbuf_cursize;
+
+	prog->functions[fnum].totaltime += (Sys_DoubleTime() - calltime);
+
+	SV_FlushBroadcastMessages();
 }
