@@ -25,7 +25,7 @@ void r_lightningbeams_start(void)
 
 void r_lightningbeams_setupqmbtexture(void)
 {
-	r_lightningbeamqmbtexture = loadtextureimage(r_lightningbeamtexturepool, "textures/particles/lightning.pcx", 0, 0, false, TEXF_ALPHA | TEXF_PRECACHE);
+	r_lightningbeamqmbtexture = loadtextureimage(r_lightningbeamtexturepool, "textures/particles/lightning.pcx", false, TEXF_ALPHA | TEXF_PRECACHE, false);
 	if (r_lightningbeamqmbtexture == NULL)
 		Cvar_SetValueQuick(&r_lightningbeam_qmbtexture, false);
 }
@@ -51,63 +51,63 @@ void r_lightningbeams_setuptexture(void)
 
 	for (imagenumber = 0, maxpathstrength = 0.0339476;maxpathstrength < 0.5;imagenumber++, maxpathstrength += 0.01)
 	{
-	for (i = 0;i < PATHPOINTS;i++)
-	{
-		path[i].x = lhrandom(0, 1);
-		path[i].y = lhrandom(0.2, 0.8);
-		path[i].strength = lhrandom(0, 1);
-	}
-	for (i = 0;i < PATHPOINTS;i++)
-	{
-		for (j = i + 1;j < PATHPOINTS;j++)
+		for (i = 0;i < PATHPOINTS;i++)
 		{
-			if (path[j].x < path[i].x)
+			path[i].x = lhrandom(0, 1);
+			path[i].y = lhrandom(0.2, 0.8);
+			path[i].strength = lhrandom(0, 1);
+		}
+		for (i = 0;i < PATHPOINTS;i++)
+		{
+			for (j = i + 1;j < PATHPOINTS;j++)
 			{
-				temppath = path[j];
-				path[j] = path[i];
-				path[i] = temppath;
+				if (path[j].x < path[i].x)
+				{
+					temppath = path[j];
+					path[j] = path[i];
+					path[i] = temppath;
+				}
 			}
 		}
-	}
-	particlex = path[0].x;
-	particley = path[0].y;
-	particlexv = lhrandom(0, 0.02);
-	particlexv = lhrandom(-0.02, 0.02);
-	memset(image, 0, BEAMWIDTH * BEAMHEIGHT * sizeof(int));
-	for (i = 0;i < 65536;i++)
-	{
-		for (nearestpathindex = 0;nearestpathindex < PATHPOINTS;nearestpathindex++)
-			if (path[nearestpathindex].x > particlex)
-				break;
-		nearestpathindex %= PATHPOINTS;
-		dx = path[nearestpathindex].x + lhrandom(-0.01, 0.01);dx = bound(0, dx, 1) - particlex;if (dx < 0) dx += 1;
-		dy = path[nearestpathindex].y + lhrandom(-0.01, 0.01);dy = bound(0, dy, 1) - particley;
-		s = path[nearestpathindex].strength / sqrt(dx*dx+dy*dy);
-		particlexv = particlexv /* (1 - lhrandom(0.08, 0.12))*/ + dx * s;
-		particleyv = particleyv /* (1 - lhrandom(0.08, 0.12))*/ + dy * s;
-		particlex += particlexv * maxpathstrength;particlex -= (int) particlex;
-		particley += particleyv * maxpathstrength;particley = bound(0, particley, 1);
-		px = particlex * BEAMWIDTH;
-		py = particley * BEAMHEIGHT;
-		if (px >= 0 && py >= 0 && px < BEAMWIDTH && py < BEAMHEIGHT)
-			image[py*BEAMWIDTH+px] += 16;
-	}
-
-	for (py = 0;py < BEAMHEIGHT;py++)
-	{
-		for (px = 0;px < BEAMWIDTH;px++)
+		particlex = path[0].x;
+		particley = path[0].y;
+		particlexv = lhrandom(0, 0.02);
+		particlexv = lhrandom(-0.02, 0.02);
+		memset(image, 0, BEAMWIDTH * BEAMHEIGHT * sizeof(int));
+		for (i = 0;i < 65536;i++)
 		{
-			pixels[(py*BEAMWIDTH+px)*4+0] = bound(0, image[py*BEAMWIDTH+px] * 1.0f, 255.0f);
-			pixels[(py*BEAMWIDTH+px)*4+1] = bound(0, image[py*BEAMWIDTH+px] * 1.0f, 255.0f);
-			pixels[(py*BEAMWIDTH+px)*4+2] = bound(0, image[py*BEAMWIDTH+px] * 1.0f, 255.0f);
-			pixels[(py*BEAMWIDTH+px)*4+3] = 255;
+			for (nearestpathindex = 0;nearestpathindex < PATHPOINTS;nearestpathindex++)
+				if (path[nearestpathindex].x > particlex)
+					break;
+			nearestpathindex %= PATHPOINTS;
+			dx = path[nearestpathindex].x + lhrandom(-0.01, 0.01);dx = bound(0, dx, 1) - particlex;if (dx < 0) dx += 1;
+			dy = path[nearestpathindex].y + lhrandom(-0.01, 0.01);dy = bound(0, dy, 1) - particley;
+			s = path[nearestpathindex].strength / sqrt(dx*dx+dy*dy);
+			particlexv = particlexv /* (1 - lhrandom(0.08, 0.12))*/ + dx * s;
+			particleyv = particleyv /* (1 - lhrandom(0.08, 0.12))*/ + dy * s;
+			particlex += particlexv * maxpathstrength;particlex -= (int) particlex;
+			particley += particleyv * maxpathstrength;particley = bound(0, particley, 1);
+			px = particlex * BEAMWIDTH;
+			py = particley * BEAMHEIGHT;
+			if (px >= 0 && py >= 0 && px < BEAMWIDTH && py < BEAMHEIGHT)
+				image[py*BEAMWIDTH+px] += 16;
 		}
+
+		for (py = 0;py < BEAMHEIGHT;py++)
+		{
+			for (px = 0;px < BEAMWIDTH;px++)
+			{
+				pixels[(py*BEAMWIDTH+px)*4+2] = bound(0, image[py*BEAMWIDTH+px] * 1.0f, 255.0f);
+				pixels[(py*BEAMWIDTH+px)*4+1] = bound(0, image[py*BEAMWIDTH+px] * 1.0f, 255.0f);
+				pixels[(py*BEAMWIDTH+px)*4+0] = bound(0, image[py*BEAMWIDTH+px] * 1.0f, 255.0f);
+				pixels[(py*BEAMWIDTH+px)*4+3] = 255;
+			}
+		}
+
+		Image_WriteTGABGRA(va("lightningbeam%i.tga", imagenumber), BEAMWIDTH, BEAMHEIGHT, pixels);
 	}
 
-	Image_WriteTGARGBA(va("lightningbeam%i.tga", imagenumber), BEAMWIDTH, BEAMHEIGHT, pixels);
-	}
-
-	r_lightningbeamtexture = R_LoadTexture2D(r_lightningbeamtexturepool, "lightningbeam", BEAMWIDTH, BEAMHEIGHT, pixels, TEXTYPE_RGBA, TEXF_PRECACHE, NULL);
+	r_lightningbeamtexture = R_LoadTexture2D(r_lightningbeamtexturepool, "lightningbeam", BEAMWIDTH, BEAMHEIGHT, pixels, TEXTYPE_BGRA, TEXF_PRECACHE, NULL);
 
 	Mem_Free(pixels);
 	Mem_Free(image);
@@ -138,14 +138,14 @@ void r_lightningbeams_setuptexture(void)
 			r = intensity * 1.0f;
 			g = intensity * 1.0f;
 			b = intensity * 1.0f;
-			data[(y * BEAMWIDTH + x) * 4 + 0] = (unsigned char)(bound(0, r, 1) * 255.0f);
+			data[(y * BEAMWIDTH + x) * 4 + 2] = (unsigned char)(bound(0, r, 1) * 255.0f);
 			data[(y * BEAMWIDTH + x) * 4 + 1] = (unsigned char)(bound(0, g, 1) * 255.0f);
-			data[(y * BEAMWIDTH + x) * 4 + 2] = (unsigned char)(bound(0, b, 1) * 255.0f);
+			data[(y * BEAMWIDTH + x) * 4 + 0] = (unsigned char)(bound(0, b, 1) * 255.0f);
 			data[(y * BEAMWIDTH + x) * 4 + 3] = (unsigned char)255;
 		}
 	}
 
-	r_lightningbeamtexture = R_LoadTexture2D(r_lightningbeamtexturepool, "lightningbeam", BEAMWIDTH, BEAMHEIGHT, data, TEXTYPE_RGBA, TEXF_PRECACHE, NULL);
+	r_lightningbeamtexture = R_LoadTexture2D(r_lightningbeamtexturepool, "lightningbeam", BEAMWIDTH, BEAMHEIGHT, data, TEXTYPE_BGRA, TEXF_PRECACHE, NULL);
 	Mem_Free(noise1);
 	Mem_Free(noise2);
 	Mem_Free(data);
@@ -242,24 +242,25 @@ void R_DrawLightningBeam_TransparentCallback(const entity_render_t *ent, const r
 	GL_DepthRange(0, 1);
 	GL_PolygonOffset(r_refdef.polygonfactor, r_refdef.polygonoffset);
 	GL_DepthTest(true);
+	GL_CullFace(GL_NONE);
 	if (r_lightningbeam_qmbtexture.integer && r_lightningbeamqmbtexture == NULL)
 		r_lightningbeams_setupqmbtexture();
 	if (!r_lightningbeam_qmbtexture.integer && r_lightningbeamtexture == NULL)
 		r_lightningbeams_setuptexture();
 
 	R_Mesh_VertexPointer(vertex3f, 0, 0);
-	// FIXME: fixed function path can't properly handle r_view.colorscale > 1
+	R_SetupGenericShader(true);
+	// FIXME: fixed function path can't properly handle r_refdef.view.colorscale > 1
 	if (r_refdef.fogenabled)
 	{
 		// per vertex colors if fog is used
 		R_Mesh_ColorPointer(color4f, 0, 0);
-		R_FogLightningBeam_Vertex3f_Color4f(vertex3f, color4f, 12, r_lightningbeam_color_red.value * r_view.colorscale, r_lightningbeam_color_green.value * r_view.colorscale, r_lightningbeam_color_blue.value * r_view.colorscale, 1);
 	}
 	else
 	{
 		// solid color if fog is not used
 		R_Mesh_ColorPointer(NULL, 0, 0);
-		GL_Color(r_lightningbeam_color_red.value * r_view.colorscale, r_lightningbeam_color_green.value * r_view.colorscale, r_lightningbeam_color_blue.value * r_view.colorscale, 1);
+		GL_Color(r_lightningbeam_color_red.value * r_refdef.view.colorscale, r_lightningbeam_color_green.value * r_refdef.view.colorscale, r_lightningbeam_color_blue.value * r_refdef.view.colorscale, 1);
 	}
 	memset(&m, 0, sizeof(m));
 	if (r_lightningbeam_qmbtexture.integer)
@@ -289,7 +290,7 @@ void R_DrawLightningBeam_TransparentCallback(const entity_render_t *ent, const r
 
 		// calculate up vector such that it points toward viewer, and rotates around the beamdir
 		// get direction from start of beam to viewer
-		VectorSubtract(r_view.origin, start, up);
+		VectorSubtract(r_refdef.view.origin, start, up);
 		// remove the portion of the vector that moves along the beam
 		// (this leaves only a vector pointing directly away from the beam)
 		t1 = -DotProduct(up, beamdir);
@@ -301,7 +302,7 @@ void R_DrawLightningBeam_TransparentCallback(const entity_render_t *ent, const r
 		VectorNormalize(up);
 
 		// calculate T coordinate scrolling (start and end texcoord along the beam)
-		t1 = r_refdef.time * -r_lightningbeam_scroll.value;// + beamrepeatscale * DotProduct(start, beamdir);
+		t1 = r_refdef.scene.time * -r_lightningbeam_scroll.value;// + beamrepeatscale * DotProduct(start, beamdir);
 		t1 = t1 - (int) t1;
 		t2 = t1 + beamrepeatscale * length;
 
@@ -362,7 +363,7 @@ void R_DrawLightningBeams(void)
 			CL_Beam_CalculatePositions(b, start, end);
 			// calculate the nearest point on the line (beam) for depth sorting
 			VectorSubtract(end, start, dir);
-			dist = (DotProduct(r_view.origin, dir) - DotProduct(start, dir)) / (DotProduct(end, dir) - DotProduct(start, dir));
+			dist = (DotProduct(r_refdef.view.origin, dir) - DotProduct(start, dir)) / (DotProduct(end, dir) - DotProduct(start, dir));
 			dist = bound(0, dist, 1);
 			VectorLerp(start, dist, end, org);
 			// now we have the nearest point on the line, so sort with it
