@@ -16,11 +16,11 @@ You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
+#include "quakedef.h"
 
 #include <math.h>
 #include <SDL.h>
 
-#include "quakedef.h"
 #include "snd_main.h"
 
 
@@ -39,6 +39,12 @@ static void Buffer_Callback (void *userdata, Uint8 *stream, int len)
 
 	RequestedFrames = (unsigned int)len / factor;
 
+	if (snd_usethreadedmixing)
+	{
+		S_MixToBuffer(stream, RequestedFrames);
+		return;
+	}
+
 	// Transfert up to a chunk of samples from snd_renderbuffer to stream
 	MaxFrames = snd_renderbuffer->endframe - snd_renderbuffer->startframe;
 	if (MaxFrames > RequestedFrames)
@@ -63,7 +69,7 @@ static void Buffer_Callback (void *userdata, Uint8 *stream, int len)
 	snd_renderbuffer->startframe += FrameCount;
 
 	if (FrameCount < RequestedFrames && developer.integer >= 1000 && vid_activewindow)
-		Con_DPrintf("SDL sound: %u sample frames missing\n", RequestedFrames - FrameCount);
+		Con_Printf("SDL sound: %u sample frames missing\n", RequestedFrames - FrameCount);
 
 	sdlaudiotime += RequestedFrames;
 }
@@ -83,7 +89,9 @@ qboolean SndSys_Init (const snd_format_t* requested, snd_format_t* suggested)
 	SDL_AudioSpec wantspec;
 	SDL_AudioSpec obtainspec;
 
-	Con_Print ("SndSys_Init: using the SDL module\n");
+	snd_threaded = false;
+
+	Con_DPrint ("SndSys_Init: using the SDL module\n");
 
 	// Init the SDL Audio subsystem
 	if( SDL_InitSubSystem( SDL_INIT_AUDIO ) ) {
@@ -101,7 +109,7 @@ qboolean SndSys_Init (const snd_format_t* requested, snd_format_t* suggested)
 	wantspec.channels = requested->channels;
 	wantspec.samples = CeilPowerOf2(buffersize);  // needs to be a power of 2 on some platforms.
 
-	Con_DPrintf("Wanted audio Specification:\n"
+	Con_Printf("Wanted audio Specification:\n"
 				"\tChannels  : %i\n"
 				"\tFormat    : 0x%X\n"
 				"\tFrequency : %i\n"
@@ -114,7 +122,7 @@ qboolean SndSys_Init (const snd_format_t* requested, snd_format_t* suggested)
 		return false;
 	}
 
-	Con_DPrintf("Obtained audio specification:\n"
+	Con_Printf("Obtained audio specification:\n"
 				"\tChannels  : %i\n"
 				"\tFormat    : 0x%X\n"
 				"\tFrequency : %i\n"
@@ -140,6 +148,8 @@ qboolean SndSys_Init (const snd_format_t* requested, snd_format_t* suggested)
 		return false;
 	}
 
+	snd_threaded = true;
+
 	snd_renderbuffer = Snd_CreateRingBuffer(requested, 0, NULL);
 	if (snd_channellayout.integer == SND_CHANNELLAYOUT_AUTO)
 	{
