@@ -223,10 +223,10 @@ static wavinfo_t GetWavinfo (char *name, unsigned char *wav, int wavlength)
 WAV_FetchSound
 ====================
 */
-static const snd_buffer_t* WAV_FetchSound (channel_t* ch, unsigned int *start, unsigned int nbsampleframes)
+static const snd_buffer_t* WAV_FetchSound (void *sfxfetcher, void **chfetcherpointer, unsigned int *start, unsigned int nbsampleframes)
 {
 	*start = 0;
-	return (snd_buffer_t *)ch->sfx->fetcher_data;
+	return (snd_buffer_t *)sfxfetcher;
 }
 
 /*
@@ -234,16 +234,11 @@ static const snd_buffer_t* WAV_FetchSound (channel_t* ch, unsigned int *start, u
 WAV_FreeSfx
 ====================
 */
-static void WAV_FreeSfx (sfx_t* sfx)
+static void WAV_FreeSfx (void *sfxfetcherdata)
 {
-	snd_buffer_t* sb = (snd_buffer_t *)sfx->fetcher_data;
-
+	snd_buffer_t* sb = (snd_buffer_t *)sfxfetcherdata;
 	// Free the sound buffer
-	sfx->memsize -= (sb->maxframes * sb->format.channels * sb->format.width) + sizeof (*sb) - sizeof (sb->samples);
 	Mem_Free(sb);
-
-	sfx->fetcher_data = NULL;
-	sfx->fetcher = NULL;
 }
 
 /*
@@ -289,7 +284,8 @@ qboolean S_LoadWavFile (const char *filename, sfx_t *sfx)
 		return false;
 	}
 
-	Con_DPrintf ("Loading WAV file \"%s\"\n", filename);
+	if (developer_loading.integer >= 2)
+		Con_Printf ("Loading WAV file \"%s\"\n", filename);
 
 	info = GetWavinfo (sfx->name, data, (int)filesize);
 	if (info.channels < 1 || info.channels > 2)  // Stereo sounds are allowed (intended for music)
@@ -334,7 +330,7 @@ qboolean S_LoadWavFile (const char *filename, sfx_t *sfx)
 	if (info.loopstart < 0)
 		sfx->loopstart = sfx->total_length;
 	else
-		sfx->loopstart = (double)info.loopstart * (double)snd_renderbuffer->format.speed / (double)sb->format.speed;
+		sfx->loopstart = (double)info.loopstart * (double)sb->format.speed / (double)info.rate;
 	sfx->loopstart = min(sfx->loopstart, sfx->total_length);
 	sfx->flags &= ~SFXFLAG_STREAMED;
 
