@@ -20,21 +20,11 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 // sv_user.c -- server code for moving users
 
 #include "quakedef.h"
-
-cvar_t sv_edgefriction = {0, "edgefriction", "2", "how much you slow down when nearing a ledge you might fall off"};
-cvar_t sv_idealpitchscale = {0, "sv_idealpitchscale","0.8", "how much to look up/down slopes and stairs when not using freelook"};
-cvar_t sv_maxspeed = {CVAR_NOTIFY, "sv_maxspeed", "320", "maximum speed a player can accelerate to when on ground (can be exceeded by tricks)"};
-cvar_t sv_maxairspeed = {0, "sv_maxairspeed", "30", "maximum speed a player can accelerate to when airborn (note that it is possible to completely stop by moving the opposite direction)"};
-cvar_t sv_accelerate = {0, "sv_accelerate", "10", "rate at which a player accelerates to sv_maxspeed"};
-cvar_t sv_airaccelerate = {0, "sv_airaccelerate", "-1", "rate at which a player accelerates to sv_maxairspeed while in the air, if less than 0 the sv_accelerate variable is used instead"};
-cvar_t sv_wateraccelerate = {0, "sv_wateraccelerate", "-1", "rate at which a player accelerates to sv_maxspeed while in the air, if less than 0 the sv_accelerate variable is used instead"};
-cvar_t sv_clmovement_enable = {0, "sv_clmovement_enable", "1", "whether to allow clients to use cl_movement prediction, which can cause choppy movement on the server which may annoy other players"};
-cvar_t sv_clmovement_minping = {0, "sv_clmovement_minping", "100", "if client ping is below this time in milliseconds, then their ability to use cl_movement prediction is disabled for a while (as they don't need it)"};
-cvar_t sv_clmovement_minping_disabletime = {0, "sv_clmovement_minping_disabletime", "1000", "when client falls below minping, disable their prediction for this many milliseconds (should be at least 1000 or else their prediction may turn on/off frequently)"};
-cvar_t sv_clmovement_waitforinput = {0, "sv_clmovement_waitforinput", "2", "when a client does not send input for this many frames, force them to move anyway (unlike QuakeWorld)"};
+#include "sv_demo.h"
+#define DEBUGMOVES 0
 
 static usercmd_t cmd;
-
+extern cvar_t sv_autodemo_perclient;
 
 /*
 ===============
@@ -68,7 +58,7 @@ void SV_SetIdealPitch (void)
 		bottom[1] = top[1];
 		bottom[2] = top[2] - 160;
 
-		tr = SV_Move (top, vec3_origin, vec3_origin, bottom, MOVE_NOMONSTERS, host_client->edict);
+		tr = SV_Move (top, vec3_origin, vec3_origin, bottom, MOVE_NOMONSTERS, host_client->edict, SUPERCONTENTS_SOLID);
 		// if looking at a wall, leave ideal the way is was
 		if (tr.startsolid)
 			return;
@@ -134,7 +124,7 @@ void SV_UserFriction (void)
 	start[2] = host_client->edict->fields.server->origin[2] + host_client->edict->fields.server->mins[2];
 	stop[2] = start[2] - 34;
 
-	trace = SV_Move (start, vec3_origin, vec3_origin, stop, MOVE_NOMONSTERS, host_client->edict);
+	trace = SV_Move (start, vec3_origin, vec3_origin, stop, MOVE_NOMONSTERS, host_client->edict, SV_GenericHitSuperContentsMask(host_client->edict));
 
 	if (trace.fraction == 1.0)
 		friction = sv_friction.value*sv_edgefriction.value;
@@ -208,7 +198,7 @@ void DropPunchAngle (void)
 		len = 0;
 	VectorScale (host_client->edict->fields.server->punchangle, len, host_client->edict->fields.server->punchangle);
 
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_punchvector)))
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.punchvector)))
 	{
 		len = VectorNormalizeLength (val->vector);
 
@@ -277,7 +267,7 @@ void SV_WaterMove (void)
 	speed = VectorLength(host_client->edict->fields.server->velocity);
 	if (speed)
 	{
-		newspeed = speed - sv.frametime * speed * sv_waterfriction.value;
+		newspeed = speed - sv.frametime * speed * (sv_waterfriction.value < 0 ? sv_friction.value : sv_waterfriction.value);
 		if (newspeed < 0)
 			newspeed = 0;
 		temp = newspeed/speed;
@@ -359,7 +349,7 @@ void SV_AirMove (void)
 		// noclip
 		VectorCopy (wishvel, host_client->edict->fields.server->velocity);
 	}
-	else if (onground && (!sv_gameplayfix_qwplayerphysics.integer || !(host_client->edict->fields.server->button2 || !((int)host_client->edict->fields.server->flags & FL_JUMPRELEASED))))
+	else if (onground)
 	{
 		SV_UserFriction ();
 		SV_Accelerate ();
@@ -383,6 +373,22 @@ void SV_ClientThink (void)
 {
 	vec3_t v_angle;
 
+	//Con_Printf("clientthink for %ims\n", (int) (sv.frametime * 1000));
+
+	SV_ApplyClientMove();
+	// make sure the velocity is sane (not a NaN)
+	SV_CheckVelocity(host_client->edict);
+
+	// LordHavoc: QuakeC replacement for SV_ClientThink (player movement)
+	if (prog->funcoffsets.SV_PlayerPhysics && sv_playerphysicsqc.integer)
+	{
+		prog->globals.server->time = sv.time;
+		prog->globals.server->self = PRVM_EDICT_TO_PROG(host_client->edict);
+		PRVM_ExecuteProgram (prog->funcoffsets.SV_PlayerPhysics, "QC function SV_PlayerPhysics is missing");
+		SV_CheckVelocity(host_client->edict);
+		return;
+	}
+
 	if (host_client->edict->fields.server->movetype == MOVETYPE_NONE)
 		return;
 
@@ -409,6 +415,7 @@ void SV_ClientThink (void)
 	if ( (int)host_client->edict->fields.server->flags & FL_WATERJUMP )
 	{
 		SV_WaterJump ();
+		SV_CheckVelocity(host_client->edict);
 		return;
 	}
 
@@ -426,10 +433,12 @@ void SV_ClientThink (void)
 	if ((host_client->edict->fields.server->waterlevel >= 2) && (host_client->edict->fields.server->movetype != MOVETYPE_NOCLIP))
 	{
 		SV_WaterMove ();
+		SV_CheckVelocity(host_client->edict);
 		return;
 	}
 
 	SV_AirMove ();
+	SV_CheckVelocity(host_client->edict);
 }
 
 /*
@@ -437,45 +446,36 @@ void SV_ClientThink (void)
 SV_ReadClientMove
 ===================
 */
-qboolean SV_ReadClientMove (void)
+int sv_numreadmoves = 0;
+usercmd_t sv_readmoves[CL_MAX_USERCMDS];
+void SV_ReadClientMove (void)
 {
-	qboolean kickplayer = false;
 	int i;
-	double oldmovetime;
-#ifdef NUM_PING_TIMES
-	double total;
-#endif
-	usercmd_t *move = &host_client->cmd;
+	usercmd_t newmove;
+	usercmd_t *move = &newmove;
 
-	oldmovetime = move->time;
-
-	// if this move has been applied, clear it, and start accumulating new data
-	if (move->applied)
-		memset(move, 0, sizeof(*move));
+	memset(move, 0, sizeof(*move));
 
 	if (msg_badread) Con_Printf("SV_ReadClientMessage: badread at %s:%i\n", __FILE__, __LINE__);
 
 	// read ping time
-	if (sv.protocol != PROTOCOL_QUAKE && sv.protocol != PROTOCOL_QUAKEDP && sv.protocol != PROTOCOL_NEHAHRAMOVIE && sv.protocol != PROTOCOL_DARKPLACES1 && sv.protocol != PROTOCOL_DARKPLACES2 && sv.protocol != PROTOCOL_DARKPLACES3 && sv.protocol != PROTOCOL_DARKPLACES4 && sv.protocol != PROTOCOL_DARKPLACES5 && sv.protocol != PROTOCOL_DARKPLACES6)
+	if (sv.protocol != PROTOCOL_QUAKE && sv.protocol != PROTOCOL_QUAKEDP && sv.protocol != PROTOCOL_NEHAHRAMOVIE && sv.protocol != PROTOCOL_NEHAHRABJP && sv.protocol != PROTOCOL_NEHAHRABJP2 && sv.protocol != PROTOCOL_NEHAHRABJP3 && sv.protocol != PROTOCOL_DARKPLACES1 && sv.protocol != PROTOCOL_DARKPLACES2 && sv.protocol != PROTOCOL_DARKPLACES3 && sv.protocol != PROTOCOL_DARKPLACES4 && sv.protocol != PROTOCOL_DARKPLACES5 && sv.protocol != PROTOCOL_DARKPLACES6)
 		move->sequence = MSG_ReadLong ();
 	move->time = MSG_ReadFloat ();
 	if (msg_badread) Con_Printf("SV_ReadClientMessage: badread at %s:%i\n", __FILE__, __LINE__);
 	move->receivetime = (float)sv.time;
 
-	// calculate average ping time
-	host_client->ping = move->receivetime - move->time;
-#ifdef NUM_PING_TIMES
-	host_client->ping_times[host_client->num_pings % NUM_PING_TIMES] = move->receivetime - move->time;
-	host_client->num_pings++;
-	for (i=0, total = 0;i < NUM_PING_TIMES;i++)
-		total += host_client->ping_times[i];
-	host_client->ping = total / NUM_PING_TIMES;
+#if DEBUGMOVES
+	Con_Printf("%s move%i #%i %ims (%ims) %i %i '%i %i %i' '%i %i %i'\n", move->time > move->receivetime ? "^3read future" : "^4read normal", sv_numreadmoves + 1, move->sequence, (int)floor((move->time - host_client->cmd.time) * 1000.0 + 0.5), (int)floor(move->time * 1000.0 + 0.5), move->impulse, move->buttons, (int)move->viewangles[0], (int)move->viewangles[1], (int)move->viewangles[2], (int)move->forwardmove, (int)move->sidemove, (int)move->upmove);
 #endif
+	// limit reported time to current time
+	// (incase the client is trying to cheat)
+	move->time = min(move->time, move->receivetime + sv.frametime);
 
 	// read current angles
 	for (i = 0;i < 3;i++)
 	{
-		if (sv.protocol == PROTOCOL_QUAKE || sv.protocol == PROTOCOL_QUAKEDP || sv.protocol == PROTOCOL_NEHAHRAMOVIE)
+		if (sv.protocol == PROTOCOL_QUAKE || sv.protocol == PROTOCOL_QUAKEDP || sv.protocol == PROTOCOL_NEHAHRAMOVIE || sv.protocol == PROTOCOL_NEHAHRABJP || sv.protocol == PROTOCOL_NEHAHRABJP2 || sv.protocol == PROTOCOL_NEHAHRABJP3)
 			move->viewangles[i] = MSG_ReadAngle8i();
 		else if (sv.protocol == PROTOCOL_DARKPLACES1)
 			move->viewangles[i] = MSG_ReadAngle16i();
@@ -495,20 +495,18 @@ qboolean SV_ReadClientMove (void)
 	// read buttons
 	// be sure to bitwise OR them into the move->buttons because we want to
 	// accumulate button presses from multiple packets per actual move
-	if (sv.protocol == PROTOCOL_QUAKE || sv.protocol == PROTOCOL_QUAKEDP || sv.protocol == PROTOCOL_NEHAHRAMOVIE || sv.protocol == PROTOCOL_DARKPLACES1 || sv.protocol == PROTOCOL_DARKPLACES2 || sv.protocol == PROTOCOL_DARKPLACES3 || sv.protocol == PROTOCOL_DARKPLACES4 || sv.protocol == PROTOCOL_DARKPLACES5)
-		move->buttons |= MSG_ReadByte ();
+	if (sv.protocol == PROTOCOL_QUAKE || sv.protocol == PROTOCOL_QUAKEDP || sv.protocol == PROTOCOL_NEHAHRAMOVIE || sv.protocol == PROTOCOL_NEHAHRABJP || sv.protocol == PROTOCOL_NEHAHRABJP2 || sv.protocol == PROTOCOL_NEHAHRABJP3 || sv.protocol == PROTOCOL_DARKPLACES1 || sv.protocol == PROTOCOL_DARKPLACES2 || sv.protocol == PROTOCOL_DARKPLACES3 || sv.protocol == PROTOCOL_DARKPLACES4 || sv.protocol == PROTOCOL_DARKPLACES5)
+		move->buttons = MSG_ReadByte ();
 	else
-		move->buttons |= MSG_ReadLong ();
+		move->buttons = MSG_ReadLong ();
 	if (msg_badread) Con_Printf("SV_ReadClientMessage: badread at %s:%i\n", __FILE__, __LINE__);
 
 	// read impulse
-	i = MSG_ReadByte ();
-	if (i)
-		move->impulse = i;
+	move->impulse = MSG_ReadByte ();
 	if (msg_badread) Con_Printf("SV_ReadClientMessage: badread at %s:%i\n", __FILE__, __LINE__);
 
 	// PRYDON_CLIENTCURSOR
-	if (sv.protocol != PROTOCOL_QUAKE && sv.protocol != PROTOCOL_QUAKEDP && sv.protocol != PROTOCOL_NEHAHRAMOVIE && sv.protocol != PROTOCOL_DARKPLACES1 && sv.protocol != PROTOCOL_DARKPLACES2 && sv.protocol != PROTOCOL_DARKPLACES3 && sv.protocol != PROTOCOL_DARKPLACES4 && sv.protocol != PROTOCOL_DARKPLACES5)
+	if (sv.protocol != PROTOCOL_QUAKE && sv.protocol != PROTOCOL_QUAKEDP && sv.protocol != PROTOCOL_NEHAHRAMOVIE && sv.protocol != PROTOCOL_NEHAHRABJP && sv.protocol != PROTOCOL_NEHAHRABJP2 && sv.protocol != PROTOCOL_NEHAHRABJP3 && sv.protocol != PROTOCOL_DARKPLACES1 && sv.protocol != PROTOCOL_DARKPLACES2 && sv.protocol != PROTOCOL_DARKPLACES3 && sv.protocol != PROTOCOL_DARKPLACES4 && sv.protocol != PROTOCOL_DARKPLACES5)
 	{
 		// 30 bytes
 		move->cursor_screen[0] = MSG_ReadShort() * (1.0f / 32767.0f);
@@ -532,46 +530,129 @@ qboolean SV_ReadClientMove (void)
 		if (msg_badread) Con_Printf("SV_ReadClientMessage: badread at %s:%i\n", __FILE__, __LINE__);
 	}
 
-	// disable clientside movement prediction in some cases
-	if (ceil((move->receivetime - move->time) * 1000.0) < sv_clmovement_minping.integer)
-		host_client->clmovement_disabletimeout = realtime + sv_clmovement_minping_disabletime.value / 1000.0;
-	if (!sv_clmovement_enable.integer || host_client->clmovement_disabletimeout > realtime)
-		move->sequence = 0;
-
-	if (!host_client->spawned)
-		memset(move, 0, sizeof(*move));
-	else if (move->sequence && (float)move->time > (float)sv.time + 0.125f) // add a little fuzz factor due to float precision issues
+	// if the previous move has not been applied yet, we need to accumulate
+	// the impulse/buttons from it
+	if (!host_client->cmd.applied)
 	{
-		Con_DPrintf("client move->time %f > sv.time %f, kicking\n", (float)move->time, (float)sv.time);
-		// if the client is lying about time, we have definitively detected a
-		// speed cheat attempt of the worst sort, and we can immediately kick
-		// the offending player off.
-		// this fixes the timestamp to prevent a speed cheat from working
-		move->time = sv.time;
-		// but we kick the player for good measure
-		kickplayer = true;
+		if (!move->impulse)
+			move->impulse = host_client->cmd.impulse;
+		move->buttons |= host_client->cmd.buttons;
+	}
+
+	// now store this move for later execution
+	// (we have to buffer the moves because of old ones being repeated)
+	if (sv_numreadmoves < CL_MAX_USERCMDS)
+		sv_readmoves[sv_numreadmoves++] = *move;
+}
+
+void SV_ExecuteClientMoves(void)
+{
+	int moveindex;
+	float moveframetime;
+	double oldframetime;
+	double oldframetime2;
+#ifdef NUM_PING_TIMES
+	double total;
+#endif
+	prvm_eval_t *val;
+	if (sv_numreadmoves < 1)
+		return;
+	// only start accepting input once the player is spawned
+	if (!host_client->spawned)
+		return;
+#if DEBUGMOVES
+	Con_Printf("SV_ExecuteClientMoves: read %i moves at sv.time %f\n", sv_numreadmoves, (float)sv.time);
+#endif
+	// disable clientside movement prediction in some cases
+	if (ceil(max(sv_readmoves[sv_numreadmoves-1].receivetime - sv_readmoves[sv_numreadmoves-1].time, 0) * 1000.0) < sv_clmovement_minping.integer)
+		host_client->clmovement_disabletimeout = realtime + sv_clmovement_minping_disabletime.value / 1000.0;
+	// several conditions govern whether clientside movement prediction is allowed
+	if (sv_readmoves[sv_numreadmoves-1].sequence && sv_clmovement_enable.integer && sv_clmovement_waitforinput.integer > 0 && host_client->clmovement_disabletimeout <= realtime && host_client->edict->fields.server->movetype == MOVETYPE_WALK && (!(val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.disableclientprediction)) || !val->_float))
+	{
+		// process the moves in order and ignore old ones
+		// but always trust the latest move
+		// (this deals with bogus initial move sequences after level change,
+		//  where the client will eventually catch up with the level change
+		//  and reset its move sequence)
+		for (moveindex = 0;moveindex < sv_numreadmoves;moveindex++)
+		{
+			usercmd_t *move = sv_readmoves + moveindex;
+			if (host_client->movesequence < move->sequence || moveindex == sv_numreadmoves - 1)
+			{
+#if DEBUGMOVES
+				Con_Printf("%smove #%i %ims (%ims) %i %i '%i %i %i' '%i %i %i'\n", (move->time - host_client->cmd.time) > sv.frametime * 1.01 ? "^1" : "^2", move->sequence, (int)floor((move->time - host_client->cmd.time) * 1000.0 + 0.5), (int)floor(move->time * 1000.0 + 0.5), move->impulse, move->buttons, (int)move->viewangles[0], (int)move->viewangles[1], (int)move->viewangles[2], (int)move->forwardmove, (int)move->sidemove, (int)move->upmove);
+#endif
+				// this is a new move
+				move->time = bound(sv.time - 1, move->time, sv.time); // prevent slowhack/speedhack combos
+				move->time = max(move->time, host_client->cmd.time); // prevent backstepping of time
+				moveframetime = bound(0, move->time - host_client->cmd.time, 0.1);
+				//Con_Printf("movesequence = %i (%i lost), moveframetime = %f\n", move->sequence, move->sequence ? move->sequence - host_client->movesequence - 1 : 0, moveframetime);
+				host_client->cmd = *move;
+				host_client->movesequence = move->sequence;
+
+				// if using prediction, we need to perform moves when packets are
+				// received, even if multiple occur in one frame
+				// (they can't go beyond the current time so there is no cheat issue
+				//  with this approach, and if they don't send input for a while they
+				//  start moving anyway, so the longest 'lagaport' possible is
+				//  determined by the sv_clmovement_waitforinput cvar)
+				if (moveframetime <= 0)
+					continue;
+				oldframetime = prog->globals.server->frametime;
+				oldframetime2 = sv.frametime;
+				// update ping time for qc to see while executing this move
+				host_client->ping = host_client->cmd.receivetime - host_client->cmd.time;
+				// the server and qc frametime values must be changed temporarily
+				prog->globals.server->frametime = sv.frametime = moveframetime;
+				// if move is more than 50ms, split it into two moves (this matches QWSV behavior and the client prediction)
+				if (sv.frametime > 0.05)
+				{
+					prog->globals.server->frametime = sv.frametime = moveframetime * 0.5f;
+					SV_Physics_ClientMove();
+				}
+				SV_Physics_ClientMove();
+				sv.frametime = oldframetime2;
+				prog->globals.server->frametime = oldframetime;
+				host_client->clmovement_skipphysicsframes = sv_clmovement_waitforinput.integer;
+			}
+		}
 	}
 	else
 	{
-		// apply the latest accepted move to the entity fields
-		host_client->movesequence = move->sequence;
-		if (host_client->movesequence && sv_clmovement_waitforinput.integer > 0)
+		// try to gather button bits from old moves, but only if their time is
+		// advancing (ones with the same timestamp can't be trusted)
+		for (moveindex = 0;moveindex < sv_numreadmoves-1;moveindex++)
 		{
-			double frametime = bound(0, move->time - oldmovetime, 0.1);
-			double oldframetime = prog->globals.server->frametime;
-			double oldframetime2 = sv.frametime;
-			//if (move->time - oldmovetime >= 0.1001)
-			//	Con_DPrintf("client move exceeds 100ms!  (time %f -> time %f)\n", oldmovetime, move->time);
-			// the server and qc frametime values must be changed temporarily
-			sv.frametime = frametime;
-			prog->globals.server->frametime = frametime;
-			SV_Physics_ClientEntity(host_client->edict);
-			sv.frametime = oldframetime2;
-			prog->globals.server->frametime = oldframetime;
-			host_client->clmovement_skipphysicsframes = sv_clmovement_waitforinput.integer;
+			usercmd_t *move = sv_readmoves + moveindex;
+			if (host_client->cmd.time < move->time)
+			{
+				sv_readmoves[sv_numreadmoves-1].buttons |= move->buttons;
+				if (move->impulse)
+					sv_readmoves[sv_numreadmoves-1].impulse = move->impulse;
+			}
 		}
+		// now copy the new move
+		host_client->cmd = sv_readmoves[sv_numreadmoves-1];
+		host_client->cmd.time = max(host_client->cmd.time, sv.time);
+			// physics will run up to sv.time, so allow no predicted moves
+			// before that otherwise, there is a speedhack by turning
+			// prediction on and off repeatedly on client side because the
+			// engine would run BOTH client and server physics for the same
+			// time
+		host_client->movesequence = 0;
+		// make sure that normal physics takes over immediately
+		host_client->clmovement_skipphysicsframes = 0;
 	}
-	return kickplayer;
+
+	// calculate average ping time
+	host_client->ping = host_client->cmd.receivetime - host_client->cmd.time;
+#ifdef NUM_PING_TIMES
+	host_client->ping_times[host_client->num_pings % NUM_PING_TIMES] = host_client->cmd.receivetime - host_client->cmd.time;
+	host_client->num_pings++;
+	for (i=0, total = 0;i < NUM_PING_TIMES;i++)
+		total += host_client->ping_times[i];
+	host_client->ping = total / NUM_PING_TIMES;
+#endif
 }
 
 void SV_ApplyClientMove (void)
@@ -597,29 +678,29 @@ void SV_ApplyClientMove (void)
 	// only send the impulse to qc once
 	move->impulse = 0;
 	VectorCopy(move->viewangles, host_client->edict->fields.server->v_angle);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button3))) val->_float = ((move->buttons >> 2) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button4))) val->_float = ((move->buttons >> 3) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button5))) val->_float = ((move->buttons >> 4) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button6))) val->_float = ((move->buttons >> 5) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button7))) val->_float = ((move->buttons >> 6) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button8))) val->_float = ((move->buttons >> 7) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button9))) val->_float = ((move->buttons >> 11) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button10))) val->_float = ((move->buttons >> 12) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button11))) val->_float = ((move->buttons >> 13) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button12))) val->_float = ((move->buttons >> 14) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button13))) val->_float = ((move->buttons >> 15) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button14))) val->_float = ((move->buttons >> 16) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button15))) val->_float = ((move->buttons >> 17) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_button16))) val->_float = ((move->buttons >> 18) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_buttonuse))) val->_float = ((move->buttons >> 8) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_buttonchat))) val->_float = ((move->buttons >> 9) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_cursor_active))) val->_float = ((move->buttons >> 10) & 1);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_movement))) VectorSet(val->vector, move->forwardmove, move->sidemove, move->upmove);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_cursor_screen))) VectorCopy(move->cursor_screen, val->vector);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_cursor_trace_start))) VectorCopy(move->cursor_start, val->vector);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_cursor_trace_endpos))) VectorCopy(move->cursor_impact, val->vector);
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_cursor_trace_ent))) val->edict = PRVM_EDICT_TO_PROG(PRVM_EDICT_NUM(move->cursor_entitynumber));
-	if ((val = PRVM_GETEDICTFIELDVALUE(host_client->edict, eval_ping))) val->_float = host_client->ping * 1000.0;
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button3))) val->_float = ((move->buttons >> 2) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button4))) val->_float = ((move->buttons >> 3) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button5))) val->_float = ((move->buttons >> 4) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button6))) val->_float = ((move->buttons >> 5) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button7))) val->_float = ((move->buttons >> 6) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button8))) val->_float = ((move->buttons >> 7) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button9))) val->_float = ((move->buttons >> 11) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button10))) val->_float = ((move->buttons >> 12) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button11))) val->_float = ((move->buttons >> 13) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button12))) val->_float = ((move->buttons >> 14) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button13))) val->_float = ((move->buttons >> 15) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button14))) val->_float = ((move->buttons >> 16) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button15))) val->_float = ((move->buttons >> 17) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.button16))) val->_float = ((move->buttons >> 18) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.buttonuse))) val->_float = ((move->buttons >> 8) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.buttonchat))) val->_float = ((move->buttons >> 9) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.cursor_active))) val->_float = ((move->buttons >> 10) & 1);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.movement))) VectorSet(val->vector, move->forwardmove, move->sidemove, move->upmove);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.cursor_screen))) VectorCopy(move->cursor_screen, val->vector);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.cursor_trace_start))) VectorCopy(move->cursor_start, val->vector);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.cursor_trace_endpos))) VectorCopy(move->cursor_impact, val->vector);
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.cursor_trace_ent))) val->edict = PRVM_EDICT_TO_PROG(PRVM_EDICT_NUM(move->cursor_entitynumber));
+	if ((val = PRVM_EDICTFIELDVALUE(host_client->edict, prog->fieldoffsets.ping))) val->_float = host_client->ping * 1000.0;
 }
 
 void SV_FrameLost(int framenum)
@@ -644,12 +725,17 @@ SV_ReadClientMessage
 ===================
 */
 extern void SV_SendServerinfo(client_t *client);
+extern sizebuf_t vm_tempstringsbuf;
 void SV_ReadClientMessage(void)
 {
-	int cmd, num;
-	char *s;
+	int cmd, num, start;
+	char *s, *p, *q;
+
+	if(sv_autodemo_perclient.integer >= 2)
+		SV_WriteDemoMessage(host_client, &(host_client->netconnection->message), true);
 
 	//MSG_BeginReading ();
+	sv_numreadmoves = 0;
 
 	for(;;)
 	{
@@ -671,6 +757,8 @@ void SV_ReadClientMessage(void)
 		if (cmd == -1)
 		{
 			// end of message
+			// apply the moves that were read this frame
+			SV_ExecuteClientMoves();
 			break;
 		}
 
@@ -685,36 +773,113 @@ void SV_ReadClientMessage(void)
 			break;
 
 		case clc_stringcmd:
+			// allow reliable messages now as the client is done with initial loading
+			if (host_client->sendsignon == 2)
+				host_client->sendsignon = 0;
 			s = MSG_ReadString ();
+			q = NULL;
+			for(p = s; *p; ++p) switch(*p)
+			{
+				case 10:
+				case 13:
+					if(!q)
+						q = p;
+					break;
+				default:
+					if(q)
+						goto clc_stringcmd_invalid; // newline seen, THEN something else -> possible exploit
+					break;
+			}
+			if(q)
+				*q = 0;
 			if (strncasecmp(s, "spawn", 5) == 0
 			 || strncasecmp(s, "begin", 5) == 0
 			 || strncasecmp(s, "prespawn", 8) == 0)
 				Cmd_ExecuteString (s, src_client);
-			else if (SV_ParseClientCommandQC)
+			else if (prog->funcoffsets.SV_ParseClientCommand)
 			{
-				PRVM_G_INT(OFS_PARM0) = PRVM_SetEngineString(s);
+				int restorevm_tempstringsbuf_cursize;
+				restorevm_tempstringsbuf_cursize = vm_tempstringsbuf.cursize;
+				PRVM_G_INT(OFS_PARM0) = PRVM_SetTempString(s);
 				prog->globals.server->self = PRVM_EDICT_TO_PROG(host_client->edict);
-				PRVM_ExecuteProgram ((func_t)(SV_ParseClientCommandQC - prog->functions), "QC function SV_ParseClientCommand is missing");
+				PRVM_ExecuteProgram (prog->funcoffsets.SV_ParseClientCommand, "QC function SV_ParseClientCommand is missing");
+				vm_tempstringsbuf.cursize = restorevm_tempstringsbuf_cursize;
 			}
 			else
 				Cmd_ExecuteString (s, src_client);
 			break;
 
+clc_stringcmd_invalid:
+			Con_Printf("Received invalid stringcmd from %s\n", host_client->name);
+			if(developer.integer)
+				Com_HexDumpToConsole((unsigned char *) s, strlen(s));
+			break;
+
 		case clc_disconnect:
 			SV_DropClient (false); // client wants to disconnect
 			return;
 
 		case clc_move:
-			// if ReadClientMove returns true, the client tried to speed cheat
-			if (SV_ReadClientMove ())
-				SV_DropClient (false);
+			SV_ReadClientMove();
+			break;
+
+		case clc_ackdownloaddata:
+			start = MSG_ReadLong();
+			num = MSG_ReadShort();
+			if (host_client->download_file && host_client->download_started)
+			{
+				if (host_client->download_expectedposition == start)
+				{
+					int size = (int)FS_FileSize(host_client->download_file);
+					// a data block was successfully received by the client,
+					// update the expected position on the next data block
+					host_client->download_expectedposition = start + num;
+					// if this was the last data block of the file, it's done
+					if (host_client->download_expectedposition >= FS_FileSize(host_client->download_file))
+					{
+						// tell the client that the download finished
+						// we need to calculate the crc now
+						//
+						// note: at this point the OS probably has the file
+						// entirely in memory, so this is a faster operation
+						// now than it was when the download started.
+						//
+						// it is also preferable to do this at the end of the
+						// download rather than the start because it reduces
+						// potential for Denial Of Service attacks against the
+						// server.
+						int crc;
+						unsigned char *temp;
+						FS_Seek(host_client->download_file, 0, SEEK_SET);
+						temp = Mem_Alloc(tempmempool, size);
+						FS_Read(host_client->download_file, temp, size);
+						crc = CRC_Block(temp, size);
+						Mem_Free(temp);
+						// calculated crc, send the file info to the client
+						// (so that it can verify the data)
+						Host_ClientCommands(va("\ncl_downloadfinished %i %i %s\n", size, crc, host_client->download_name));
+						Con_DPrintf("Download of %s by %s has finished\n", host_client->download_name, host_client->name);
+						FS_Close(host_client->download_file);
+						host_client->download_file = NULL;
+						host_client->download_name[0] = 0;
+						host_client->download_expectedposition = 0;
+						host_client->download_started = false;
+					}
+				}
+				else
+				{
+					// a data block was lost, reset to the expected position
+					// and resume sending from there
+					FS_Seek(host_client->download_file, host_client->download_expectedposition, SEEK_SET);
+				}
+			}
 			break;
 
 		case clc_ackframe:
 			if (msg_badread) Con_Printf("SV_ReadClientMessage: badread at %s:%i\n", __FILE__, __LINE__);
 			num = MSG_ReadLong();
 			if (msg_badread) Con_Printf("SV_ReadClientMessage: badread at %s:%i\n", __FILE__, __LINE__);
-			if (developer_networkentities.integer >= 1)
+			if (developer_networkentities.integer >= 10)
 				Con_Printf("recv clc_ackframe %i\n", num);
 			// if the client hasn't progressed through signons yet,
 			// ignore any clc_ackframes we get (they're probably from the
