@@ -42,6 +42,8 @@ int gl_support_clamptoedge = false;
 // GL_EXT_texture_filter_anisotropic
 int gl_support_anisotropy = false;
 int gl_max_anisotropy = 1;
+// OpenGL2.0 core glStencilOpSeparate, or GL_ATI_separate_stencil
+int gl_support_separatestencil = false;
 // GL_EXT_stencil_two_side
 int gl_support_stenciltwoside = false;
 // GL_EXT_blend_minmax
@@ -56,6 +58,10 @@ int gl_support_shading_language_100 = false;
 int gl_support_vertex_shader = false;
 // GL_ARB_fragment_shader
 int gl_support_fragment_shader = false;
+//GL_ARB_vertex_buffer_object
+int gl_support_arb_vertex_buffer_object = false;
+//GL_ARB_texture_compression
+int gl_support_texture_compression = false;
 
 // LordHavoc: if window is hidden, don't update screen
 qboolean vid_hidden = true;
@@ -76,19 +82,26 @@ cvar_t vid_fullscreen = {CVAR_SAVE, "vid_fullscreen", "1", "use fullscreen (1) o
 cvar_t vid_width = {CVAR_SAVE, "vid_width", "640", "resolution"};
 cvar_t vid_height = {CVAR_SAVE, "vid_height", "480", "resolution"};
 cvar_t vid_bitsperpixel = {CVAR_SAVE, "vid_bitsperpixel", "32", "how many bits per pixel to render at (32 or 16, 32 is recommended)"};
+cvar_t vid_samples = {CVAR_SAVE, "vid_samples", "1", "how many anti-aliasing samples per pixel to request from the graphics driver (4 is recommended, 1 is faster)"};
 cvar_t vid_refreshrate = {CVAR_SAVE, "vid_refreshrate", "60", "refresh rate to use, in hz (higher values flicker less, if supported by your monitor)"};
+cvar_t vid_userefreshrate = {CVAR_SAVE, "vid_userefreshrate", "0", "set this to 1 to make vid_refreshrate used, or to 0 to let the engine choose a sane default"};
 cvar_t vid_stereobuffer = {CVAR_SAVE, "vid_stereobuffer", "0", "enables 'quad-buffered' stereo rendering for stereo shutterglasses, HMD (head mounted display) devices, or polarized stereo LCDs, if supported by your drivers"};
 
 cvar_t vid_vsync = {CVAR_SAVE, "vid_vsync", "0", "sync to vertical blank, prevents 'tearing' (seeing part of one frame and part of another on the screen at the same time), automatically disabled when doing timedemo benchmarks"};
 cvar_t vid_mouse = {CVAR_SAVE, "vid_mouse", "1", "whether to use the mouse in windowed mode (fullscreen always does)"};
+cvar_t vid_grabkeyboard = {CVAR_SAVE, "vid_grabkeyboard", "0", "whether to grab the keyboard when mouse is active (prevents use of volume control keys, music player keys, etc on some keyboards)"};
 cvar_t vid_minwidth = {0, "vid_minwidth", "0", "minimum vid_width that is acceptable (to be set in default.cfg in mods)"};
 cvar_t vid_minheight = {0, "vid_minheight", "0", "minimum vid_height that is acceptable (to be set in default.cfg in mods)"};
 cvar_t gl_combine = {0, "gl_combine", "1", "faster rendering by using GL_ARB_texture_env_combine extension (part of OpenGL 1.3 and above)"};
 cvar_t gl_finish = {0, "gl_finish", "0", "make the cpu wait for the graphics processor at the end of each rendered frame (can help with strange input or video lag problems on some machines)"};
 
+cvar_t vid_stick_mouse = {CVAR_SAVE, "vid_stick_mouse", "0", "have the mouse stuck in the center of the screen" };
+cvar_t vid_resizable = {CVAR_SAVE, "vid_resizable", "0", "0: window not resizable, 1: resizable, 2: window can be resized but the framebuffer isn't adjusted" };
+
 cvar_t v_gamma = {CVAR_SAVE, "v_gamma", "1", "inverse gamma correction value, a brightness effect that does not affect white or black, and tends to make the image grey and dull"};
 cvar_t v_contrast = {CVAR_SAVE, "v_contrast", "1", "brightness of white (values above 1 give a brighter image with increased color saturation, unlike v_gamma)"};
 cvar_t v_brightness = {CVAR_SAVE, "v_brightness", "0", "brightness of black, useful for monitors that are too dark"};
+cvar_t v_contrastboost = {CVAR_SAVE, "v_contrastboost", "1", "by how much to multiply the contrast in dark areas (1 is no change)"};
 cvar_t v_color_enable = {CVAR_SAVE, "v_color_enable", "0", "enables black-grey-white color correction curve controls"};
 cvar_t v_color_black_r = {CVAR_SAVE, "v_color_black_r", "0", "desired color of black"};
 cvar_t v_color_black_g = {CVAR_SAVE, "v_color_black_g", "0", "desired color of black"};
@@ -100,6 +113,7 @@ cvar_t v_color_white_r = {CVAR_SAVE, "v_color_white_r", "1", "desired color of w
 cvar_t v_color_white_g = {CVAR_SAVE, "v_color_white_g", "1", "desired color of white"};
 cvar_t v_color_white_b = {CVAR_SAVE, "v_color_white_b", "1", "desired color of white"};
 cvar_t v_hwgamma = {CVAR_SAVE, "v_hwgamma", "1", "enables use of hardware gamma correction ramps if available (note: does not work very well on Windows2000 and above), values are 0 = off, 1 = attempt to use hardware gamma, 2 = use hardware gamma whether it works or not"};
+cvar_t v_glslgamma = {CVAR_SAVE, "v_glslgamma", "0", "enables use of GLSL to apply gamma correction ramps if available (note: overrides v_hwgamma)"};
 cvar_t v_psycho = {0, "v_psycho", "0", "easter egg (does not work on Windows2000 or above)"};
 
 // brand of graphics chip
@@ -214,6 +228,7 @@ void (GLAPIENTRY *qglTexEnvi)(GLenum target, GLenum pname, GLint param);
 void (GLAPIENTRY *qglTexParameterf)(GLenum target, GLenum pname, GLfloat param);
 void (GLAPIENTRY *qglTexParameterfv)(GLenum target, GLenum pname, GLfloat *params);
 void (GLAPIENTRY *qglTexParameteri)(GLenum target, GLenum pname, GLint param);
+void (GLAPIENTRY *qglHint)(GLenum target, GLenum mode);
 
 void (GLAPIENTRY *qglGenTextures)(GLsizei n, GLuint *textures);
 void (GLAPIENTRY *qglDeleteTextures)(GLsizei n, const GLuint *textures);
@@ -248,12 +263,17 @@ void (GLAPIENTRY *qglPolygonOffset)(GLfloat factor, GLfloat units);
 void (GLAPIENTRY *qglPolygonMode)(GLenum face, GLenum mode);
 void (GLAPIENTRY *qglPolygonStipple)(const GLubyte *mask);
 
+void (GLAPIENTRY *qglClipPlane)(GLenum plane, const GLdouble *equation);
+void (GLAPIENTRY *qglGetClipPlane)(GLenum plane, GLdouble *equation);
+
 //[515]: added on 29.07.2005
 void (GLAPIENTRY *qglLineWidth)(GLfloat width);
 void (GLAPIENTRY *qglPointSize)(GLfloat size);
 
 void (GLAPIENTRY *qglBlendEquationEXT)(GLenum);
 
+void (GLAPIENTRY *qglStencilOpSeparate)(GLenum, GLenum, GLenum, GLenum);
+void (GLAPIENTRY *qglStencilFuncSeparate)(GLenum, GLenum, GLint, GLuint);
 void (GLAPIENTRY *qglActiveStencilFaceEXT)(GLenum);
 
 void (GLAPIENTRY *qglDeleteObjectARB)(GLhandleARB obj);
@@ -343,19 +363,37 @@ GLint (GLAPIENTRY *qglGetAttribLocationARB)(GLhandleARB programObj, const GLchar
 //void (GLAPIENTRY *qglGetVertexAttribivARB)(GLuint index, GLenum pname, GLint *params);
 //void (GLAPIENTRY *qglGetVertexAttribPointervARB)(GLuint index, GLenum pname, GLvoid **pointer);
 
+//GL_ARB_vertex_buffer_object
+void (GLAPIENTRY *qglBindBufferARB) (GLenum target, GLuint buffer);
+void (GLAPIENTRY *qglDeleteBuffersARB) (GLsizei n, const GLuint *buffers);
+void (GLAPIENTRY *qglGenBuffersARB) (GLsizei n, GLuint *buffers);
+GLboolean (GLAPIENTRY *qglIsBufferARB) (GLuint buffer);
+GLvoid* (GLAPIENTRY *qglMapBufferARB) (GLenum target, GLenum access);
+GLboolean (GLAPIENTRY *qglUnmapBufferARB) (GLenum target);
+void (GLAPIENTRY *qglBufferDataARB) (GLenum target, GLsizeiptrARB size, const GLvoid *data, GLenum usage);
+void (GLAPIENTRY *qglBufferSubDataARB) (GLenum target, GLintptrARB offset, GLsizeiptrARB size, const GLvoid *data);
+
+void (GLAPIENTRY *qglCompressedTexImage3DARB)(GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, GLint border, GLsizei imageSize, const void *data);
+void (GLAPIENTRY *qglCompressedTexImage2DARB)(GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLint border,  GLsizei imageSize, const void *data);
+void (GLAPIENTRY *qglCompressedTexImage1DARB)(GLenum target, GLint level, GLenum internalformat, GLsizei width, GLint border, GLsizei imageSize, const void *data);
+void (GLAPIENTRY *qglCompressedTexSubImage3DARB)(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLsizei width, GLsizei height, GLsizei depth, GLenum format, GLsizei imageSize, const void *data);
+void (GLAPIENTRY *qglCompressedTexSubImage2DARB)(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLsizei imageSize, const void *data);
+void (GLAPIENTRY *qglCompressedTexSubImage1DARB)(GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLsizei imageSize, const void *data);
+void (GLAPIENTRY *qglGetCompressedTexImageARB)(GLenum target, GLint lod, void *img);
+
 int GL_CheckExtension(const char *name, const dllfunction_t *funcs, const char *disableparm, int silent)
 {
 	int failed = false;
 	const dllfunction_t *func;
 
-	Con_Printf("checking for %s...  ", name);
+	Con_DPrintf("checking for %s...  ", name);
 
 	for (func = funcs;func && func->name;func++)
 		*func->funcvariable = NULL;
 
 	if (disableparm && (COM_CheckParm(disableparm) || COM_CheckParm("-safe")))
 	{
-		Con_Print("disabled by commandline\n");
+		Con_DPrint("disabled by commandline\n");
 		return false;
 	}
 
@@ -367,19 +405,19 @@ int GL_CheckExtension(const char *name, const dllfunction_t *funcs, const char *
 			if (!(*func->funcvariable = (void *) GL_GetProcAddress(func->name)))
 			{
 				if (!silent)
-					Con_Printf("OpenGL extension \"%s\" is missing function \"%s\" - broken driver!\n", name, func->name);
+					Con_DPrintf("OpenGL extension \"%s\" is missing function \"%s\" - broken driver!\n", name, func->name);
 				failed = true;
 			}
 		}
 		// delay the return so it prints all missing functions
 		if (failed)
 			return false;
-		Con_Print("enabled\n");
+		Con_DPrint("enabled\n");
 		return true;
 	}
 	else
 	{
-		Con_Print("not detected\n");
+		Con_DPrint("not detected\n");
 		return false;
 	}
 }
@@ -458,6 +496,7 @@ static dllfunction_t opengl110funcs[] =
 	{"glTexParameterf", (void **) &qglTexParameterf},
 	{"glTexParameterfv", (void **) &qglTexParameterfv},
 	{"glTexParameteri", (void **) &qglTexParameteri},
+	{"glHint", (void **) &qglHint},
 	{"glPixelStoref", (void **) &qglPixelStoref},
 	{"glPixelStorei", (void **) &qglPixelStorei},
 	{"glGenTextures", (void **) &qglGenTextures},
@@ -478,6 +517,8 @@ static dllfunction_t opengl110funcs[] =
 	{"glPolygonOffset", (void **) &qglPolygonOffset},
 	{"glPolygonMode", (void **) &qglPolygonMode},
 	{"glPolygonStipple", (void **) &qglPolygonStipple},
+	{"glClipPlane", (void **) &qglClipPlane},
+	{"glGetClipPlane", (void **) &qglGetClipPlane},
 	{NULL, NULL}
 };
 
@@ -519,6 +560,20 @@ static dllfunction_t texture3dextfuncs[] =
 	{NULL, NULL}
 };
 
+static dllfunction_t atiseparatestencilfuncs[] =
+{
+	{"glStencilOpSeparateATI", (void **) &qglStencilOpSeparate},
+	{"glStencilFuncSeparateATI", (void **) &qglStencilFuncSeparate},
+	{NULL, NULL}
+};
+
+static dllfunction_t gl2separatestencilfuncs[] =
+{
+	{"glStencilOpSeparate", (void **) &qglStencilOpSeparate},
+	{"glStencilFuncSeparate", (void **) &qglStencilFuncSeparate},
+	{NULL, NULL}
+};
+
 static dllfunction_t stenciltwosidefuncs[] =
 {
 	{"glActiveStencilFaceEXT", (void **) &qglActiveStencilFaceEXT},
@@ -626,6 +681,30 @@ static dllfunction_t vertexshaderfuncs[] =
 	{NULL, NULL}
 };
 
+static dllfunction_t vbofuncs[] =
+{
+	{"glBindBufferARB"    , (void **) &qglBindBufferARB},
+	{"glDeleteBuffersARB" , (void **) &qglDeleteBuffersARB},
+	{"glGenBuffersARB"    , (void **) &qglGenBuffersARB},
+	{"glIsBufferARB"      , (void **) &qglIsBufferARB},
+	{"glMapBufferARB"     , (void **) &qglMapBufferARB},
+	{"glUnmapBufferARB"   , (void **) &qglUnmapBufferARB},
+	{"glBufferDataARB"    , (void **) &qglBufferDataARB},
+	{"glBufferSubDataARB" , (void **) &qglBufferSubDataARB},
+	{NULL, NULL}
+};
+
+static dllfunction_t texturecompressionfuncs[] =
+{
+	{"glCompressedTexImage3DARB",    (void **) &qglCompressedTexImage3DARB},
+	{"glCompressedTexImage2DARB",    (void **) &qglCompressedTexImage2DARB},
+	{"glCompressedTexImage1DARB",    (void **) &qglCompressedTexImage1DARB},
+	{"glCompressedTexSubImage3DARB", (void **) &qglCompressedTexSubImage3DARB},
+	{"glCompressedTexSubImage2DARB", (void **) &qglCompressedTexSubImage2DARB},
+	{"glCompressedTexSubImage1DARB", (void **) &qglCompressedTexSubImage1DARB},
+	{"glGetCompressedTexImageARB",   (void **) &qglGetCompressedTexImageARB},
+	{NULL, NULL}
+};
 
 void VID_CheckExtensions(void)
 {
@@ -646,6 +725,7 @@ void VID_CheckExtensions(void)
 	gl_support_clamptoedge = false;
 	gl_support_anisotropy = false;
 	gl_max_anisotropy = 1;
+	gl_support_separatestencil = false;
 	gl_support_stenciltwoside = false;
 	gl_support_ext_blend_minmax = false;
 	gl_support_ext_blend_subtract = false;
@@ -653,20 +733,16 @@ void VID_CheckExtensions(void)
 	gl_support_shading_language_100 = false;
 	gl_support_vertex_shader = false;
 	gl_support_fragment_shader = false;
+	gl_support_arb_vertex_buffer_object = false;
+	gl_support_texture_compression = false;
 
 	if (!GL_CheckExtension("OpenGL 1.1.0", opengl110funcs, NULL, false))
 		Sys_Error("OpenGL 1.1.0 functions not found");
 
-	Con_Printf("GL_VENDOR: %s\n", gl_vendor);
-	Con_Printf("GL_RENDERER: %s\n", gl_renderer);
-	Con_Printf("GL_VERSION: %s\n", gl_version);
-	Con_Printf("GL_EXTENSIONS: %s\n", gl_extensions);
-	Con_Printf("%s_EXTENSIONS: %s\n", gl_platform, gl_platformextensions);
-
 	CHECKGLERROR
 	qglGetIntegerv(GL_MAX_TEXTURE_SIZE, &gl_max_texture_size);
 
-	Con_Print("Checking OpenGL extensions...\n");
+	Con_DPrint("Checking OpenGL extensions...\n");
 
 // COMMANDLINEOPTION: GL: -nodrawrangeelements disables GL_EXT_draw_range_elements (renders faster)
 	if (!GL_CheckExtension("glDrawRangeElements", drawrangeelementsfuncs, "-nodrawrangeelements", true))
@@ -696,7 +772,11 @@ void VID_CheckExtensions(void)
 // COMMANDLINEOPTION: GL: -nocubemap disables GL_ARB_texture_cube_map (required for bumpmapping)
 	if ((gl_texturecubemap = GL_CheckExtension("GL_ARB_texture_cube_map", NULL, "-nocubemap", false)))
 		qglGetIntegerv(GL_MAX_CUBE_MAP_TEXTURE_SIZE_ARB, &gl_max_cube_map_texture_size);
+	// TODO: blacklist this extension on Radeon X1600-X1950 hardware (they support it only with certain filtering/repeat modes)
+// COMMANDLINEOPTION: GL: -notexturenonpoweroftwo disables GL_ARB_texture_non_power_of_two (which saves video memory if it is supported, but crashes on some buggy drivers)
 	gl_support_arb_texture_non_power_of_two = GL_CheckExtension("GL_ARB_texture_non_power_of_two", NULL, "-notexturenonpoweroftwo", false);
+// COMMANDLINEOPTION: GL: -notexturecompression disables GL_ARB_texture_compression (which saves video memory if it is supported, but can also degrade image quality, see gl_texturecompression cvar documentation for more information)
+	gl_support_texture_compression = GL_CheckExtension("GL_ARB_texture_compression", texturecompressionfuncs, "-notexturecompression", false);
 // COMMANDLINEOPTION: GL: -nocva disables GL_EXT_compiled_vertex_array (renders faster)
 	gl_supportslockarrays = GL_CheckExtension("GL_EXT_compiled_vertex_array", compiledvertexarrayfuncs, "-nocva", false);
 // COMMANDLINEOPTION: GL: -noedgeclamp disables GL_EXT_texture_edge_clamp or GL_SGIS_texture_edge_clamp (recommended, some cards do not support the other texture clamp method)
@@ -709,9 +789,15 @@ void VID_CheckExtensions(void)
 	gl_support_ext_blend_minmax = GL_CheckExtension("GL_EXT_blend_minmax", blendequationfuncs, "-noblendminmax", false);
 	gl_support_ext_blend_subtract = GL_CheckExtension("GL_EXT_blend_subtract", blendequationfuncs, "-noblendsubtract", false);
 
-// COMMANDLINEOPTION: GL: -nostenciltwoside disables GL_EXT_stencil_two_side (accelerates shadow rendering)
+// COMMANDLINEOPTION: GL: -noseparatestencil disables use of OpenGL2.0 glStencilOpSeparate and GL_ATI_separate_stencil extensions (which accelerate shadow rendering)
+	if (!(gl_support_separatestencil = GL_CheckExtension("glStencilOpSeparate", gl2separatestencilfuncs, "-noseparatestencil", true)))
+		gl_support_separatestencil = GL_CheckExtension("GL_ATI_separate_stencil", atiseparatestencilfuncs, "-noseparatestencil", false);
+// COMMANDLINEOPTION: GL: -nostenciltwoside disables GL_EXT_stencil_two_side (which accelerate shadow rendering)
 	gl_support_stenciltwoside = GL_CheckExtension("GL_EXT_stencil_two_side", stenciltwosidefuncs, "-nostenciltwoside", false);
 
+// COMMANDLINEOPTION: GL: -novbo disables GL_ARB_vertex_buffer_object (which accelerates rendering)
+	gl_support_arb_vertex_buffer_object = GL_CheckExtension("GL_ARB_vertex_buffer_object", vbofuncs, "-novbo", false);
+
 	// we don't care if it's an extension or not, they are identical functions, so keep it simple in the rendering code
 	if (qglDrawRangeElements == NULL)
 		qglDrawRangeElements = qglDrawRangeElementsEXT;
@@ -733,51 +819,151 @@ void Force_CenterView_f (void)
 }
 
 static int gamma_forcenextframe = false;
-static float cachegamma, cachebrightness, cachecontrast, cacheblack[3], cachegrey[3], cachewhite[3];
+static float cachegamma, cachebrightness, cachecontrast, cacheblack[3], cachegrey[3], cachewhite[3], cachecontrastboost;
 static int cachecolorenable, cachehwgamma;
-#define BOUNDCVAR(cvar, m1, m2) c = &(cvar);f = bound(m1, c->value, m2);if (c->value != f) Cvar_SetValueQuick(c, f);
+
+unsigned int vid_gammatables_serial = 0; // so other subsystems can poll if gamma parameters have changed
+qboolean vid_gammatables_trivial = true;
+void VID_BuildGammaTables(unsigned short *ramps, int rampsize)
+{
+	if (cachecolorenable)
+	{
+		BuildGammaTable16(1.0f, invpow(0.5, 1 - cachegrey[0]), cachewhite[0], cacheblack[0], cachecontrastboost, ramps, rampsize);
+		BuildGammaTable16(1.0f, invpow(0.5, 1 - cachegrey[1]), cachewhite[1], cacheblack[1], cachecontrastboost, ramps + rampsize, rampsize);
+		BuildGammaTable16(1.0f, invpow(0.5, 1 - cachegrey[2]), cachewhite[2], cacheblack[2], cachecontrastboost, ramps + rampsize*2, rampsize);
+	}
+	else
+	{
+		BuildGammaTable16(1.0f, cachegamma, cachecontrast, cachebrightness, cachecontrastboost, ramps, rampsize);
+		BuildGammaTable16(1.0f, cachegamma, cachecontrast, cachebrightness, cachecontrastboost, ramps + rampsize, rampsize);
+		BuildGammaTable16(1.0f, cachegamma, cachecontrast, cachebrightness, cachecontrastboost, ramps + rampsize*2, rampsize);
+	}
+
+	// LordHavoc: this code came from Ben Winslow and Zinx Verituse, I have
+	// immensely butchered it to work with variable framerates and fit in with
+	// the rest of darkplaces.
+	if (v_psycho.integer)
+	{
+		int x, y;
+		float t;
+		static float n[3], nd[3], nt[3];
+		static int init = true;
+		unsigned short *ramp;
+		gamma_forcenextframe = true;
+		if (init)
+		{
+			init = false;
+			for (x = 0;x < 3;x++)
+			{
+				n[x] = lhrandom(0, 1);
+				nd[x] = (rand()&1)?-0.25:0.25;
+				nt[x] = lhrandom(1, 8.2);
+			}
+		}
+
+		for (x = 0;x < 3;x++)
+		{
+			nt[x] -= cl.realframetime;
+			if (nt[x] < 0)
+			{
+				nd[x] = -nd[x];
+				nt[x] += lhrandom(1, 8.2);
+			}
+			n[x] += nd[x] * cl.realframetime;
+			n[x] -= floor(n[x]);
+		}
+
+		for (x = 0, ramp = ramps;x < 3;x++)
+			for (y = 0, t = n[x] - 0.75f;y < rampsize;y++, t += 0.75f * (2.0f / rampsize))
+				*ramp++ = (unsigned short)(cos(t*(M_PI*2.0)) * 32767.0f + 32767.0f);
+	}
+}
+
 void VID_UpdateGamma(qboolean force, int rampsize)
 {
 	cvar_t *c;
 	float f;
+	int wantgamma;
+	qboolean gamma_changed = false;
 
 	// LordHavoc: don't mess with gamma tables if running dedicated
 	if (cls.state == ca_dedicated)
 		return;
 
-	if (!force
-	 && !gamma_forcenextframe
-	 && !v_psycho.integer
-	 && cachehwgamma == (vid_activewindow ? v_hwgamma.integer : 0)
-	 && v_gamma.value == cachegamma
-	 && v_contrast.value == cachecontrast
-	 && v_brightness.value == cachebrightness
-	 && cachecolorenable == v_color_enable.integer
-	 && cacheblack[0] == v_color_black_r.value
-	 && cacheblack[1] == v_color_black_g.value
-	 && cacheblack[2] == v_color_black_b.value
-	 && cachegrey[0] == v_color_grey_r.value
-	 && cachegrey[1] == v_color_grey_g.value
-	 && cachegrey[2] == v_color_grey_b.value
-	 && cachewhite[0] == v_color_white_r.value
-	 && cachewhite[1] == v_color_white_g.value
-	 && cachewhite[2] == v_color_white_b.value)
-		return;
+	wantgamma = v_hwgamma.integer;
+	if(r_glsl.integer && v_glslgamma.integer)
+		wantgamma = 0;
+	if(!vid_activewindow)
+		wantgamma = 0;
+#define BOUNDCVAR(cvar, m1, m2) c = &(cvar);f = bound(m1, c->value, m2);if (c->value != f) Cvar_SetValueQuick(c, f);
+	BOUNDCVAR(v_gamma, 0.1, 5);
+	BOUNDCVAR(v_contrast, 1, 5);
+	BOUNDCVAR(v_brightness, 0, 0.8);
+	BOUNDCVAR(v_contrastboost, 0.0625, 16);
+	BOUNDCVAR(v_color_black_r, 0, 0.8);
+	BOUNDCVAR(v_color_black_g, 0, 0.8);
+	BOUNDCVAR(v_color_black_b, 0, 0.8);
+	BOUNDCVAR(v_color_grey_r, 0, 0.95);
+	BOUNDCVAR(v_color_grey_g, 0, 0.95);
+	BOUNDCVAR(v_color_grey_b, 0, 0.95);
+	BOUNDCVAR(v_color_white_r, 1, 5);
+	BOUNDCVAR(v_color_white_g, 1, 5);
+	BOUNDCVAR(v_color_white_b, 1, 5);
+#undef BOUNDCVAR
 
-	BOUNDCVAR(v_gamma, 0.1, 5);cachegamma = v_gamma.value;
-	BOUNDCVAR(v_contrast, 1, 5);cachecontrast = v_contrast.value;
-	BOUNDCVAR(v_brightness, 0, 0.8);cachebrightness = v_brightness.value;
-	BOUNDCVAR(v_color_black_r, 0, 0.8);cacheblack[0] = v_color_black_r.value;
-	BOUNDCVAR(v_color_black_g, 0, 0.8);cacheblack[1] = v_color_black_g.value;
-	BOUNDCVAR(v_color_black_b, 0, 0.8);cacheblack[2] = v_color_black_b.value;
-	BOUNDCVAR(v_color_grey_r, 0, 0.95);cachegrey[0] = v_color_grey_r.value;
-	BOUNDCVAR(v_color_grey_g, 0, 0.95);cachegrey[1] = v_color_grey_g.value;
-	BOUNDCVAR(v_color_grey_b, 0, 0.95);cachegrey[2] = v_color_grey_b.value;
-	BOUNDCVAR(v_color_white_r, 1, 5);cachewhite[0] = v_color_white_r.value;
-	BOUNDCVAR(v_color_white_g, 1, 5);cachewhite[1] = v_color_white_g.value;
-	BOUNDCVAR(v_color_white_b, 1, 5);cachewhite[2] = v_color_white_b.value;
-	cachecolorenable = v_color_enable.integer;
-	cachehwgamma = vid_activewindow ? v_hwgamma.integer : 0;
+	// set vid_gammatables_trivial to true if the current settings would generate the identity gamma table
+	vid_gammatables_trivial = false;
+	if(v_psycho.integer == 0)
+	if(v_contrastboost.value == 1)
+	{
+		if(v_color_enable.integer)
+		{
+			if(v_color_black_r.value == 0)
+			if(v_color_black_g.value == 0)
+			if(v_color_black_b.value == 0)
+			if(fabs(v_color_grey_r.value - 0.5) < 1e-6)
+			if(fabs(v_color_grey_g.value - 0.5) < 1e-6)
+			if(fabs(v_color_grey_b.value - 0.5) < 1e-6)
+			if(v_color_white_r.value == 1)
+			if(v_color_white_g.value == 1)
+			if(v_color_white_b.value == 1)
+				vid_gammatables_trivial = true;
+		}
+		else
+		{
+			if(v_gamma.value == 1)
+			if(v_contrast.value == 1)
+			if(v_brightness.value == 0)
+				vid_gammatables_trivial = true;
+		}
+	}
+
+#define GAMMACHECK(cache, value) if (cache != (value)) gamma_changed = true;cache = (value)
+	if(v_psycho.integer)
+		gamma_changed = true;
+	GAMMACHECK(cachegamma      , v_gamma.value);
+	GAMMACHECK(cachecontrast   , v_contrast.value);
+	GAMMACHECK(cachebrightness , v_brightness.value);
+	GAMMACHECK(cachecontrastboost, v_contrastboost.value);
+	GAMMACHECK(cachecolorenable, v_color_enable.integer);
+	GAMMACHECK(cacheblack[0]   , v_color_black_r.value);
+	GAMMACHECK(cacheblack[1]   , v_color_black_g.value);
+	GAMMACHECK(cacheblack[2]   , v_color_black_b.value);
+	GAMMACHECK(cachegrey[0]    , v_color_grey_r.value);
+	GAMMACHECK(cachegrey[1]    , v_color_grey_g.value);
+	GAMMACHECK(cachegrey[2]    , v_color_grey_b.value);
+	GAMMACHECK(cachewhite[0]   , v_color_white_r.value);
+	GAMMACHECK(cachewhite[1]   , v_color_white_g.value);
+	GAMMACHECK(cachewhite[2]   , v_color_white_b.value);
+
+	if(gamma_changed)
+		++vid_gammatables_serial;
+
+	GAMMACHECK(cachehwgamma    , wantgamma);
+#undef GAMMACHECK
+
+	if (!force && !gamma_forcenextframe && !gamma_changed)
+		return;
 
 	gamma_forcenextframe = false;
 
@@ -797,57 +983,7 @@ void VID_UpdateGamma(qboolean force, int rampsize)
 			VID_GetGamma(vid_systemgammaramps, vid_gammarampsize);
 		}
 
-		if (cachecolorenable)
-		{
-			BuildGammaTable16(1.0f, invpow(0.5, 1 - cachegrey[0]), cachewhite[0], cacheblack[0], vid_gammaramps, rampsize);
-			BuildGammaTable16(1.0f, invpow(0.5, 1 - cachegrey[1]), cachewhite[1], cacheblack[1], vid_gammaramps + vid_gammarampsize, rampsize);
-			BuildGammaTable16(1.0f, invpow(0.5, 1 - cachegrey[2]), cachewhite[2], cacheblack[2], vid_gammaramps + vid_gammarampsize*2, rampsize);
-		}
-		else
-		{
-			BuildGammaTable16(1.0f, cachegamma, cachecontrast, cachebrightness, vid_gammaramps, rampsize);
-			BuildGammaTable16(1.0f, cachegamma, cachecontrast, cachebrightness, vid_gammaramps + vid_gammarampsize, rampsize);
-			BuildGammaTable16(1.0f, cachegamma, cachecontrast, cachebrightness, vid_gammaramps + vid_gammarampsize*2, rampsize);
-		}
-
-		// LordHavoc: this code came from Ben Winslow and Zinx Verituse, I have
-		// immensely butchered it to work with variable framerates and fit in with
-		// the rest of darkplaces.
-		if (v_psycho.integer)
-		{
-			int x, y;
-			float t;
-			static float n[3], nd[3], nt[3];
-			static int init = true;
-			unsigned short *ramp;
-			gamma_forcenextframe = true;
-			if (init)
-			{
-				init = false;
-				for (x = 0;x < 3;x++)
-				{
-					n[x] = lhrandom(0, 1);
-					nd[x] = (rand()&1)?-0.25:0.25;
-					nt[x] = lhrandom(1, 8.2);
-				}
-			}
-
-			for (x = 0;x < 3;x++)
-			{
-				nt[x] -= cl.realframetime;
-				if (nt[x] < 0)
-				{
-					nd[x] = -nd[x];
-					nt[x] += lhrandom(1, 8.2);
-				}
-				n[x] += nd[x] * cl.realframetime;
-				n[x] -= floor(n[x]);
-			}
-
-			for (x = 0, ramp = vid_gammaramps;x < 3;x++)
-				for (y = 0, t = n[x] - 0.75f;y < vid_gammarampsize;y++, t += 0.75f * (2.0f / vid_gammarampsize))
-					*ramp++ = (unsigned short)(cos(t*(M_PI*2.0)) * 32767.0f + 32767.0f);
-		}
+		VID_BuildGammaTables(vid_gammaramps, vid_gammarampsize);
 
 		// set vid_hardwaregammasupported to true if VID_SetGamma succeeds, OR if vid_hwgamma is >= 2 (forced gamma - ignores driver return value)
 		Cvar_SetValueQuick(&vid_hardwaregammasupported, VID_SetGamma(vid_gammaramps, vid_gammarampsize) || cachehwgamma >= 2);
@@ -887,6 +1023,7 @@ void VID_Shared_Init(void)
 	Cvar_RegisterVariable(&vid_hardwaregammasupported);
 	Cvar_RegisterVariable(&v_gamma);
 	Cvar_RegisterVariable(&v_brightness);
+	Cvar_RegisterVariable(&v_contrastboost);
 	Cvar_RegisterVariable(&v_contrast);
 
 	Cvar_RegisterVariable(&v_color_enable);
@@ -901,6 +1038,7 @@ void VID_Shared_Init(void)
 	Cvar_RegisterVariable(&v_color_white_b);
 
 	Cvar_RegisterVariable(&v_hwgamma);
+	Cvar_RegisterVariable(&v_glslgamma);
 
 	Cvar_RegisterVariable(&v_psycho);
 
@@ -908,10 +1046,15 @@ void VID_Shared_Init(void)
 	Cvar_RegisterVariable(&vid_width);
 	Cvar_RegisterVariable(&vid_height);
 	Cvar_RegisterVariable(&vid_bitsperpixel);
+	Cvar_RegisterVariable(&vid_samples);
 	Cvar_RegisterVariable(&vid_refreshrate);
+	Cvar_RegisterVariable(&vid_userefreshrate);
 	Cvar_RegisterVariable(&vid_stereobuffer);
 	Cvar_RegisterVariable(&vid_vsync);
 	Cvar_RegisterVariable(&vid_mouse);
+	Cvar_RegisterVariable(&vid_grabkeyboard);
+	Cvar_RegisterVariable(&vid_stick_mouse);
+	Cvar_RegisterVariable(&vid_resizable);
 	Cvar_RegisterVariable(&vid_minwidth);
 	Cvar_RegisterVariable(&vid_minheight);
 	Cvar_RegisterVariable(&gl_combine);
@@ -922,22 +1065,27 @@ void VID_Shared_Init(void)
 		Cvar_Set("gl_combine", "0");
 }
 
-int VID_Mode(int fullscreen, int width, int height, int bpp, int refreshrate, int stereobuffer)
+int VID_Mode(int fullscreen, int width, int height, int bpp, int refreshrate, int stereobuffer, int samples)
 {
-	Con_Printf("Video: %s %dx%dx%dx%dhz%s\n", fullscreen ? "fullscreen" : "window", width, height, bpp, refreshrate, stereobuffer ? " stereo" : "");
-	if (VID_InitMode(fullscreen, width, height, bpp, refreshrate, stereobuffer))
+	cl_ignoremousemoves = 2;
+	Con_Printf("Initializing Video Mode: %s %dx%dx%dx%dhz%s%s\n", fullscreen ? "fullscreen" : "window", width, height, bpp, refreshrate, stereobuffer ? " stereo" : "", samples > 1 ? va(" (%ix AA)", samples) : "");
+	if (VID_InitMode(fullscreen, width, height, bpp, vid_userefreshrate.integer ? max(1, refreshrate) : 0, stereobuffer, samples))
 	{
 		vid.fullscreen = fullscreen;
 		vid.width = width;
 		vid.height = height;
 		vid.bitsperpixel = bpp;
+		vid.samples = samples;
 		vid.refreshrate = refreshrate;
 		vid.stereobuffer = stereobuffer;
+		vid.userefreshrate = vid_userefreshrate.integer;
 		Cvar_SetValueQuick(&vid_fullscreen, fullscreen);
 		Cvar_SetValueQuick(&vid_width, width);
 		Cvar_SetValueQuick(&vid_height, height);
 		Cvar_SetValueQuick(&vid_bitsperpixel, bpp);
-		Cvar_SetValueQuick(&vid_refreshrate, refreshrate);
+		Cvar_SetValueQuick(&vid_samples, samples);
+		if(vid_userefreshrate.integer)
+			Cvar_SetValueQuick(&vid_refreshrate, refreshrate);
 		Cvar_SetValueQuick(&vid_stereobuffer, stereobuffer);
 		return true;
 	}
@@ -965,20 +1113,31 @@ void VID_Restart_f(void)
 	if (vid_commandlinecheck)
 		return;
 
-	Con_Printf("VID_Restart: changing from %s %dx%dx%dbpp, to %s %dx%dx%dbpp.\n",
-		vid.fullscreen ? "fullscreen" : "window", vid.width, vid.height, vid.bitsperpixel,
-		vid_fullscreen.integer ? "fullscreen" : "window", vid_width.integer, vid_height.integer, vid_bitsperpixel.integer);
+	Con_Printf("VID_Restart: changing from %s %dx%dx%dbpp%s%s, to %s %dx%dx%dbpp%s%s.\n",
+		vid.fullscreen ? "fullscreen" : "window", vid.width, vid.height, vid.bitsperpixel, vid.fullscreen && vid_userefreshrate.integer ? va("x%ihz", vid.refreshrate) : "", vid.samples > 1 ? va(" (%ix AA)", vid.samples) : "",
+		vid_fullscreen.integer ? "fullscreen" : "window", vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_fullscreen.integer && vid_userefreshrate.integer ? va("x%ihz", vid_refreshrate.integer) : "", vid_samples.integer > 1 ? va(" (%ix AA)", vid_samples.integer) : "");
 	VID_CloseSystems();
 	VID_Shutdown();
-	if (!VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_refreshrate.integer, vid_stereobuffer.integer))
+	if (!VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_refreshrate.integer, vid_stereobuffer.integer, vid_samples.integer))
 	{
 		Con_Print("Video mode change failed\n");
-		if (!VID_Mode(vid.fullscreen, vid.width, vid.height, vid.bitsperpixel, vid.refreshrate, vid.stereobuffer))
+		if (!VID_Mode(vid.fullscreen, vid.width, vid.height, vid.bitsperpixel, vid.refreshrate, vid.stereobuffer, vid.samples))
 			Sys_Error("Unable to restore to last working video mode");
 	}
 	VID_OpenSystems();
 }
 
+const char *vidfallbacks[][2] =
+{
+	{"vid_stereobuffer", "0"},
+	{"vid_samples", "1"},
+	{"vid_userefreshrate", "0"},
+	{"vid_width", "640"},
+	{"vid_height", "480"},
+	{"vid_bitsperpixel", "16"},
+	{NULL, NULL}
+};
+
 // this is only called once by Host_StartVideo
 void VID_Start(void)
 {
@@ -1014,20 +1173,15 @@ void VID_Start(void)
 			Cvar_SetQuick(&vid_bitsperpixel, com_argv[i+1]);
 	}
 
-	Con_Print("Starting video system\n");
-	success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_refreshrate.integer, vid_stereobuffer.integer);
+	success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_refreshrate.integer, vid_stereobuffer.integer, vid_samples.integer);
 	if (!success)
 	{
 		Con_Print("Desired video mode fail, trying fallbacks...\n");
-		success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, 60, vid_stereobuffer.integer);
-		if (!success && vid_stereobuffer.integer)
-			success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_refreshrate.integer, false);
-		if (!success && vid_bitsperpixel.integer > 16)
-			success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, 16, 60, false);
-		if (!success && (vid_width.integer > 640 || vid_height.integer > 480))
-			success = VID_Mode(vid_fullscreen.integer, 640, 480, 16, 60, false);
-		if (!success && vid_fullscreen.integer)
-			success = VID_Mode(false, 640, 480, 16, 60, false);
+		for (i = 0;!success && vidfallbacks[i][0] != NULL;i++)
+		{
+			Cvar_Set(vidfallbacks[i][0], vidfallbacks[i][1]);
+			success = VID_Mode(vid_fullscreen.integer, vid_width.integer, vid_height.integer, vid_bitsperpixel.integer, vid_refreshrate.integer, vid_stereobuffer.integer, vid_samples.integer);
+		}
 		if (!success)
 			Sys_Error("Video modes failed");
 	}
